@@ -13,10 +13,7 @@
 
 #-sb-assembling ; avoid redefinition warning
 (macrolet ((static-fun-addr (name)
-             #-immobile-code
-             `(progn
-                (inst mov rax-tn (thread-slot-ea sb-vm::thread-linkage-table-slot))
-                (ea (make-fixup ,name :linkage-cell) rax-tn))
+             #-immobile-code `(ea (make-fixup ,name :linkage-cell) null-tn)
              ;; Caution: this looks like it jumps to the linkage cell's address,
              ;; and that is indeed what it would do if it were not for the fact that
              ;; genesis recognizes that this isn't right, and instead does what you mean.
@@ -176,6 +173,9 @@
 
 ;;;; comparison
 
+(eval-when (:compile-toplevel)
+  (assert (minusp (static-symbol-offset t))))
+
 (macrolet ((define-cond-assem-rtn (name translate static-fn test)
              `(define-assembly-routine (,name
                                         (:translate ,translate)
@@ -196,12 +196,12 @@
 
                 DO-STATIC-FUN
                 (call-static-fun ',static-fn 2)
-                ;; HACK: We depend on NIL having the lowest address of all
-                ;; static symbols (including T)
+                ;; X now holds T or NIL corresponding to the answer but it needs
+                ;; to be returned as :L or :G in EFLAGS. We rely on address of T
+                ;; being less address of NIL (asserted above)
                 ,@(ecase test
-                    (:l `((inst mov y (1+ nil-value))
-                          (inst cmp y x)))
-                    (:g `((inst cmp x (1+ nil-value))))))))
+                    (:l `((inst cmp x null-tn)))
+                    (:g `((inst cmp null-tn x)))))))
   (define-cond-assem-rtn generic-< < two-arg-< :l)
   (define-cond-assem-rtn generic-> > two-arg-> :g))
 
@@ -224,7 +224,8 @@
 
   DO-STATIC-FUN
   (call-static-fun 'two-arg-= 2)
-  (inst cmp x (+ nil-value (static-symbol-offset t))))
+  (inst sub x null-tn)
+  (inst cmp x (static-symbol-offset t)))
 
 #+sb-assembling
 (define-assembly-routine (logcount)
