@@ -9,24 +9,32 @@
 ;;; but is more friendly to gdb. There may be some subtle bugs with regard to
 ;;; blocking/unblocking of async signals which arrive nearly around the same
 ;;; time as a synchronous trap.
-#+sw-int-avoidance ; "software interrupt avoidance"
+;;; Partal avoidance will use a call for handle-pending-interrupt after a
+;;; pseudo-atomic sequence, and will emit trapping instructions for TYPE-ERROR
+;;; and other traps.
+;;; For either feature, use at your own risk.
+#+(or sw-int-avoidance partial-sw-int-avoidance) ; "software interrupt avoidance"
 (define-assembly-routine (synchronous-trap) ()
   (inst pushf)
   (inst push rbp-tn)
   (inst mov rbp-tn rsp-tn)
-  (inst and rsp-tn (- 16))
-  (inst sub rsp-tn 8) ; PUSHing an odd number of GPRs
+  (inst and rsp-tn (- 32))
   ;; Arrange in the utterly confusing order that a linux signal context has them
   ;; so that we can memcpy() into a context. Push RBX twice to maintain alignment.
-  (regs-pushlist rcx rax rdx rbx rbx rsi rdi r15 r14 r13 r12 r11 r10 r9 r8)
-  ;;                             ^^^ technically this is the slot for RBP
-  (inst sub rsp-tn (* 16 16))
-  (dotimes (i 16) (inst movdqa (ea (* i 16) rsp-tn) (sb-x86-64-asm::get-fpr :xmm i)))
-  (inst lea rdi-tn (ea 24 rbp-tn)) ; stack-pointer at moment of "interrupt"
-  (inst mov rsi-tn rsp-tn)         ; pointer to saved CPU state
-  (inst call (make-fixup "synchronous_trap" :foreign))
-  (dotimes (i 16) (inst movdqa (sb-x86-64-asm::get-fpr :xmm i) (ea (* i 16) rsp-tn)))
-  (inst add rsp-tn (* 16 16))
+  ;; This enum is usually in "/usr/include/x86_64-linux-gnu/sys/ucontext.h"
+  (regs-pushlist rsp rcx rax rdx rbx rbx rsi rdi r15 r14 r13 r12 r11 r10 r9 r8)
+  ;;                                 ^^^ technically this is the slot for RBP
+
+  (do ((i 15 (1- i))) ((< i 0))
+    (when (member i '(15 11 7 3)) (inst sub rsp-tn (* 4 32))) ; 4 32-byte regs
+    (inst vmovaps (ea (* (mod i 4) 32) rsp-tn) (sb-x86-64-asm::get-fpr :ymm i)))
+
+  (call-c "synchronous_trap" rsp-tn (addressof (ea 24 rbp-tn)))
+
+  (dotimes (i 16)
+    (inst vmovaps (sb-x86-64-asm::get-fpr :ymm i) (ea (* (mod i 4) 32) rsp-tn))
+    (when (member i '(15 11 7 3)) (inst add rsp-tn (* 4 32)))) ; 4 32-byte regs
+
   (regs-poplist rcx rax rdx rbx rbx rsi rdi r15 r14 r13 r12 r11 r10 r9 r8)
   (inst leave)
   (inst popf))
@@ -88,7 +96,7 @@
   (with-registers-preserved (c :except (rsi rdi))
     (pseudo-atomic ()
       #-system-tlabs (inst break halt-trap)
-      #+system-tlabs (call-c (make-fixup "switch_to_arena" :foreign) #+win32 rdi-tn #+win32 rsi-tn))))
+      #+system-tlabs (call-c "switch_to_arena" #+win32 rdi-tn #+win32 rsi-tn))))
 
 (macrolet ((def-routine-pair (name&options vars &body code)
              `(progn
@@ -104,41 +112,30 @@
 
 (def-routine-pair (alloc-tramp) ()
   (with-registers-preserved (c)
-    (call-c (make-fixup "alloc" :foreign)
-            (ea 16 rbp-tn)
-            system-tlab-p)
+    (call-c "alloc" (ea 16 rbp-tn) system-tlab-p)
     (inst mov (ea 16 rbp-tn) rax-tn))) ; result onto stack
 
 (def-routine-pair (list-alloc-tramp) () ; CONS, ACONS, LIST, LIST*
   (with-registers-preserved (c)
-    (call-c (make-fixup "alloc_list" :foreign)
-            (ea 16 rbp-tn)
-            system-tlab-p)
+    (call-c "alloc_list" (ea 16 rbp-tn) system-tlab-p)
     (inst mov (ea 16 rbp-tn) rax-tn))) ; result onto stack
 
 (def-routine-pair (listify-&rest (:return-style :none)) ()
   (with-registers-preserved (c)
-    (call-c (make-fixup "listify_rest_arg" :foreign)
-            (ea 16 rbp-tn)
-            (ea 24 rbp-tn)
-            system-tlab-p)
+    (call-c "listify_rest_arg" (ea 16 rbp-tn) (ea 24 rbp-tn) system-tlab-p)
     (inst mov (ea 24 rbp-tn) rax-tn))   ; result
   (inst ret 8)) ; pop one argument; the unpopped word now holds the result
 
 (def-routine-pair (make-list (:return-style :none)) ()
   (with-registers-preserved (c)
-    (call-c (make-fixup "make_list" :foreign)
-            (ea 16 rbp-tn)
-            (ea 24 rbp-tn)
-            system-tlab-p)
+    (call-c "make_list" (ea 16 rbp-tn) (ea 24 rbp-tn) system-tlab-p)
     (inst mov (ea 24 rbp-tn) rax-tn)) ; result
   (inst ret 8)) ; pop one argument; the unpopped word now holds the result
 )
 
 (define-assembly-routine (alloc-funinstance) ()
   (with-registers-preserved (c)
-    (call-c (make-fixup "alloc_funinstance" :foreign)
-            (ea 16 rbp-tn))
+    (call-c "alloc_funinstance" (ea 16 rbp-tn))
     (inst mov (ea 16 rbp-tn) rax-tn)))
 
 ;;; These routines are for the deterministic consing profiler.
@@ -147,22 +144,13 @@
   (with-registers-preserved (c)
     #+sb-thread
     (pseudo-atomic ()
-      (call-c (make-fixup "allocation_tracker_counted" :foreign)
-              (* (ea 8 rbp-tn))))))
+      (call-c "allocation_tracker_counted" (addressof (ea 8 rbp-tn))))))
 
 (define-assembly-routine (enable-sized-alloc-counter) ()
   (with-registers-preserved (c)
     #+sb-thread
     (pseudo-atomic ()
-      (call-c (make-fixup "allocation_tracker_sized" :foreign)
-              (* (ea 8 rbp-tn))))))
-
-(define-assembly-routine (undefined-tramp (:return-style :none))
-    ((:temp rax descriptor-reg rax-offset))
-  (inst pop (ea n-word-bytes rbp-tn))
-  (emit-error-break nil cerror-trap (error-number-or-lose 'undefined-fun-error) (list rax))
-  (inst push (ea n-word-bytes rbp-tn))
-  (inst jmp (ea (- (* closure-fun-slot n-word-bytes) fun-pointer-lowtag) rax)))
+      (call-c "allocation_tracker_sized" (addressof (ea 8 rbp-tn))))))
 
 #+win32
 (define-assembly-routine
@@ -209,25 +197,6 @@
   ;; (I guess we can only THROW at this point, so it doesn't matter)
   (error-call nil 'undefined-alien-fun-error rbx-tn))
 
-;;; the closure trampoline - entered when a global function is a closure
-;;; and the function is called "by name" (normally, as when it is the
-;;; head of a form) via an FDEFN. Register %RAX holds the fdefn address,
-;;; but the simple-fun which underlies the closure expects %RAX to be the
-;;; closure itself. So we grab the closure out of the fdefn pointed to,
-;;; then jump to the simple-fun that the closure points to.
-;;;
-;;; Immobile code uses a different strategy to call a closure that has been
-;;; installed as a globally named function. The fdefn contains a jump opcode
-;;; to a tiny code component specific to the particular closure.
-;;; The trampoline is responsible for loading RAX, since named calls don't.
-;;; However, #+immobile-code might still need CLOSURE-TRAMP for any fdefn
-;;; for which the compiler chooses not to use "direct" call convention.
-(define-assembly-routine
-    (closure-tramp (:return-style :none))
-    ()
-  (loadw rax-tn rax-tn fdefn-fun-slot other-pointer-lowtag)
-  (inst jmp (object-slot-ea rax-tn closure-fun-slot fun-pointer-lowtag)))
-
 #+debug-gc-barriers
 (define-assembly-routine (check-barrier (:return-style :none)) ()
   (inst push rax-tn)
@@ -239,10 +208,7 @@
   (inst ret 24)
   check
   (with-registers-preserved (c :except fp) ;; shouldn't have any fp operations
-    (call-c (make-fixup "check_barrier" :foreign)
-            (ea 16 rbp-tn)
-            (ea 24 rbp-tn)
-            (ea 32 rbp-tn)))
+    (call-c "check_barrier" (ea 16 rbp-tn) (ea 24 rbp-tn) (ea 32 rbp-tn)))
   (inst ret 24))
 
 ;;; Perform a store to code, updating the GC card mark bit.
@@ -269,12 +235,6 @@
     (pseudo-atomic ()
       #+immobile-space
       (progn
-        #-sb-thread
-        (let ((fixup (make-fixup "all_threads" :foreign-dataref)))
-          ;; Load THREAD-BASE-TN from the all_threads. Does not need to be spilled
-          ;; to stack, because we do do not give the register allocator access to it.
-          (inst mov thread-tn (rip-relative-ea fixup))
-          (inst mov thread-tn (ea thread-tn)))
         (inst mov rax object)
         (inst sub rax (thread-slot-ea thread-text-space-addr-slot))
         (inst shr rax (1- (integer-length immobile-card-bytes)))
@@ -285,9 +245,7 @@
         (inst jmp store))
       TRY-DYNAMIC-SPACE
       (inst mov rax object)
-      (inst shr rax gencgc-card-shift)
-      (inst and :dword rax card-index-mask)
-      (inst mov :byte (ea gc-card-table-reg-tn rax) CARD-MARKED)
+      (mark-gc-card rax)
       STORE
       (inst mov rdi object)
       (inst mov rdx word-index)
