@@ -53,7 +53,7 @@
   (progn
     (setq *break-on-signals* nil)
     (setq sb-kernel::*current-error-depth* 0))
-  (setq *stack-top-hint* nil))
+  (setq sb-debug:*stack-top-hint* nil))
 
 (defun !printer-control-init ()
   (setq *print-readably* nil
@@ -63,6 +63,7 @@
         *print-radix* nil
         *print-vector-length* nil
         *print-circle* nil
+        *print-circle-not-shared* nil
         *print-case* :upcase
         *print-array* t
         *print-gensym* t
@@ -75,33 +76,28 @@
 
 ;;; Create a stream that works early.
 (defun !make-cold-stderr-stream ()
-  (let ((stderr
-          #-win32 2
-          #+win32 (sb-win32::get-std-handle-or-null sb-win32::+std-error-handle+))
-        (buf (make-string 1 :element-type 'base-char :initial-element #\Space)))
-    (%make-fd-stream
-     :cout (lambda (stream ch)
+  (let* ((stderr
+           #-win32 2
+           #+win32 (sb-win32::get-std-handle-or-null sb-win32::+std-error-handle+))
+         (ucs2 #+win32 (sb-win32::console-handle-p stderr))
+         (char-size (if ucs2 4 1))
+         (buf (make-string 1 :element-type (if ucs2 'character 'base-char) :initial-element #\Space)))
+    (flet ((cout (stream ch)
              (declare (ignore stream))
              (setf (char buf 0) ch)
-             (sb-unix:unix-write stderr buf 0 1)
-             ch)
-     :sout (lambda (stream string start end)
-             (declare (ignore stream))
-             (flet ((out (s start len)
-                      (when (plusp len)
-                        (setf (char buf 0) (char s (+ start len -1))))
-                      (sb-unix:unix-write stderr s start len)))
-               (if (typep string 'simple-base-string)
-                   (out string start (- end start))
-                   (let ((n (- end start)))
-                     ;; will croak if there is any non-BASE-CHAR in the string
-                     (out (replace (make-array n :element-type 'base-char)
-                                   string :start2 start) 0 n)))))
-     :misc (lambda (stream operation arg1)
-             (declare (ignore stream arg1))
-             (stream-misc-case (operation :default nil)
-               (:charpos ; impart just enough smarts to make FRESH-LINE dtrt
-                (if (eql (char buf 0) #\newline) 0 1)))))))
+             (sb-unix:unix-write stderr buf 0 char-size)
+             ch))
+      (%make-fd-stream
+       :cout #'cout
+       :sout (lambda (stream string start end)
+               (declare (simple-string string))
+               (loop for i from start below end
+                     do (cout stream (char string i))))
+       :misc (lambda (stream operation arg1)
+               (declare (ignore stream arg1))
+               (stream-misc-case (operation :default nil)
+                                 (:charpos ; impart just enough smarts to make FRESH-LINE dtrt
+                                  (if (eql (char buf 0) #\newline) 0 1))))))))
 
 (defun !xc-sanity-checks ()
   ;; Verify on startup that some constants were dumped reflecting the
@@ -116,14 +112,13 @@
                                       :if-does-not-exist nil)
                 (when stream
                   (let ((*package* (find-package "SB-KERNEL"))) (read stream))))))
+    (dolist (item list)
+      (destructuring-bind (object hash) item
+        (let ((calc (sxhash object)))
+          (unless (= hash calc)
+            (error "cross-compiler SXHASH failure on ~S: ~X vs ~X" object hash calc)))))
     (when list
-      (dolist (item list)
-        (destructuring-bind (object hash) item
-          (unless (= (sxhash object) hash)
-            (error "SXHASH computed wrong answer for ~S. Got ~x should be ~x"
-                   object hash (sxhash object)))))
-      (format t "~&cross-compiler SXHASH tests passed: ~D cases~%"
-              (length list)))))
+      (format t "~&cross-compiler SXHASH tests passed: ~D cases~%" (length list)))))
 
 ;;; called when a cold system starts up
 (defun !cold-init ()
@@ -149,19 +144,14 @@
   ;; debugging.
   (show-and-call !format-cold-init)
   (unless (!c-runtime-noinform-p)
-    ;; I'd like FORMAT to remain working in cold-init, where it does work,
-    ;; hence the conditional.
-    #+(or x86 x86-64) (format t "COLD-INIT... ")
-    #-(or x86 x86-64) (write-string "COLD-INIT... "))
+    (write-string "COLD-INIT... "))
 
   ;; Anyone might call RANDOM to initialize a hash value or something;
   ;; and there's nothing which needs to be initialized in order for
   ;; this to be initialized, so we initialize it right away.
   (show-and-call !random-cold-init)
 
-  ;; We can't do such much as a simple PROCLAIM without this global
-  ;; hash-table (because of WARN-IF-INLINE-FAILED/PROCLAIM)
-  (setf sb-c::*emitted-full-calls* (make-hash-table :test 'equal :synchronized t))
+  (setq sb-c::*compilation-unit* nil) ; its DEFVAR is not processed yet
 
   ;; All sorts of things need INFO and/or (SETF INFO).
   (/show0 "about to SHOW-AND-CALL !GLOBALDB-COLD-INIT")
@@ -172,10 +162,14 @@
   ;; And now *CURRENT-THREAD*
   (sb-thread::init-main-thread)
 
+  (show-and-call !hash-table-cold-init)
+
   ;; not sure why this is needed on some architectures. Dark magic.
+  #-linkage-space
   (setf (fdefn-fun (find-or-create-fdefn '%coerce-callable-for-call))
         #'%coerce-callable-to-fun)
   (show-and-call !loader-cold-init)
+  #+linkage-space (show-and-call sb-vm::!initialize-lisp-linkage-table)
   ;; Assert that FBOUNDP doesn't choke when its answer is NIL.
   ;; It was fine if T because in that case the legality of the arg is certain.
   ;; And be extra paranoid - ensure that it really gets called.
@@ -202,6 +196,13 @@
   ;; Must be done before any non-opencoded array references are made.
   (show-and-call sb-vm::!hairy-data-vector-reffer-init)
 
+  ;; It's not unreasonable to call PARSE-INTEGER in cold-init, which demands that
+  ;; WHITESPACE[1]P work properly, so it needs *STANDARD-READTABLE* established.
+  (let ((*readtable* (make-readtable)))
+    (show-and-call !reader-cold-init)
+    ;; *STANDARD-READTABLE* is assigned last to avoid an error about altering it.
+    (setf *standard-readtable* *readtable*))
+
   ;; Various toplevel forms call MAKE-ARRAY, which calls SUBTYPEP, so
   ;; the basic type machinery needs to be initialized before toplevel
   ;; forms run.
@@ -219,8 +220,7 @@
   (show-and-call sb-kernel::!set-up-structure-object-class)
 
   (unless (!c-runtime-noinform-p)
-    #+(or x86 x86-64) (format t "[Length(TLFs)=~D]" (length *!cold-toplevels*))
-    #-(or x86 x86-64) (write `("Length(TLFs)=" ,(length *!cold-toplevels*)) :escape nil))
+    (write `("Length(TLFs)=" ,(length *!cold-toplevels*)) :escape nil))
 
   (setq sb-pcl::*!docstrings* nil) ; needed before any documentation is set
   (setq sb-c::*queued-proclaims* nil) ; needed before any proclaims are run
@@ -259,12 +259,6 @@
 
   (makunbound '*!cold-toplevels*) ; so it gets GC'd
 
-  ;; Need the static-space replica of the assembly routine jump vector
-  ;; filled in, and the static space vector of static fdefns.
-  ;; This matters only for code that gets compiled to dynamic space,
-  ;; so it's OK that it occurs somewhat late in cold-init.
-  #+x86-64 (sb-vm::validate-asm-routine-vector)
-
   #+win32 (show-and-call reinit-internal-real-time)
 
   ;; Set sane values again, so that the user sees sane values instead
@@ -281,8 +275,15 @@
 
   ;; We run through queued-up type and ftype proclaims that were made
   ;; before the type system was initialized, and (since it is now
-  ;; initalized) reproclaim them..
-  (mapcar #'proclaim sb-c::*queued-proclaims*)
+  ;; initalized) reproclaim them.
+  (loop for claim in sb-c::*queued-proclaims*
+        do
+        (when (eq (car claim) 'ftype)
+          ;; Avoid warnings about mismatched types.
+          (loop for name in (cddr claim)
+                do (setf (info :function :where-from name) :assumed)))
+        (proclaim claim))
+
   (makunbound 'sb-c::*queued-proclaims*)
 
   (show-and-call os-cold-init-or-reinit)
@@ -305,9 +306,6 @@
     (show-and-call !reader-cold-init)
     (show-and-call !sharpm-cold-init)
     (show-and-call !backq-cold-init)
-    ;; The *STANDARD-READTABLE* is assigned at last because the above
-    ;; functions would operate on the standard readtable otherwise---
-    ;; which would result in an error.
     (setf *standard-readtable* *readtable*))
   (setf *readtable* (copy-readtable *standard-readtable*))
   (setf sb-debug:*debug-readtable* (copy-readtable *standard-readtable*))
@@ -322,6 +320,7 @@
   (show-and-call sb-disassem::!compile-inst-printers)
 
   ;; Toggle some readonly bits
+  #-sb-devel
   (dovector (sc sb-c:*backend-sc-numbers*)
     (when sc
       (logically-readonlyize (sb-c::sc-move-funs sc))
@@ -407,7 +406,6 @@ process to continue normally."
     ;; can be called, as pretty much anything can assume that it is set.
     (when total ; newly started process, and not a failed save attempt
       (sb-thread::init-main-thread)
-      #+x86-64 (sb-vm::validate-asm-routine-vector)
       (rebuild-package-vector))
     ;; Initialize streams next, so that any errors can be printed
     (stream-reinit t)
@@ -418,6 +416,7 @@ process to continue normally."
     (setf (extern-alien "internal_errors_enabled" int) 1)
     (float-cold-init-or-reinit))
   (gc-reinit)
+  (finalizers-reinit)
   (foreign-reinit)
   #+win32 (reinit-internal-real-time)
   ;; If the debugger was disabled in the saved core, we need to
