@@ -329,6 +329,8 @@
          :operation '/
          :operands (list number 0)))
 
+(defvar *type-error-no-check-restart* nil)
+
 (defun restart-type-error (type condition &optional pc-offset)
   (let ((tn-offset (car *current-internal-error-args*)))
     (labels ((retry-value (value)
@@ -340,7 +342,7 @@
                                         :context "while restarting a type error."))))
              (set-value (value)
                (sb-di::sub-set-debug-var-slot
-                nil tn-offset (retry-value value)
+                nil tn-offset value
                 *current-internal-error-context*)
                (when pc-offset
                  (sb-vm::incf-context-pc *current-internal-error-context*
@@ -352,8 +354,9 @@
                    :report (lambda (stream)
                              (format stream "Use specified value."))
                    :interactive read-evaluated-form
-                   (set-value value)))))
-      (try condition))))
+                   (set-value (retry-value value))))))
+      (let ((*type-error-no-check-restart* #'set-value))
+        (try condition)))))
 
 (defun object-not-type-error (object type &optional (context nil context-p))
   (if (invalid-array-p object)
@@ -395,9 +398,9 @@
                                                       (cdr context)
                                                       context)))))
               (cond ((typep context '(cons integer))
-                     (restart-type-error type condition (car context)))
+                     (restart-type-error expected-type condition (car context)))
                     ((eq context 'cerror)
-                     (restart-type-error type condition))
+                     (restart-type-error expected-type condition))
                     (t
                      (error condition))))))))
 
@@ -593,6 +596,10 @@
           (object-not-type-error (- x) type nil)
           (object-not-type-error x 'number nil)))))
 
+(deferr op-not-type2-error (a b)
+  (destructuring-bind (type . op) (sb-di:error-context)
+   (object-not-type-error (funcall op a b) type nil)))
+
 (deferr fill-pointer-error (array)
   (declare (notinline fill-pointer-error))
   (if (and (arrayp array)
