@@ -165,7 +165,7 @@
   (def symbol-package-id)
   (def symbol-hash)
   (def symbol-%info) ; primitive reader always needs a stub
-  #-(or x86 x86-64) (def lra-code-header)
+  #-(or x86 x86-64 arm64 riscv) (def lra-code-header)
   (def %make-lisp-obj)
   (def get-lisp-obj-address)
   #+x86-64
@@ -191,6 +191,10 @@
   (def %simd-pack-256-2)
   (def %simd-pack-256-3))
 
+(defun %set-symbol-global-value (symbol value)
+  (declare (type (and symbol (not null)) symbol))
+  (%primitive %set-symbol-global-value symbol value))
+
 (defun spin-loop-hint ()
   "Hints the processor that the current thread is spin-looping."
   (spin-loop-hint))
