@@ -34,9 +34,6 @@
   (declare (explicit-check))
   (%%typep object (specifier-type specifier)))
 
-;;; probably not the right place for this declamation. The benefits
-;;; should be more widespread.
-(declaim (freeze-type ctype))
 (defun %%typep (object type &optional (strict t))
  (declare (type ctype type))
  (typep-impl-macro (object :defaults nil)
@@ -47,7 +44,7 @@
        ((funcallable-instance) (funcallable-instance-p object))
        ((extended-sequence) (extended-sequence-p object))
        ((nil) nil)))
-    (numeric-type (number-typep object type))
+    (numeric-union-type (numeric-union-typep object type))
     (array-type
      (and (arrayp object)
           (or (eq (array-type-complexp type) :maybe)
@@ -162,19 +159,6 @@
           (setf (car cache) fun)
           (funcall fun cache object)))))
 
-;;; Do a type test from a class cell, allowing forward reference and
-;;; redefinition.
-(defun classoid-cell-typep (cell object)
-  (let ((layout (typecase object
-                  (instance (%instance-layout object))
-                  (funcallable-instance (%fun-layout object))
-                  (t (return-from classoid-cell-typep))))
-        (classoid (classoid-cell-classoid cell)))
-    (unless classoid
-      (error "The class ~S has not yet been defined."
-             (classoid-cell-name cell)))
-    (classoid-typep layout classoid object)))
-
 ;;; Return true of any object which is either a funcallable-instance,
 ;;; or an ordinary instance that is not a structure-object.
 (declaim (inline %pcl-instance-p))
@@ -262,6 +246,19 @@
             (when (eq (svref obj-inherits i) layout)
               (return t)))))))
 
+;;; Do a type test from a class cell, allowing forward reference and
+;;; redefinition.
+(defun classoid-cell-typep (cell object)
+  (let ((layout (typecase object
+                  (instance (%instance-layout object))
+                  (funcallable-instance (%fun-layout object))
+                  (t (return-from classoid-cell-typep))))
+        (classoid (classoid-cell-classoid (truly-the classoid-cell cell))))
+    (unless classoid
+      (error "The class ~S has not yet been defined."
+             (classoid-cell-name cell)))
+    (classoid-typep layout classoid object)))
+
 (declaim (end-block))
 
 ;;; If TYPE is a type that we can do a compile-time test on, then
@@ -276,7 +273,7 @@
 (defun ctypep (obj type)
   (declare (type ctype type))
   (typep-impl-macro (obj)
-    ((or numeric-type
+    ((or numeric-union-type
          named-type
          member-type
          character-set-type
@@ -335,6 +332,7 @@
 ;;; :SB-XREF-FOR-INTERNALS hangs on to more symbols. It is not also the intent
 ;;; to retain all toplevel definitions whether subsequently needed or not.
 ;;; That's an unfortunate side-effect; this macro is done being used now.
+#-sb-devel
 (fmakunbound 'typep-impl-macro)
 
 
@@ -435,35 +433,85 @@ Experimental."
        ;; to pass to %MAKE-ARRAY-TYPE but with care we can usually avoid consing.
        (let ((etype (sb-vm::array-element-ctype x))
              (rank (array-rank x))
-             (complexp (not (simple-array-p x))))
-         (case rank
-           (0 (%make-array-type '() complexp etype etype))
-           ((1 2 3)
-            (dx-let ((dims (list (array-dimension x 0) nil nil)))
-              (case rank
-                (1 (setf (cdr dims) nil))
-                (2 (setf (cadr dims) (array-dimension x 1)
-                         (cddr dims) nil))
-                (t (setf (cadr dims) (array-dimension x 1)
-                         (caddr dims) (array-dimension x 2))))
-              (%make-array-type dims complexp etype etype)))
-           (t
-            (let ((dims (make-list rank)))
-              ;; Need ALLOCATE-LIST-ON-STACK for this decl. Can't use vop-exists-p
-              ;; because can't macroexpand into DECLARE. Maybe sharp-dot it ?
-              #+x86-64 (declare (truly-dynamic-extent dims))
-              (dotimes (i rank)
-                (setf (nth i dims) (array-dimension x i)))
-           (%make-array-type dims complexp etype etype))))))
-      (cons (specifier-type 'cons))
-      (character ; Why not return an EQL type?
-       (typecase x
-         (standard-char (specifier-type 'standard-char))
-         (base-char (specifier-type 'base-char))
-         (t (specifier-type 'extended-char))))
+             (complexp (and (not (simple-array-p x))
+                            :maybe)))
+         ;; Complex arrays get turned into simple arrays when compiling to a fasl.
+         (if complexp
+             (%make-array-type (make-list rank :initial-element '*) complexp etype etype)
+             (case rank
+               (0 (%make-array-type '() complexp etype etype))
+               ((1 2 3)
+                (dx-let ((dims (list (array-dimension x 0) nil nil)))
+                  (case rank
+                    (1 (setf (cdr dims) nil))
+                    (2 (setf (cadr dims) (array-dimension x 1)
+                             (cddr dims) nil))
+                    (t (setf (cadr dims) (array-dimension x 1)
+                             (caddr dims) (array-dimension x 2))))
+                  (%make-array-type dims complexp etype etype)))
+               (t
+                (let ((dims (make-list rank)))
+                  ;; Need ALLOCATE-LIST-ON-STACK for this decl. Can't use vop-exists-p
+                  ;; because can't macroexpand into DECLARE. Maybe sharp-dot it ?
+                  #+x86-64 (declare (dynamic-extent dims))
+                  (dotimes (i rank)
+                    (setf (nth i dims) (array-dimension x i)))
+                  (%make-array-type dims complexp etype etype)))))))
+      (cons
+       (let ((car (car x))
+             (cdr (cdr x)))
+         (make-cons-type (cond ((eq car x)
+                                (specifier-type 'cons))
+                               ;; Creates complicated unions
+                               ((functionp car)
+                                (specifier-type 'function))
+                               (t
+                                (ctype-of car)))
+                         (cond ((consp cdr)
+                                (specifier-type 'cons))
+                               ((functionp cdr)
+                                (specifier-type 'function))
+                               (t
+                                (ctype-of cdr))))))
+      (character
+       (character-set-type-from-characters (list x)))
       #+sb-simd-pack
       (simd-pack (simd-subtype (%simd-pack-tag x) simd-pack))
       #+sb-simd-pack-256
       (simd-pack-256 (simd-subtype (%simd-pack-256-tag x) simd-pack-256))
       (t
        (classoid-of x)))))
+
+;;; The stub for sb-c::%structure-is-a should really use layout-id in the same way
+;;; that the vop does, however, because the all 64-bit architectures other than
+;;; x86-64 need to use with-pinned-objects to extract a layout-id, it is cheaper not to.
+;;; I should add a vop for uint32 access to raw slots.
+(defun sb-c::%structure-is-a (object-layout test-layout)
+  (or (eq object-layout test-layout)
+      (let ((depthoid (layout-depthoid test-layout))
+            (inherits (layout-inherits object-layout)))
+        (and (> (length inherits) depthoid)
+             (eq (svref inherits depthoid) test-layout)))))
+
+(defun sb-c::structure-typep (object test-layout)
+  (and (%instancep object)
+       (let ((object-layout (%instance-layout object)))
+        (or (eq object-layout test-layout)
+            (let ((depthoid (layout-depthoid test-layout))
+                  (inherits (layout-inherits object-layout)))
+              (and (> (length inherits) depthoid)
+                   (eq (svref inherits depthoid) test-layout)))))))
+
+;;; TODO: this could be further generalized to handle any type where layout-of correctly
+;;; chooses a clause. It would of course not work for types like MEMBER or (MOD 5)
+(defun %typecase-index (layout-lists object sealed)
+  (declare (ignore sealed))
+  (when (%instancep object)
+    (let ((object-layout (%instance-layout object)))
+      (let ((clause-index 1))
+        (dovector (layouts layout-lists)
+          (dolist (layout layouts)
+            (when (sb-c::%structure-is-a object-layout layout)
+              (return-from %typecase-index clause-index)))
+          (incf clause-index)))))
+  0)
