@@ -93,8 +93,6 @@
     (proclaim-target-optimization)
     (funcall fun)))
 
-(setf *target-compile-file* #'sb-xc:compile-file)
-(setf *target-assemble-file* #'sb-c:assemble-file)
 (setf *in-target-compilation-mode-fn* #'in-target-cross-compilation-mode)
 
 ;; Update the xc-readtable
