@@ -388,7 +388,7 @@
     (:args (value :scs (any-reg descriptor-reg zero) :to :save)
            (symbol :scs (descriptor-reg)
                    :target alloc-tls-symbol))
-    (:temporary (:sc descriptor-reg :offset r8-offset :from (:argument 1)) alloc-tls-symbol)
+    (:temporary (:sc descriptor-reg :offset r10-offset :from (:argument 1)) alloc-tls-symbol)
     (:temporary (:sc non-descriptor-reg :offset nl0-offset) tls-index)
     (:temporary (:sc non-descriptor-reg :offset nl1-offset) free-tls-index)
     (:temporary (:sc non-descriptor-reg :offset lr-offset) lr)
@@ -715,14 +715,14 @@
         (inst lsl pa-flag pa-flag card)
 
         (cond ((member :arm-v8.1 *backend-subfeatures*)
-               (inst ldset pa-flag pa-flag temp))
+               (inst ldset pa-flag zr-tn temp))
               (t
                (let ((loop (gen-label)))
                  (emit-label LOOP)
                  (inst ldaxr card temp)
                  (inst orr card card pa-flag)
-                 (inst stlxr pa-flag card temp)
-                 (inst cbnz (32-bit-reg pa-flag) LOOP))))
+                 (inst stlxr tmp-tn card temp)
+                 (inst cbnz (32-bit-reg tmp-tn) LOOP))))
         (inst b store))
       TRY-DYNAMIC-SPACE
       (load-foreign-symbol temp "gc_card_table_mask" :dataref t)
