@@ -143,6 +143,14 @@
 
 ;;;; Storage allocation:
 
+(defun generate-stack-overflow-check (vop size)
+  (let ((overflow (generate-error-code vop
+                                       'stack-allocated-object-overflows-stack-error
+                                       size)))
+    (loadw tmp-tn thread-tn thread-control-stack-end-slot)
+    (inst sub tmp-tn tmp-tn csp-tn)
+    (inst cmp tmp-tn size)
+    (inst b :le overflow)))
 
 ;;; This is the main mechanism for allocating memory in the lisp heap.
 ;;;
@@ -158,12 +166,14 @@
 ;;; surround a call to ALLOCATION anyway), and to indicate that the
 ;;; P-A FLAG-TN is also acceptable here.
 
-#+gencgc
-(defun allocation-tramp (type alloc-tn size back-label)
+#+generational
+(defun allocation-tramp (type alloc-tn size back-label systemp)
   (if (integerp size)
       (load-immediate-word tmp-tn size)
       (inst mov tmp-tn size))
-  (let ((asm-routine (if (eq type 'list) 'list-alloc-tramp 'alloc-tramp)))
+  (let ((asm-routine (if systemp
+                         (if (eq type 'list) 'sys-list-alloc-tramp 'sys-alloc-tramp)
+                         (if (eq type 'list) 'list-alloc-tramp 'alloc-tramp))))
     (invoke-asm-routine asm-routine alloc-tn))
   (inst b back-label))
 
@@ -172,17 +182,31 @@
 (defun allocation (type size lowtag result-tn
                    &key flag-tn
                         stack-allocate-p
-                        overflow)
+                        overflow
+                        (systemp (system-tlab-p type (sb-c::vop-node sb-assem::*current-vop*))))
   (declare (ignorable type))
   ;; Normal allocation to the heap.
   (if stack-allocate-p
       (assemble ()
-        (inst add tmp-tn csp-tn lowtag-mask)
-        (inst and tmp-tn tmp-tn (lognot lowtag-mask))
-        (inst add csp-tn tmp-tn (add-sub-immediate size result-tn))
-        (inst add result-tn tmp-tn lowtag))
+        (cond ((aligned-stack-p stack-allocate-p)
+               (assemble ()
+                 (inst tst csp-tn lowtag-mask)
+                 (inst b :eq skip)
+                 (inst add csp-tn csp-tn 8)
+                 (error-call nil 'sb-kernel::unreachable-error)
+                 skip)
+               (move tmp-tn csp-tn)
+               (inst add csp-tn csp-tn (add-sub-immediate size result-tn))
+               (inst add result-tn tmp-tn lowtag))
+              (t
+               (inst add tmp-tn csp-tn lowtag-mask)
+               (inst and tmp-tn tmp-tn (lognot lowtag-mask))
+               (inst add csp-tn tmp-tn (add-sub-immediate size result-tn))
+               (inst add result-tn tmp-tn lowtag))))
       (let ((alloc (gen-label))
-            #+sb-thread (tlab (if (eq type 'list) thread-cons-tlab-slot thread-mixed-tlab-slot))
+            #+sb-thread (tlab (if systemp
+                                  (if (eq type 'list) thread-sys-cons-tlab-slot thread-sys-mixed-tlab-slot)
+                                  (if (eq type 'list) thread-cons-tlab-slot thread-mixed-tlab-slot)))
             #-sb-thread (region-offset (if (eq type 'list)
                                            cons-region-offset
                                            mixed-region-offset))
@@ -208,7 +232,8 @@
               (allocation-tramp type
                                 result-tn
                                 size
-                                BACK-FROM-ALLOC))))))
+                                BACK-FROM-ALLOC
+                                systemp))))))
 
 (defmacro with-fixed-allocation ((result-tn flag-tn type-code size
                                             &key (lowtag other-pointer-lowtag)
@@ -297,14 +322,14 @@
 #+sb-safepoint
 (defun emit-safepoint ()
   (inst ldr zr-tn (@ null-tn
-                     (- (+ gc-safepoint-trap-offset n-word-bytes
-                           other-pointer-lowtag)))))
+                     (load-store-offset
+                      (- (+ nil-value-offset gc-safepoint-trap-offset))))))
 
 ;;; handy macro for making sequences look atomic
 (defmacro pseudo-atomic ((flag-tn &key elide-if (sync t)) &body forms)
   (declare (ignorable sync))
   #+sb-safepoint
-  `(progn ,@forms (emit-safepoint))
+  `(progn ,flag-tn (assemble () ,@forms) (unless ,elide-if (emit-safepoint)))
   #-sb-safepoint
   `(progn
      (unless ,elide-if
@@ -342,106 +367,132 @@
 
 (defmacro define-full-reffer (name type offset lowtag scs el-type
                               &optional translate)
-  (let ((lip (if (or (memq 'any-reg scs)
-                     (memq 'descriptor-reg scs))
-                 'lip
-                 'value)))
-    `(define-vop (,name)
-       ,@(when translate
-           `((:translate ,translate)))
-       (:policy :fast-safe)
-       (:args (object :scs (descriptor-reg))
-              (index :scs (any-reg unsigned-reg signed-reg immediate)))
-       (:arg-types ,type tagged-num)
-       ,@(when (eq lip 'lip)
-           '((:temporary (:scs (non-descriptor-reg)
-                          :unused-if (sc-is index immediate)) lip)))
-       (:results (value :scs ,scs))
-       (:result-types ,el-type)
-       (:generator 5
-         (sc-case index
-           (immediate
-            (inst ldr value (@ object (load-store-offset
-                                       (- (ash (+ ,offset (tn-value index)) word-shift)
-                                          ,lowtag)))))
-           (t
-            (inst add ,lip object (lsl index (- word-shift (if (sc-is index any-reg)
-                                                               n-fixnum-tag-bits
-                                                               0))))
-            (loadw value ,lip ,offset ,lowtag)))))))
-
-(defmacro define-full-setter (name type offset lowtag scs el-type
-                              &optional translate)
   `(define-vop (,name)
      ,@(when translate
-             `((:translate ,translate)))
+         `((:translate ,translate)))
      (:policy :fast-safe)
      (:args (object :scs (descriptor-reg))
-            (index :scs (any-reg unsigned-reg signed-reg immediate))
-            (value :scs (,@scs zero)))
-     (:arg-types ,type tagged-num ,el-type)
-     (:temporary (:scs (non-descriptor-reg)
-                  :unused-if (sc-is index immediate)) lip)
-     (:generator 2
+            (index :scs (any-reg unsigned-reg signed-reg immediate)))
+     (:arg-types ,type tagged-num)
+     (:results (value :scs ,scs))
+     (:result-types ,el-type)
+     (:generator 5
        (sc-case index
          (immediate
-          (inst str value (@ object (load-store-offset
+          (inst ldr value (@ object (load-store-offset
                                      (- (ash (+ ,offset (tn-value index)) word-shift)
                                         ,lowtag)))))
          (t
-          (inst add lip object (lsl index (- word-shift (if (sc-is index any-reg)
-                                                            n-fixnum-tag-bits
-                                                            0))))
-          (storew value lip ,offset ,lowtag))))))
+          (inst add tmp-tn object (lsl index (- word-shift (if (sc-is index any-reg)
+                                                               n-fixnum-tag-bits
+                                                               0))))
+          (loadw value tmp-tn ,offset ,lowtag))))))
+
+(defmacro define-full-setter (name type offset lowtag scs el-type
+                              &optional translate
+                              &aux (barrierp (member 'descriptor-reg scs)))
+  (case name
+    ((data-vector-set/simple-vector %weakvec-set)
+     `(define-vop (,name)
+        (:translate ,translate)
+        (:policy :fast-safe)
+        (:args (object :scs (descriptor-reg))
+               (index :scs (any-reg unsigned-reg signed-reg immediate))
+               (value :scs (,@scs zero)))
+        (:arg-types ,type tagged-num ,el-type)
+        (:temporary (:sc non-descriptor-reg) ea)
+        (:vop-var vop)
+        (:generator 2
+          (cond ((require-gengc-barrier-p object (vop-nth-arg 2 vop) value)
+                 (sc-case index
+                   (immediate
+                    (inst add ea object (load-store-offset
+                                         (- (ash (+ ,offset (tn-value index)) word-shift)
+                                            ,lowtag))))
+                   (t
+                    ;; Scale the index
+                    (let ((unshift (if (sc-is index any-reg) n-fixnum-tag-bits 0)))
+                      (inst add ea object (lsl index (- word-shift unshift))))
+                    ;; Calculate the exact cell address to ensure the right card is marked
+                    (inst add ea ea (- (ash vector-data-offset word-shift) other-pointer-lowtag))))
+                 (emit-gengc-barrier object ea tmp-tn t)
+                 (storew value ea 0 0))
+                (t
+                 (sc-case index
+                   (immediate
+                    (inst str value (@ object (load-store-offset
+                                               (- (ash (+ ,offset (tn-value index)) word-shift)
+                                                  ,lowtag)))))
+                   (t
+                    (let ((unshift (if (sc-is index any-reg) n-fixnum-tag-bits 0)))
+                      (inst add tmp-tn object (lsl index (- word-shift unshift))))
+                    (storew value tmp-tn ,offset ,lowtag))))))))
+    (t
+     `(define-vop (,name)
+        (:translate ,translate)
+        (:policy :fast-safe)
+        (:args (object :scs (descriptor-reg))
+               (index :scs (any-reg unsigned-reg signed-reg immediate))
+               (value :scs (,@scs zero)))
+        (:arg-types ,type tagged-num ,el-type)
+        (:vop-var vop)
+        ,@(if barrierp '((:gc-barrier 0 2)
+                         (:info barrier)))
+        (:generator 2
+          ,@(when barrierp
+              '((when barrier
+                  (emit-gengc-barrier object nil tmp-tn t))))
+          (sc-case index
+            (immediate
+             (inst str value (@ object (load-store-offset
+                                        (- (ash (+ ,offset (tn-value index)) word-shift)
+                                           ,lowtag)))))
+            (t
+             (let ((unshift (if (sc-is index any-reg) n-fixnum-tag-bits 0)))
+               (inst add tmp-tn object (lsl index (- word-shift unshift))))
+             (storew value tmp-tn ,offset ,lowtag))))))))
 
 (defmacro define-partial-reffer (name type size signed offset lowtag scs
                                  el-type &optional translate)
-  (let ((lip (if (or (memq 'any-reg scs)
-                     (memq 'descriptor-reg scs))
-                 'lip
-                 'value)))
-    `(define-vop (,name)
-       ,@(when translate
-           `((:translate ,translate)))
-       (:policy :fast-safe)
-       (:args (object :scs (descriptor-reg))
-              (index :scs (any-reg unsigned-reg signed-reg immediate)))
-       (:arg-types ,type tagged-num)
-       (:results (value :scs ,scs))
-       (:result-types ,el-type)
-       ,@(when (eq lip 'lip)
-           '((:temporary (:scs (non-descriptor-reg)
-                          :unused-if (sc-is index immediate)) lip)))
-       (:generator 5
-         ,@(multiple-value-bind (op shift)
-               (ecase size
-                 (:byte
-                  (values (if signed 'ldrsb 'ldrb) 0))
-                 (:short
-                  (values (if signed 'ldrsh 'ldrh) 1))
-                 (:word
-                  (values (if signed 'ldrsw 'ldr) 2))
-                 (:single-float
-                  (values 'ldr 2)))
-             (let ((value (if (and (eq size :word)
-                                   (not signed))
-                              '(32-bit-reg value)
-                              'value)))
-               `((sc-case index
-                   (immediate
-                    (inst ,op ,value (@ object (load-store-offset
-                                                (+
-                                                 (ash (tn-value index) ,shift)
-                                                 (- (* ,offset n-word-bytes) ,lowtag))))))
-                   (t
-                    (let ((shift ,shift))
-                      (when (sc-is index any-reg)
-                        (decf shift n-fixnum-tag-bits))
-                      (inst add ,lip object (if (minusp shift)
+  `(define-vop (,name)
+     ,@(when translate
+         `((:translate ,translate)))
+     (:policy :fast-safe)
+     (:args (object :scs (descriptor-reg))
+            (index :scs (any-reg unsigned-reg signed-reg immediate)))
+     (:arg-types ,type tagged-num)
+     (:results (value :scs ,scs))
+     (:result-types ,el-type)
+     (:generator 5
+       ,@(multiple-value-bind (op shift)
+             (ecase size
+               (:byte
+                (values (if signed 'ldrsb 'ldrb) 0))
+               (:short
+                (values (if signed 'ldrsh 'ldrh) 1))
+               (:word
+                (values (if signed 'ldrsw 'ldr) 2))
+               (:single-float
+                (values 'ldr 2)))
+           (let ((value (if (and (eq size :word)
+                                 (not signed))
+                            '(32-bit-reg value)
+                            'value)))
+             `((sc-case index
+                 (immediate
+                  (inst ,op ,value (@ object (load-store-offset
+                                              (+
+                                               (ash (tn-value index) ,shift)
+                                               (- (* ,offset n-word-bytes) ,lowtag))))))
+                 (t
+                  (let ((shift ,shift))
+                    (when (sc-is index any-reg)
+                      (decf shift n-fixnum-tag-bits))
+                    (inst add tmp-tn object (if (minusp shift)
                                                 (asr index (- shift))
                                                 (lsl index shift)))
-                      (inst ,op
-                            ,value (@ ,lip (- (* ,offset n-word-bytes) ,lowtag)))))))))))))
+                    (inst ,op
+                          ,value (@ tmp-tn (- (* ,offset n-word-bytes) ,lowtag))))))))))))
 
 (defmacro define-partial-setter (name type size offset lowtag scs el-type
                                  &optional translate)
@@ -465,8 +516,6 @@
                   tagged-num
                   ,@(unless setf-p
                       `(,el-type)))
-      (:temporary (:scs (non-descriptor-reg)
-                   :unused-if (sc-is index immediate)) lip)
       (:generator 5
         (when (sc-is value immediate)
           (setf value zr-tn))
@@ -491,11 +540,11 @@
                    (let ((shift ,shift))
                      (when (sc-is index any-reg)
                        (decf shift n-fixnum-tag-bits))
-                     (inst add lip object (if (minusp shift)
-                                              (asr index (- shift))
-                                              (lsl index shift)))
+                     (inst add tmp-tn object (if (minusp shift)
+                                                 (asr index (- shift))
+                                                 (lsl index shift)))
                      (inst ,op
-                           ,value (@ lip (- (* ,offset n-word-bytes) ,lowtag)))))))))))))
+                           ,value (@ tmp-tn (- (* ,offset n-word-bytes) ,lowtag)))))))))))))
 
 (defun load-inline-constant (dst &rest constant-descriptor)
   (inst load-from-label dst (cdr (apply #'register-inline-constant constant-descriptor))))
@@ -525,3 +574,31 @@
   `(inst str ,reg (@ thread-tn ,(info :variable :wired-tls symbol)))
   #-sb-thread
   `(store-symbol-value ,reg ,symbol))
+
+;;; Load constants, stack-values, reusing when possible
+(defmacro maybe-load (tn &optional (temp 'temp))
+  (once-only ((tn tn))
+    `(sc-case ,tn
+       ((any-reg descriptor-reg)
+        ,tn)
+       ((immediate constant)
+        (cond ((not (sb-c::tn-leaf ,tn))
+               (load-constant vop ,tn ,temp)
+               ,temp)
+              ((eql (tn-value ,tn) 0)
+               zr-tn)
+              ((or (eql prev-constant (tn-value ,tn))
+                   (progn
+                     (setf prev-constant (tn-value ,tn))
+                     nil))
+               temp)
+              ((sc-is ,tn constant)
+               (load-constant vop ,tn ,temp)
+               ,temp)
+              (t
+               (load-immediate vop ,tn ,temp)
+               ,temp)))
+       (control-stack
+        (setf prev-constant nil)
+        (load-stack-tn ,temp ,tn)
+        ,temp))))
