@@ -331,10 +331,9 @@
   (etypecase y
     (ctype
        (awhen (lambda-var-ctype-constraints x)
-         (dolist (con (gethash (sb-kernel::type-class y) it) nil)
+         (dolist (con (gethash y it) nil)
            (when (and (eq (constraint-kind con) kind)
-                      (eq (constraint-not-p con) not-p)
-                      (type= (constraint-y con) y))
+                      (eq (constraint-not-p con) not-p))
              (return-from find-constraint con)))
          nil))
     (lvar
@@ -375,7 +374,7 @@
       (ctype
        (let ((index (ensure-hash (lambda-var-ctype-constraints x)))
              (vec   (ensure-vec  (lambda-var-inheritable-constraints x))))
-         (push con (gethash (sb-kernel::type-class y) index))
+         (push con (gethash y index))
          (vector-push-extend con vec)))
       (lvar
        (let ((index (ensure-hash (lambda-var-eq-constraints x))))
@@ -416,7 +415,7 @@
 (declaim (inline type-for-constraints-p))
 (defun type-for-constraints-p (type)
   (not (or (eq type *universal-type*)
-           (contains-hairy-type-p type))))
+           (opaque-type-p type))))
 
 ;;; Actual conset interface
 ;;;
@@ -767,17 +766,20 @@
                         (add-equality-constraints name args
                                                   constraints consequent-constraints alternative-constraints)
                         (case name
-                          ((%typep %instance-typep)
+                          ((typep %typep %instance-typep)
                            (let ((type (second args)))
                              (when (constant-lvar-p type)
-                               (let ((val (lvar-value type)))
-                                 (add 'typep
-                                      (ok-lvar-lambda-var (first args) constraints)
-                                      (if (ctype-p val)
-                                          val
-                                          (let ((*compiler-error-context* node))
-                                            (specifier-type val)))
-                                      nil)))))
+                               (let* ((val (lvar-value type))
+                                      (ctype (if (ctype-p val)
+                                                 val
+                                                 (careful-specifier-type val))))
+                                 (when (and ctype
+                                            (type-for-constraints-p ctype))
+                                   (add 'typep
+                                        (ok-lvar-lambda-var (first args) constraints)
+                                        ctype
+                                        nil
+                                        (first args)))))))
                           ((eq eql)
                            (let* ((arg1 (first args))
                                   (var1 (ok-lvar-lambda-var arg1 constraints))
@@ -855,6 +857,13 @@
        (eq (numeric-type-class x) 'integer)
        (eq (numeric-type-complexp x) :real)))
 
+(defun ratio-type-p (x)
+  (declare (type ctype x))
+  (and (numeric-type-p x)
+       (eq (numeric-type-class x) 'rational)
+       (eq (numeric-type-complexp x) :real)
+       (csubtypep x (specifier-type 'ratio))))
+
 ;;; Given that an inequality holds on values of type X and Y, return a
 ;;; new type for X. If GREATER is true, then X was greater than Y,
 ;;; otherwise less. If OR-EQUAL is true, then the inequality was
@@ -920,14 +929,16 @@
                     (sb-xc:> (type-bound-number ref) (type-bound-number x))))))
     (let* ((x-bound (bound x))
            (y-bound (exclude (bound y)))
-           (new-bound (cond ((not x-bound)
-                             y-bound)
-                            ((not y-bound)
-                             x-bound)
-                            ((tighter-p y-bound x-bound)
-                             y-bound)
-                            (t
-                             x-bound))))
+           (new-bound (coerce-for-bound
+                       (cond ((not x-bound)
+                              y-bound)
+                             ((not y-bound)
+                              x-bound)
+                             ((tighter-p y-bound x-bound)
+                              y-bound)
+                             (t
+                              x-bound))
+                       (numeric-type-format x))))
       (if greater
           (modified-numeric-type x :low new-bound)
           (modified-numeric-type x :high new-bound)))))
@@ -1035,7 +1046,7 @@
                      (setf not-xset (alloc-xset)))
                    (add-to-xset x not-xset))))
            (intersect-result (other-type)
-             (setf type (type-approx-intersection2 type other-type))))
+             (setf type (type-intersection type other-type))))
       (declare (inline intersect-result))
       (do-propagatable-constraints (con (constraints variable))
         (let* ((kind (constraint-kind con))
@@ -1181,19 +1192,10 @@
       (when type
         (unless set
           (setf (lambda-var-unused-initial-value leaf) nil))
-        ;; CHANGE-CLASS can change the type, lower down to standard-object,
+        ;; CHANGE-CLASS can change the type,
         ;; type propagation for classes is not as important anyway.
-        (cond #-sb-xc-host
-              ((and
-                (eq sb-pcl::**boot-state** 'sb-pcl::complete)
-                (block nil
-                  (let ((standard-object (find-classoid 'standard-object)))
-                    (sb-kernel::map-type
-                     (lambda (type)
-                       (when (and (classoid-p type)
-                                  (csubtypep type standard-object))
-                         (return t)))
-                     type)))))
+        (cond ((logtest sb-kernel::ctype-contains-class
+                        (sb-kernel::type-flags type)))
               (t
                (derive-node-type ref
                                  (make-single-value-type type))
@@ -1207,7 +1209,13 @@
         ;; Find unchanged eql refs to a set variable.
         (when (lambda-var-sets leaf)
           (let (mark
-                (eq (lambda-var-eq-constraints leaf)))
+                (old-mark (ref-same-refs ref))
+                (eq (lambda-var-eq-constraints leaf))
+                ;; It's cheap to compute but not cheap to reoptimize
+                ;; everything if it's not needed.
+                ;; Some transforms ask for same-leaf-ref-p after :ir1-phases.
+                (compute (lambda-var-compute-same-refs leaf))
+                reoptimize)
             (when eq
               (loop for other-ref in (leaf-refs leaf)
                     unless (eq other-ref ref)
@@ -1217,9 +1225,14 @@
                            (unless mark
                              (setf mark (list 0))
                              (setf (ref-same-refs ref) mark))
+                           (when (and compute
+                                       (not (and old-mark
+                                                 (eq old-mark (ref-same-refs other-ref)))))
+                             (setf reoptimize t)
+                             (reoptimize-lvar (node-lvar other-ref)))
                            (setf (ref-same-refs other-ref) mark)))))
-            (when mark
-              (reoptimize-node ref))))))))
+            (when reoptimize
+              (reoptimize-lvar (node-lvar ref)))))))))
 
 ;;;; Flow analysis
 
@@ -1253,15 +1266,23 @@
        (let ((fun (bind-lambda node)))
          (functional-kind-case fun
            (let
-            (loop with call = (lvar-dest (node-lvar (first (lambda-refs fun))))
-                  for var in (lambda-vars fun)
-                  and val in (combination-args call)
-                  when (and val (lambda-var-constraints var))
-                  do (let ((type (lvar-type val)))
-                       (when (type-for-constraints-p type)
-                         (conset-add-constraint gen 'typep var type nil)))
-                     (maybe-add-eql-var-var-constraint var val gen)
-                     (add-var-result-constraints var val gen)))
+               (loop with call = (lvar-dest (node-lvar (first (lambda-refs fun))))
+                     for var in (lambda-vars fun)
+                     and val in (combination-args call)
+                     when (and val (lambda-var-constraints var))
+                     do (let ((type (lvar-type val)))
+                          (when (type-for-constraints-p type)
+                            (conset-add-constraint gen 'typep var type nil)))
+                        (maybe-add-eql-var-var-constraint var val gen)
+                        (add-var-result-constraints var val gen)))
+           ((nil optional)
+            (loop for var in (lambda-vars fun)
+                  for type = (leaf-defined-type var)
+                  do
+                  (when (and (lambda-var-constraints var)
+                             (type-for-constraints-p type)
+                             (not (lambda-var-arg-info var)))
+                    (conset-add-constraint gen 'typep var type nil))))
            (mv-let
             (add-mv-let-result-constraints (lvar-dest (node-lvar (first (lambda-refs fun)))) fun gen)))))
       (ref
@@ -1424,10 +1445,7 @@
                  (when (block-type-check (lambda-block fun))
                    ;; This is optimistic, make sure it's going to be
                    ;; processed.
-                   (setf (lambda-var-unused-initial-value var) t))
-                 (loop for ref in (lambda-var-refs var)
-
-                       do (setf (ref-same-refs ref) nil))))))
+                   (setf (lambda-var-unused-initial-value var) t))))))
       (frob fun)
       (dolist (let (lambda-lets fun))
         (frob let)))))
