@@ -56,7 +56,12 @@
   (packages () :type list)
   ;; a table mapping from the ENTRY-INFO structures for dumped XEPs to
   ;; the table offsets of the corresponding code pointers
-  (entry-table (make-hash-table :test 'eq) :type hash-table)
+  (entry-table (make-hash-table :test 'eq
+                                ;; It holds on to clambdas, retaining
+                                ;; them for the duration of
+                                ;; compile-file.
+                                #-sb-xc-host :weakness #-sb-xc-host :key)
+   :type hash-table)
   ;; a table holding back-patching info for forward references to XEPs.
   ;; The key is the ENTRY-INFO structure for the XEP, and the value is
   ;; a list of conses (<code-handle> . <offset>), where <code-handle>
@@ -1080,7 +1085,6 @@
     :code-object
     :layout :immobile-symbol
     #+linkage-space ,@'(:linkage-cell :linkage-cell-ud)
-    :symbol-value
     :layout-id)
   #'equalp)
 
@@ -1165,12 +1169,7 @@
                    (:layout-id (the layout name))
                    ((:assembly-routine
                      :symbol-tls-index
-                     ;; Only #+immobile-space can use the following two flavors.
-                     ;; An :IMMOBILE-SYMBOL fixup references the symbol itself,
-                     ;; whereas a :SYMBOL-VALUE fixup references the value of the symbol.
-                     ;; In the latter case, the symbol's address doesn't matter,
-                     ;; but its global value must be an immobile object.
-                     :immobile-symbol :symbol-value)
+                     :immobile-symbol) ; only #+immobile-space can use
                     (the symbol name))
                    (t name)))) ; function name
             (dump-object operand fasl-output)
