@@ -82,7 +82,9 @@
         ((spaces (append `((read-only ,ro-space-size)
                            #+(and win32 x86-64)
                            (seh-data ,(symbol-value '+backend-page-bytes+) win64-seh-data-addr)
-                           #-immobile-space (alien-linkage ,alien-linkage-space-size)
+                           ;; #+immobile-space implies a relocatable alien linkage space. And x86-64 always
+                           ;; has relocatable linkage tables
+                           #-(or x86-64 immobile-space) (alien-linkage ,alien-linkage-space-size)
                            ;; safepoint on 64-bit uses a relocatable trap page just below the card mark
                            ;; table, which works nicely assuming a register is wired to the card table
                            #+(and sb-safepoint (not x86-64))
@@ -116,13 +118,16 @@
                        `((defconstant ,var-name ,start))
                        (let ((start-sym (symbolicate space "-SPACE-START")))
                          `(,(defconstantish relocatable start-sym start)
-                           ,(cond ((eq space 'alien-linkage-table)) ; nothing for the -END
-                                  ((not relocatable)
-                                   `(defconstant ,(symbolicate space "-SPACE-END") ,end))
-                                  #-sb-xc-host ((eq space 'text)) ; don't emit anything
-                                  (t
-                                   `(defconstant ,(symbolicate space "-SPACE-SIZE")
-                                      ,(- end start)))))))))))
+                           ,@(cond ((member space '(alien-linkage #-sb-xc-host text))
+                                    ;; * alien-linkage-space-size is already defined above
+                                    ;; * text-space-size is only needed in 1st genesis
+                                    ;;   (probably true of most of these constants in fact)
+                                    nil) ; neither -END nor -SIZE constant
+                                   ((and (not (eq space 'static)) (not relocatable))
+                                    `((defconstant ,(symbolicate space "-SPACE-END") ,end)))
+                                   ((or (eq space 'static) relocatable)
+                                    `((defconstant ,(symbolicate space "-SPACE-SIZE")
+                                        ,(- end start))))))))))))
       `(progn
          ,@small-space-forms
          ,(defconstantish t 'dynamic-space-start
@@ -429,5 +434,10 @@
          #'equal))
 #-64-bit (defconstant-eqx sb-impl::symbol-hash-prng-byte (byte 3 0) #'equal)
 
+(defmacro symhash-xor-constant (tagged-nil)
+  (declare (ignorable tagged-nil))
+  #+x86-64 `(dpb 0 (byte n-linkage-index-bits symbol-linkage-index-pos) ,tagged-nil)
+  #-x86-64 0)
+
 (push '("SB-VM" +c-callable-fdefns+ +common-static-symbols+)
       *!removable-symbols*)
