@@ -83,17 +83,17 @@
 (define-vop (move-from-single)
   (:args (x :scs (single-reg) :to :save))
   ;; There are no moves between float regs and gprs, so use the stack
-  (:temporary (:scs (double-stack)) stack-temp)
+  (:temporary (:scs (single-stack)) stack-temp)
   (:results (y :scs (descriptor-reg)))
   (:vop-var vop)
   (:note "float to pointer coercion")
   (:generator 4
-    (inst li y single-float-widetag)
     (let ((base (current-nfp-tn vop))
           (disp (tn-byte-offset stack-temp)))
-      #+little-endian (progn (inst stw y base disp) (inst stfs x base (+ disp 4)))
-      #+big-endian    (progn (inst stfs x base disp) (inst stw y base (+ disp 4)))
-      (inst ld y base disp))))
+      (inst stfs x base disp)
+      (inst lwz temp-reg-tn base disp)
+      (inst sldi temp-reg-tn temp-reg-tn 32)
+      (inst ori y temp-reg-tn single-float-widetag))))
 
 (define-move-vop move-from-single :move
   (single-reg) (descriptor-reg))
@@ -152,18 +152,14 @@
 ;;;; Complex float move functions
 
 (defun complex-single-reg-real-tn (x)
-  (make-random-tn :kind :normal :sc (sc-or-lose 'single-reg)
-                  :offset (tn-offset x)))
+  (make-random-tn (sc-or-lose 'single-reg) (tn-offset x)))
 (defun complex-single-reg-imag-tn (x)
-  (make-random-tn :kind :normal :sc (sc-or-lose 'single-reg)
-                  :offset (1+ (tn-offset x))))
+  (make-random-tn (sc-or-lose 'single-reg) (1+ (tn-offset x))))
 
 (defun complex-double-reg-real-tn (x)
-  (make-random-tn :kind :normal :sc (sc-or-lose 'double-reg)
-                  :offset (tn-offset x)))
+  (make-random-tn (sc-or-lose 'double-reg) (tn-offset x)))
 (defun complex-double-reg-imag-tn (x)
-  (make-random-tn :kind :normal :sc (sc-or-lose 'double-reg)
-                  :offset (1+ (tn-offset x))))
+  (make-random-tn (sc-or-lose 'double-reg) (1+ (tn-offset x))))
 
 
 (define-move-fun (load-complex-single 2) (vop x y)
@@ -192,7 +188,7 @@
     (let ((real-tn (complex-double-reg-real-tn y)))
       (inst lfd real-tn nfp offset))
     (let ((imag-tn (complex-double-reg-imag-tn y)))
-      (inst lfd imag-tn nfp (+ offset (* 2 n-word-bytes))))))
+      (inst lfd imag-tn nfp (+ offset n-word-bytes)))))
 
 (define-move-fun (store-complex-double 4) (vop x y)
   ((complex-double-reg) (complex-double-stack))
@@ -201,7 +197,7 @@
     (let ((real-tn (complex-double-reg-real-tn x)))
       (inst stfd real-tn nfp offset))
     (let ((imag-tn (complex-double-reg-imag-tn x)))
-      (inst stfd imag-tn nfp (+ offset (* 2 n-word-bytes))))))
+      (inst stfd imag-tn nfp (+ offset n-word-bytes)))))
 
 
 ;;;
@@ -369,7 +365,7 @@
          (let ((real-tn (complex-double-reg-real-tn x)))
            (inst stfd real-tn nfp offset))
          (let ((imag-tn (complex-double-reg-imag-tn x)))
-           (inst stfd imag-tn nfp (+ offset (* 2 n-word-bytes)))))))))
+           (inst stfd imag-tn nfp (+ offset n-word-bytes))))))))
 (define-move-vop move-complex-double-float-arg :move-arg
   (complex-double-reg descriptor-reg) (complex-double-reg))
 
@@ -431,8 +427,12 @@
   (frob abs/single-float fabs abs single-reg single-float)
   (frob abs/double-float fabs abs double-reg double-float)
   (frob %negate/single-float fneg %negate single-reg single-float)
-  (frob %negate/double-float fneg %negate double-reg double-float))
+  (frob %negate/double-float fneg %negate double-reg double-float)
+  (frob sqrt/single-float fsqrts %sqrt single-reg single-float)
+  (frob sqrt/double-float fsqrt %sqrt double-reg double-float))
 
+(deftransform %sqrt ((x) * * :vop t)
+  (memq :fsqrt *backend-subfeatures*))
 
 ;;;; Comparison:
 
@@ -498,6 +498,7 @@
   (frob %single-float/signed %single-float single-reg single-float)
   (frob %double-float/signed %double-float double-reg double-float))
 
+#+nil ;; no fcfidu instructions
 (macrolet ((frob (name translate inst to-sc to-type)
             `(define-vop (,name)
                (:args (x :scs (unsigned-reg)))
@@ -712,9 +713,9 @@
   (:vop-var vop)
   (:generator 3
     (let ((nfp (current-nfp-tn vop)))
-      (storew new nfp (1+ (tn-offset temp)))
+      (storew new nfp (tn-offset temp))
       (inst lfd fp-temp nfp (* n-word-bytes (tn-offset temp)))
-      ;; (inst mtfsf 255 fp-temp) ; FIXME: gets sigfpe
+      (inst mtfsf 255 fp-temp)
       (move res new))))
 
 
@@ -746,7 +747,7 @@
              (offset (tn-byte-offset r)))
          (unless (location= real r)
            (inst stfs real nfp offset))
-         (inst stfs imag nfp (+ offset 4)))))))
+         (inst stfs imag nfp (+ offset n-word-bytes)))))))
 
 (define-vop (make-complex-double-float)
   (:translate complex)
@@ -800,6 +801,6 @@
                      (inst ,inst r (current-nfp-tn vop)
                            (+ (tn-byte-offset x) ,byte-offset))))))))
   (def real single 0 lfs)
-  (def imag single 4 lfs)
+  (def imag single 8 lfs)
   (def real double 0 lfd)
   (def imag double 8 lfd))
