@@ -51,6 +51,8 @@
                          :linkage-cell))))
 
 (defun compute-linkage-cell (node name res)
+  #-immobile-space (inst lea res (ea (linkage-cell-fixup name node) null-tn))
+  #+immobile-space ; this is ironically worse than #-immobile-space
   (cond ((code-immobile-p node)
          (inst lea res (rip-relative-ea (linkage-cell-fixup name node))))
         (t
@@ -352,13 +354,15 @@
                 (when 2nd-tn-live
                   (inst mov 2nd-tn null-tn))
                 (when (> nvals 2)
+                  ;; FIXME: simplify this logic- don't use 2nd-tn as a proxy
+                  ;; for NIL now that NULL-TN is a thing.
                   (loop
                     for tn-ref = (tn-ref-across 2nd-tn-ref)
                     then (tn-ref-across tn-ref)
                     for count from 2 below register-arg-count
                     unless (eq (tn-kind (tn-ref-tn tn-ref)) :unused)
                     do
-                    (inst mov :dword (tn-ref-tn tn-ref)
+                    (inst mov (tn-ref-tn tn-ref)
                           (if 2nd-tn-live 2nd-tn null-tn)))))
               (inst mov rbx rsp-tn)
               regs-defaulted))
@@ -397,9 +401,8 @@
                        (inst jmp :nc default-stack-slots))
                       (t
                        (inst jmp :c regs-defaulted)
-                       (loop for null = nil-value then (car used-registers)
-                             for reg in used-registers
-                             do (inst mov :dword reg null))
+                       (loop for reg in used-registers
+                             do (inst mov reg null-tn))
                        (inst jmp done)))
                 REGS-DEFAULTED
                 (do ((i register-arg-count (1+ i))
@@ -432,9 +435,8 @@
                       (when (or (not trust)
                                 (<= min-values 1))
                         (emit-label default-stack-slots)
-                        (loop for null = nil-value then (car used-registers)
-                              for reg in used-registers
-                              do (inst mov :dword reg null))
+                        (loop for reg in used-registers
+                              do (inst mov reg null-tn))
                         (move rbx rsp-tn))
                       (dolist (default defaults)
                         (emit-label (car default))
@@ -484,7 +486,7 @@
     ;; stack values. In this case quickly reallocate sufficient space.
     (when (<= (sb-kernel:values-type-min-value-count type)
               register-arg-count)
-      (inst cmp nargs (fixnumize register-arg-count))
+      (inst cmp :dword nargs (fixnumize register-arg-count))
       (inst jmp :g stack-values)
       #+#.(cl:if (cl:= sb-vm:word-shift sb-vm:n-fixnum-tag-bits) '(and) '(or))
       (inst sub rsp-tn nargs)
@@ -903,6 +905,9 @@
          (inst* instruction (ea rax-tn)))
         ((code-immobile-p node)
          (inst* instruction (rip-relative-ea (linkage-cell-fixup name node))))
+        #-immobile-space
+        (t (inst* instruction (ea (linkage-cell-fixup name node) null-tn)))
+        #+immobile-space ; again, this should not be worse than #-immobile-space, but it is
         (t
          ;; get the linkage table base into RAX
          (inst mov rax-tn (thread-slot-ea sb-vm::thread-linkage-table-slot))
@@ -911,8 +916,7 @@
 ;;; Invoke the function-designator FUN.
 (defun tail-call-unnamed (fun type vop)
   (let ((relative-call (code-immobile-p vop))
-        (fun-ea (ea (- (* closure-fun-slot n-word-bytes) fun-pointer-lowtag)
-                    fun)))
+        (fun-ea (object-slot-ea fun closure-fun-slot fun-pointer-lowtag)))
     (case type
       (:designator
        (assemble ()
@@ -943,8 +947,7 @@
          (invoke-asm-routine 'call 'call-symbol vop)
          (inst jmp ret))
        call
-       (inst call (ea (- (* closure-fun-slot n-word-bytes) fun-pointer-lowtag)
-                      fun))
+       (inst call (object-slot-ea fun closure-fun-slot fun-pointer-lowtag))
        ret))))
 
 ;;; This is defined separately, since it needs special code that BLT's
@@ -1042,8 +1045,8 @@
     ;; Set the multiple value return flag.
     (inst stc)
     ;; And away we go. Except that return-pc is still on the
-    ;; stack and we've changed the stack pointer. So we have to
-    ;; tell it to index off of RBX instead of RBP.
+    ;; stack and we've changed the stack pointer. So we might have to
+    ;; tell it to index off of RBX instead of RBP, depending on nvals.
     (cond ((<= nvals register-arg-count)
            (inst leave)
            (inst ret))
@@ -1057,10 +1060,8 @@
            (inst lea rsp-tn
                  (ea (frame-byte-offset (1- nvals)) rbp-tn))
            (move rbp-tn old-fp)
-           (inst push (ea (frame-byte-offset
-                           (+ sp->fp-offset (tn-offset return-pc)))
-                          rbx))
-           (inst ret)))))
+           (emit-mv-return
+            (ea (frame-byte-offset (+ sp->fp-offset (tn-offset return-pc))) rbx))))))
 
 ;;; Do unknown-values return of an arbitrary number of values (passed
 ;;; on the stack.) We check for the common case of a single return
@@ -1087,7 +1088,7 @@
     (unless (policy node (> space speed))
       ;; Check for the single case.
       (let ((not-single (gen-label)))
-        (inst cmp nvals (fixnumize 1))
+        (inst cmp :dword nvals (fixnumize 1))
         (inst jmp :ne not-single)
         ;; Return with one value.
         (loadw a0 vals -1)
