@@ -13,17 +13,23 @@
 (in-package "SB-VM")
 
 (defun symbol-slot-ea (symbol slot)
-  (ea (let ((offset (- (* slot n-word-bytes) other-pointer-lowtag)))
-             (if (static-symbol-p symbol)
-                 (+ nil-value (static-symbol-offset symbol) offset)
-                 (make-fixup symbol :immobile-symbol offset)))))
+  (let ((offset (- (* slot n-word-bytes) other-pointer-lowtag)))
+    (if (static-symbol-p symbol)
+        (ea (+ (static-symbol-offset symbol) offset) null-tn)
+        (ea (make-fixup symbol :immobile-symbol offset)))))
 
+;;; The GC card table base is either an imm8 displacement from NULL-TN, or that
+;;; plus one backend page if #+sb-safepoint. The reason for sliding the table up
+;;; with #+sb-safepoint is that the safepoint trap address wants to be an imm8 away
+;;; from NIL, as there are far more safepoint instructions than GC barriers.
 (defun mark-gc-card (addr-or-card-index &optional (compute t))
   (when compute
     (inst shr addr-or-card-index gencgc-card-shift)
     ;; :DWORD suffices because gc_allocate_ptes() asserts mask to be < 32 bits
     (inst and :dword addr-or-card-index card-index-mask))
-  (inst mov :byte (ea gc-card-table-reg-tn addr-or-card-index) CARD-MARKED))
+  (let ((disp (+ sb-vm::nil-static-space-end-offs
+                 #+sb-safepoint sb-c:+backend-page-bytes+))) ; stupidly excessive
+    (inst mov :byte (ea disp null-tn addr-or-card-index) CARD-MARKED)))
 
 ;;; TODOs:
 ;;; 1. Sometimes people write constructors like
@@ -92,6 +98,7 @@
                          bits)
                     (plausible-signed-imm32-operand-p bits))
                 (inst mov :qword ea it))
+               ((tn-p bits) (inst mov ea bits)) ; null-tn
                (t
                 (move-immediate val-temp bits)
                 (inst mov ea val-temp)))))
@@ -211,8 +218,8 @@
   (if (typep value 'sb-vm:word)
       value
       (case value
-         (:unbound (unbound-marker-bits))
-         ((nil) (bug "Should not see SPLAT NIL"))
+         (:unbound unbound-marker-widetag)
+         ((nil) null-tn)
          (t #+ubsan unwritten-vector-element-marker
             #-ubsan 0))))
 
@@ -232,12 +239,8 @@
   (:generator 1
    (progn words) ; don't put it in :ignore, which gets inherited
    (let ((bits (compute-splat-bits value)))
-     (when (integerp bits)
-       ;; it's gonna get treated as simm32 by the CPU, so it had better be
-       (aver (plausible-signed-imm32-operand-p bits)))
-     (inst mov :qword
-           (ea (- (* vector-data-offset n-word-bytes) other-pointer-lowtag) vector)
-           bits))
+     (aver (or (tn-p bits) (plausible-signed-imm32-operand-p bits)))
+     (inst mov :qword (object-slot-ea vector vector-data-offset other-pointer-lowtag) bits))
    (move result vector)))
 
 (define-vop (splat-small splat-word)
@@ -245,19 +248,22 @@
   (:temporary (:sc complex-double-reg) pattern)
   (:generator 5
    (let ((bits (compute-splat-bits value)))
-     (cond ((= bits nil-value)
-            (inst movaps pattern (ea (- nil-value list-pointer-lowtag))))
-           ((/= bits 0) (inst movddup pattern (register-inline-constant :qword bits)))
-           (t (inst xorps pattern pattern))))
-   (let ((data-addr (ea (- (* vector-data-offset n-word-bytes) other-pointer-lowtag)
-                        vector)))
-     (multiple-value-bind (quo rem) (truncate words 2)
-       (dotimes (i quo)
-         (inst movaps data-addr pattern) ; 1 byte shorter encoding than movapd
-         (setf data-addr (ea (+ (ea-disp data-addr) (* n-word-bytes 2))
-                             (ea-base data-addr))))
-       (unless (zerop rem)
-         (inst movsd data-addr pattern))))
+     (cond ((and (eq bits null-tn) (eql words 2)) ; don't use XMM register
+            (storew null-tn vector vector-data-offset other-pointer-lowtag)
+            (storew null-tn vector (1+ vector-data-offset) other-pointer-lowtag))
+           (t
+            (cond ((eq bits null-tn)
+                   (inst movaps pattern (ea (- list-pointer-lowtag) null-tn)))
+                  ((/= bits 0) (inst movddup pattern (register-inline-constant :qword bits)))
+                  (t (inst xorps pattern pattern)))
+            (let ((data-addr (object-slot-ea vector vector-data-offset other-pointer-lowtag)))
+              (multiple-value-bind (quo rem) (truncate words 2)
+                (dotimes (i quo)
+                  (inst movaps data-addr pattern) ; 1 byte shorter encoding than movapd
+                  (setf data-addr (ea (+ (ea-disp data-addr) (* n-word-bytes 2))
+                                      (ea-base data-addr))))
+                (unless (zerop rem)
+                  (inst movsd data-addr pattern)))))))
    (move result vector)))
 
 (define-vop (splat-any splat-word)
@@ -275,10 +281,9 @@
                :to (:result 0)) rax)
   (:results (result :scs (descriptor-reg)))
   (:generator 10
-   (inst lea rdi (ea (- (* vector-data-offset n-word-bytes) other-pointer-lowtag)
-                        vector))
+   (inst lea rdi (object-slot-ea vector vector-data-offset other-pointer-lowtag))
    (let ((bits (compute-splat-bits value)))
-     (cond ((and (= bits 0)
+     (cond ((and (eql bits 0)
                  (constant-tn-p words)
                  (typep (tn-value words) '(unsigned-byte 7)))
             (zeroize rax)
@@ -286,7 +291,7 @@
            (t
             ;; words could be in RAX, so read it first, then zeroize
             (inst mov rcx (or (and (constant-tn-p words) (tn-value words)) words))
-            (if (= bits 0) (zeroize rax) (inst mov rax bits)))))
+            (if (eql bits 0) (zeroize rax) (inst mov rax bits)))))
    (inst rep)
    (inst stos :qword)
    (move result vector)))
