@@ -151,8 +151,10 @@
                      (unless size
                        (setf size :dword))
                      (setf size :qword))
+                 (if (nil-relative-p bits)
+                     (move-immediate dst bits)
                  ;; Can't use ZEROIZE, since XOR will affect the flags.
-                 (inst mov dst bits))))
+                     (inst mov dst bits)))))
         (cond ((null (rest flags))
                (cond ((sc-is else immediate)
                       (load-immediate res else))
@@ -219,7 +221,7 @@
 ;;; Return NIL to give up.
 (defun computable-from-flags-p (res x y flags)
   ;; TODO: handle unsigned-reg
-  (unless (and (singleton-p flags)
+  (unless (and (singleton-p (conditional-flags-flags flags))
                (sc-is res sb-vm::any-reg sb-vm::descriptor-reg))
     (return-from computable-from-flags-p nil))
   ;; There are plenty more algebraic transforms possible,
@@ -240,8 +242,15 @@
                 (typep (fixnumize y) '(signed-byte 32))
                 (member (abs (fixnumize (- x y))) '(2 4 8))
                 'add)))
-    (or #+sb-thread (or (and (eq x t) (eq y nil) 'boolean)
-                        (and (eq x nil) (eq y t) 'boolean))
+    ;; FIXME: the BOOLEAN case has little benefit, except that converting to
+    ;; a CMOV in the general way unnecessarily loads both inputs even when
+    ;; one of them is NIL, e.g. (lambda (x) (eql x 1)) becomes
+    ;;   CMP RSI, 2
+    ;;   MOV RDX, R12 ; <-- this is completely superfluous
+    ;;   LEA RAX, [R12-56]
+    ;;   CMOVEQ RDX, RAX
+    (or (or (and (eq x t) (eq y nil) 'boolean)
+            (and (eq x nil) (eq y t) 'boolean))
         (try-shift x y)
         (try-shift y x)
         (try-add x y))))
@@ -256,21 +265,12 @@
            (y (tn-value y-tn))
            #+gs-seg (thread-tn nil)
            (hint (computable-from-flags-p res x y flags))
-           (flag (car flags)))
+           (flag (car (conditional-flags-flags flags))))
       (ecase hint
         (boolean
-         ;; FIXNUMP -> {T,NIL} could be special-cased, reducing the instruction count by
-         ;; 1 or 2 depending on whether the argument and result are in the same register.
-         ;; Best case would be "AND :dword res, arg, 1 ; MOV res, [ea]".
-         (when (eql x t)
-           ;; T is at the lower address, so to pick it out we need index=0
-           ;; which makes the condition in (IF BIT T NIL) often flipped.
-           (setq flag (negate-condition flag)))
-         (inst set flag res)
-         (inst movzx '(:byte :dword) res res)
-         (inst mov :dword res
-               (ea thread-segment-reg (ash thread-t-nil-constants-slot word-shift)
-                   thread-tn res 4)))
+         (when (eql x t) (setq flag (negate-condition flag)))
+         (load-symbol res t) ; doesn't mess up flags
+         (inst cmov flag res null-tn))
         (shl
          (when (eql x 0)
            (setq flag (negate-condition flag)))
@@ -303,12 +303,14 @@
   (:policy :fast-safe)
   (:translate eq)
   (:arg-refs x-tn-ref)
-  (:temporary (:sc unsigned-reg) temp)
+  (:temporary (:sc unsigned-reg) temp) ; TODO: add :unused-if
   (:generator 6
     (cond
       ((sc-is y constant)
        (inst cmp x (cond ((sc-is x descriptor-reg any-reg) y)
                          (t (inst mov temp y) temp))))
+      ((and (sc-is y immediate) (nil-relative-p (immediate-tn-repr y)))
+       (inst cmp x (move-immediate temp (immediate-tn-repr y))))
       ((sc-is y immediate)
        (let* ((value (encode-value-if-immediate y))
               (immediate (plausible-signed-imm32-operand-p value)))
@@ -327,7 +329,7 @@
                        (specifier-type 'cons)))
              (inst cmp :byte x (logand nil-value #xff))
              (return-from if-eq)))
-         (cond ((fixup-p value) ; immobile object
+         (cond ((or (fixup-p value) (tn-p value)) ; immobile object or NIL
                 (inst cmp x value))
                ((and (zerop value) (sc-is x any-reg descriptor-reg))
                 (inst test x x))
@@ -371,7 +373,8 @@
             :load-if (or (not (sc-is x immediate))
                          (typep (tn-value x)
                                 '(and integer
-                                  (not (signed-byte #.(- 32 n-fixnum-tag-bits))))))))
+                                  (not (signed-byte #.(- 32 n-fixnum-tag-bits)))))
+                         (nil-relative-p (immediate-tn-repr x)))))
   (:arg-types * (:constant (unsigned-byte 16)) *)
   (:info slot)
   (:translate %instance-ref-eq)
