@@ -417,19 +417,9 @@
 ;;; Sign bit and fixnum tag bit.
 (defconstant non-negative-fixnum-mask-constant
   #x8000000000000001)
-(defconstant non-negative-fixnum-mask-constant-wired-address
-  (+ static-space-start (* 12 n-word-bytes)))
-(defmacro non-negative-fixnum-mask-ea ()
-  `(ea non-negative-fixnum-mask-constant-wired-address))
-;; the preceding constant is embedded in an array,
-;; the header of which must not overlap the static alloc regions
-#-sb-thread
-(aver (>= (- non-negative-fixnum-mask-constant-wired-address (* 2 n-word-bytes))
-          (+ static-space-start
-             (max boxed-region-offset
-                  cons-region-offset
-                  mixed-region-offset)
-             (* 3 n-word-bytes))))
+;; this magic constant resides in an otherwise unused word after the symbol slots of NIL
+(defun non-negative-fixnum-mask-ea ()
+  (ea (- (ash 5 word-shift) list-pointer-lowtag) null-tn))
 
 ;;; An (unsigned-byte 64) can be represented with either a positive
 ;;; fixnum, a bignum with exactly one positive digit, or a bignum with
@@ -726,7 +716,8 @@
 ;;; but hey at least this provides the IR2 support for it.
 (define-vop (car-eq-if-listp)
   (:args (value :scs (descriptor-reg))
-         (obj :scs (immediate any-reg descriptor-reg)))
+         (obj :scs (any-reg descriptor-reg
+                            (immediate (reg-or-legal-imm32-p tn)))))
   (:temporary (:sc unsigned-reg) temp)
   (:conditional :z)
   (:policy :fast-safe)
@@ -1043,11 +1034,15 @@
   (:result-types unsigned-num)
   (:generator 1
     (when null-label
+      ;; Since the comparison against a register, not an imm8 or imm32 any more,
+      ;; there is probably no reason to distinguish the two cases here.
+      ;; The only conceivable reason would be if null-tn were in a low register
+      ;; then it's theoretically possible that a REX prefix could be avoided.
       (if (types-equal-or-intersect
            (type-difference (tn-ref-type value-ref) (specifier-type 'null))
            (specifier-type 'cons))
           (inst cmp value null-tn)
-          (inst cmp :byte value (logand nil-value #xff)))
+          (inst cmp :byte value null-tn)) ; was: (logand nil-value #xff)
       (inst jmp :e null-label))
     (cond ((other-pointer-tn-ref-p value-ref t)
            (if zero-extend
