@@ -311,9 +311,7 @@
                  (let ((tn-name (symbolicate reg-name "-TN"))
                        (offset-name (symbolicate reg-name "-OFFSET")))
                    (forms `(defglobal ,tn-name
-                             (make-random-tn :kind :normal
-                                             :sc (sc-or-lose ',sc-name)
-                                             :offset ,offset-name))))))))
+                             (make-random-tn (sc-or-lose ',sc-name) ,offset-name))))))))
 
   (def-misc-reg-tns unsigned-reg eax ebx ecx edx ebp esp edi esi)
   (def-misc-reg-tns word-reg ax bx cx dx bp sp di si)
@@ -330,9 +328,7 @@
 #|
 ;;; added by pw
 (defparameter fp-constant-tn
-  (make-random-tn :kind :normal
-                  :sc (sc-or-lose 'fp-constant)
-                  :offset 31))          ; Offset doesn't get used.
+  (make-random-tn (sc-or-lose 'fp-constant) 31))          ; Offset doesn't get used.
 |#
 
 ;;; If value can be represented as an immediate constant, then return
@@ -347,11 +343,11 @@
        immediate-sc-number))
     (single-float
        (case value
-         (($0f0 $1f0) fp-constant-sc-number)
+         ((0f0 1f0) fp-constant-sc-number)
          (t fp-single-immediate-sc-number)))
     (double-float
        (case value
-         (($0d0 $1d0) fp-constant-sc-number)
+         ((0d0 1d0) fp-constant-sc-number)
          (t fp-double-immediate-sc-number)))
     #+long-float
     (long-float
@@ -383,7 +379,7 @@
                              character-widetag))
           (structure-object
            (if (eq val sb-lockless:+tail+)
-               sb-vm::lockfree-list-tail-value
+               (+ static-space-start lockfree-list-tail-value-offset)
                (bug "immediate structure-object ~S" val)))))
       tn))
 
@@ -421,7 +417,7 @@
                                         ; related to signal context stuff
 
 ;;; This is used by the debugger.
-(defconstant single-value-return-byte-offset 2)
+(defconstant single-value-return-byte-offset 0)
 
 ;;; This function is called by debug output routines that want a pretty name
 ;;; for a TN's location. It returns a thing that can be printed with PRINC.
@@ -446,34 +442,3 @@
       (constant (format nil "Const~D" offset))
       (immediate-constant "Immed")
       (noise (symbol-name (sc-name sc))))))
-
-(defun combination-implementation-style (node)
-  (declare (type sb-c::combination node))
-  (flet ((valid-funtype (args result)
-           (sb-c::valid-fun-use node
-                                (sb-c::specifier-type
-                                 `(function ,args ,result)))))
-    (case (sb-c::combination-fun-source-name node)
-      (logtest
-       (cond
-         ((valid-funtype '(fixnum fixnum) '*)
-          (values :maybe nil))
-         ((valid-funtype '((signed-byte 32) (signed-byte 32)) '*)
-          (values :maybe nil))
-         ((valid-funtype '((unsigned-byte 32) (unsigned-byte 32)) '*)
-          (values :maybe nil))
-         (t (values :default nil))))
-      (logbitp
-       (cond
-         ((and (valid-funtype '((integer 0 29) fixnum) '*)
-               (sb-c:constant-lvar-p (first (sb-c::basic-combination-args node))))
-          (values :transform '(lambda (index integer)
-                               (%logbitp integer index))))
-         ((valid-funtype '((integer 0 31) (signed-byte 32)) '*)
-          (values :transform '(lambda (index integer)
-                               (%logbitp integer index))))
-         ((valid-funtype '((integer 0 31) (unsigned-byte 32)) '*)
-          (values :transform '(lambda (index integer)
-                               (%logbitp integer index))))
-         (t (values :default nil))))
-      (t (values :default nil)))))
