@@ -55,7 +55,8 @@
 (defun make-effective-method-function1 (generic-function form
                                         method-alist-p wrappers-p)
   (if (and (listp form)
-           (eq (car form) 'call-method))
+           (eq (car form) 'call-method)
+           (not (gf-requires-emf-keyword-checks generic-function)))
       (make-effective-method-function-simple generic-function form)
       ;; We have some sort of `real' effective method. Go off and get a
       ;; compiled function for it. Most of the real hair here is done by
@@ -118,8 +119,56 @@
                                               method-alist
                                               wrappers))))
 
+;;; methods-tracing TODO:
+;;;
+;;; 2. tracing method calls for non-fast-method-function calls
+;;;    - [DONE] the calls themselves
+;;;    - calls to the METHOD-FUNCTION of methods with fast functions
+;;;      (e.g. from something implementing CALL-NEXT-METHOD; handle this with
+;;;      some more smarts in %METHOD-FUNCTION objects?)
+;;;    - calls to the METHOD-FUNCTION of methods without fast functions
+;;;      (TRACE :METHODS T /could/ modify the METHOD-FUNCTION slot)
+;;; 4. tracing particular methods
+;;;    - need an interface.
+;;;      * (trace (method foo :around (t)))? [ how to trace the method and not
+;;;        the generic function as a whole?]
+;;;      * (trace :methods '((:around (t))) foo)? [probably not, interacts
+;;;        poorly with TRACE arg handling]
+;;; 5. supporting non-munged arguments as an option
+
+(defun method-trace-name (gf method)
+  ;; KLUDGE: we abuse NIL as second argument to mean that this is a
+  ;; combined method (i.e. something resulting from MAKE-METHOD in a
+  ;; method combination, rather than CALL-METHOD on a method object).
+  (if method
+      `(method ,(generic-function-name gf)
+               ,@(method-qualifiers method)
+               ,(unparse-specializers gf (method-specializers method)))
+      `(combined-method ,(generic-function-name gf))))
+
+(defun maybe-trace-method (gf method fun fmf-p)
+  (let ((m-name (when (plusp (hash-table-count sb-debug::*traced-funs*))
+                  ;; KLUDGE: testing if *TRACE-FUNS* has anything anything to
+                  ;; avoid calling METHOD-TRACE-NAME during PCL bootstrapping
+                  ;; when the generic-function type is not yet defined.)
+                  (method-trace-name gf method))))
+    (when m-name
+      (sb-debug::retrace-local-funs m-name))
+    (let ((info (when m-name
+                  (or (gethash m-name sb-debug::*traced-funs*)
+                      (let ((gf-info (gethash (or (generic-function-name gf) gf)
+                                              sb-debug::*traced-funs*)))
+                        (when (and gf-info (sb-debug::trace-info-methods gf-info))
+                          (let ((copy (copy-structure gf-info)))
+                            (setf (sb-debug::trace-info-what copy) m-name)
+                            copy)))))))
+      (if info
+          (lambda (&rest args)
+            (apply #'sb-debug::trace-method-call info fun fmf-p args))
+          fun))))
+
 (defun make-emf-from-method
-    (method cm-args &optional gf fmf-p method-alist wrappers)
+    (gf method cm-args fmf-p &optional method-alist wrappers)
   ;; Avoid style-warning about compiler-macro being unavailable.
   (declare (notinline make-instance))
   (multiple-value-bind (mf real-mf-p fmf pv)
@@ -132,12 +181,13 @@
                       fmf-p method-alist wrappers))
                (arg-info (method-plist-value method :arg-info))
                (default (cons nil nil))
-               (value (method-plist-value method :constant-value default)))
+               (value (method-plist-value method :constant-value default))
+               (fun (maybe-trace-method gf method fmf t)))
           (if (eq value default)
-              (make-fast-method-call :function fmf :pv pv
-                                     :next-method-call next :arg-info arg-info)
+              (make-fast-method-call
+               :function fun :pv pv :next-method-call next :arg-info arg-info)
               (make-constant-fast-method-call
-               :function fmf :pv pv :next-method-call next
+               :function fun :pv pv :next-method-call next
                :arg-info arg-info :value value)))
         (if real-mf-p
             (flet ((frob-cm-arg (arg)
@@ -158,7 +208,7 @@
                                   (let* ((fmf (fast-method-call-function emf))
                                          (fun (method-function-from-fast-method-call emf))
                                          (mf (%make-method-function fmf)))
-                                    (set-funcallable-instance-function mf fun)
+                                    (setf (%funcallable-instance-fun mf) fun)
                                     (make-instance 'standard-method
                                                    :specializers nil ; XXX
                                                    :qualifiers nil
@@ -181,11 +231,12 @@
                      ;; user-defined class don't work at all.  -- CSR,
                      ;; 2006-08-05
                      (args (cons (mapcar #'frob-cm-arg (car cm-args))
-                                 (cdr cm-args))))
+                                 (cdr cm-args)))
+                     (fun (maybe-trace-method gf method mf nil)))
                 (if (eq value default)
-                    (make-method-call :function mf :call-method-args args)
-                    (make-constant-method-call :function mf :value value
-                                               :call-method-args args))))
+                    (make-method-call :function fun :call-method-args args)
+                    (make-constant-method-call
+                     :function fun :value value :call-method-args args))))
             mf))))
 
 (defun make-effective-method-function-simple1
@@ -194,7 +245,7 @@
     (if (if (listp method)
             (eq (car method) :early-method)
             (method-p method))
-        (make-emf-from-method method cm-args gf fmf-p method-alist wrappers)
+        (make-emf-from-method gf method cm-args fmf-p method-alist wrappers)
         (if (and (consp method) (eq (car method) 'make-method))
             (make-effective-method-function gf
                                             (cadr method)
@@ -206,9 +257,8 @@
 
 (defun get-effective-method-gensym ()
   (or (pop *rebound-effective-method-gensyms*)
-      (let ((new (format-symbol *pcl-package*
-                                "EFFECTIVE-METHOD-GENSYM-~D"
-                                (length *global-effective-method-gensyms*))))
+      (let ((new (pcl-symbolicate "EFFECTIVE-METHOD-GENSYM-"
+                                    (length *global-effective-method-gensyms*))))
         (setq *global-effective-method-gensyms*
               (append *global-effective-method-gensyms* (list new)))
         new)))
@@ -222,48 +272,33 @@
   (multiple-value-bind (nreq applyp)
       (get-generic-fun-info gf)
     (let ((ll (make-fast-method-call-lambda-list nreq applyp))
-          (check-applicable-keywords
-           (when (and applyp (gf-requires-emf-keyword-checks gf))
-             '((check-applicable-keywords))))
-          (error-p (or (eq (first effective-method) '%no-primary-method)
-                       (eq (first effective-method) '%invalid-qualifiers)))
           (mc-args-p
            (when (eq **boot-state** 'complete)
              ;; Otherwise the METHOD-COMBINATION slot is not bound.
              (let ((combin (generic-function-method-combination gf)))
                (and (long-method-combination-p combin)
                     (long-method-combination-args-lambda-list combin)))))
+          ;; FIXME: this name in the lambda almost completely defeats
+          ;; the fngen cache when compiling method combinations
+          ;; since no two expressions will be EQUAL. Perhaps we should
+          ;; teach fngen to remove the name?
           (name `(emf ,(generic-function-name gf))))
       (cond
-        (error-p
-         `(named-lambda ,name (.pv. .next-method-call. &rest .args.)
-            (declare (ignore .pv. .next-method-call.))
-            (declare (ignorable .args.))
-            (flet ((%no-primary-method (gf args)
-                     (call-no-primary-method gf args))
-                   (%invalid-qualifiers (gf combin method)
-                     (invalid-qualifiers gf combin method)))
-              (declare (ignorable #'%no-primary-method #'%invalid-qualifiers))
-              ,effective-method)))
         (mc-args-p
          (let* ((required (make-dfun-required-args nreq))
                 (gf-args (if applyp
-                             `(list* ,@required
-                                     (sb-c::%listify-rest-args
-                                      .dfun-more-context.
-                                      (the (and unsigned-byte fixnum)
-                                        .dfun-more-count.)))
+                             `(list* ,@required (sb-c::%rest-list .rest.))
                              `(list ,@required))))
            `(named-lambda ,name ,ll
               (declare (ignore .pv. .next-method-call.))
               (let ((.gf-args. ,gf-args))
                 (declare (ignorable .gf-args.))
-                ,@check-applicable-keywords
                 ,effective-method))))
         (t
          `(named-lambda ,name ,ll
-            (declare (ignore ,@(if error-p ll '(.pv. .next-method-call.))))
-            ,@check-applicable-keywords
+            (declare (ignore .pv. .next-method-call.))
+            (declare (ignorable ,@(make-dfun-required-args nreq)
+                                ,@(when applyp '(.rest.))))
             ,effective-method))))))
 
 (defun expand-emf-call-method (gf form metatypes applyp env)
@@ -298,7 +333,6 @@
             generic-function form method-alist-p wrappers-p)
        (fast-method-call '.fast-call-method-list.)
        (t '.call-method-list.)))
-    (check-applicable-keywords 'check-applicable-keywords)
     (t (default-test-converter form))))
 
 ;;; CMUCL comment (2003-10-15):
@@ -324,12 +358,6 @@
        (values `(dolist (emf ,gensym nil)
                  ,(make-emf-call (length metatypes) applyp 'emf type))
                (list gensym))))
-    (check-applicable-keywords
-     (values `(check-applicable-keywords .keyargs-start.
-                                         .valid-keys.
-                                         .dfun-more-context.
-                                         .dfun-more-count.)
-             '(.keyargs-start. .valid-keys.)))
     (t
      (default-code-converter form))))
 
@@ -345,12 +373,9 @@
                            (make-effective-method-function-simple
                             generic-function form))
                          (cdr form)))))
-    (check-applicable-keywords
-     '(.keyargs-start. .valid-keys.))
     (t
      (default-constant-converter form))))
 
-(defvar *applicable-methods*)
 (defun make-effective-method-function-internal
     (generic-function effective-method method-alist-p wrappers-p)
   (multiple-value-bind (nreq applyp metatypes nkeys arg-info)
@@ -365,40 +390,34 @@
            (effective-method-lambda (expand-effective-method-function
                                      generic-function effective-method)))
       (multiple-value-bind (cfunction constants)
-          (get-fun1 effective-method-lambda
-                    (lambda (form)
-                      (memf-test-converter form generic-function
-                                           method-alist-p wrappers-p))
-                    (lambda (form)
-                      (memf-code-converter form generic-function
-                                           metatypes applyp
-                                           method-alist-p wrappers-p))
-                    (lambda (form)
-                      (memf-constant-converter form generic-function)))
+          (get-fun effective-method-lambda
+                   (lambda (form)
+                     (memf-test-converter form generic-function
+                                          method-alist-p wrappers-p))
+                   (lambda (form)
+                     (memf-code-converter form generic-function
+                                          metatypes applyp
+                                          method-alist-p wrappers-p))
+                   (lambda (form)
+                     (memf-constant-converter form generic-function)))
         (lambda (method-alist wrappers)
-          (multiple-value-bind (valid-keys keyargs-start)
-              (when (memq '.valid-keys. constants)
-                (compute-applicable-keywords
-                 generic-function *applicable-methods*))
-            (flet ((compute-constant (constant)
-                     (if (consp constant)
-                         (case (car constant)
-                           (.meth.
-                            (funcall (cdr constant) method-alist wrappers))
-                           (.meth-list.
-                            (mapcar (lambda (fn)
-                                      (funcall fn method-alist wrappers))
-                                    (cdr constant)))
-                           (t constant))
-                         (case constant
-                           (.keyargs-start. keyargs-start)
-                           (.valid-keys. valid-keys)
-                           (t constant)))))
-              (let ((fun (apply cfunction
-                                (mapcar #'compute-constant constants))))
-                (set-fun-name fun `(combined-method ,name))
-                (make-fast-method-call :function fun
-                                       :arg-info arg-info)))))))))
+          (flet ((compute-constant (constant)
+                   (if (consp constant)
+                       (case (car constant)
+                         (.meth.
+                          (funcall (cdr constant) method-alist wrappers))
+                         (.meth-list.
+                          (mapcar (lambda (fn)
+                                    (funcall fn method-alist wrappers))
+                                  (cdr constant)))
+                         (t constant))
+                       (case constant
+                         (t constant)))))
+            (let ((fun (apply cfunction
+                              (mapcar #'compute-constant constants))))
+              (set-fun-name fun `(combined-method ,name))
+              (make-fast-method-call :function (maybe-trace-method generic-function nil fun t)
+                                     :arg-info arg-info))))))))
 
 (defmacro call-method-list (&rest calls)
   `(progn ,@calls))
@@ -419,20 +438,35 @@
 (defun standard-compute-effective-method
     (generic-function combin applicable-methods)
   (collect ((before) (primary) (after) (around))
-    (flet ((invalid (gf combin m) (invalid-qualifiers gf combin m)))
+    (flet ((invalid (method)
+             (return-from standard-compute-effective-method
+               `(invalid-qualifiers ',generic-function ',combin ',method))))
       (dolist (m applicable-methods)
         (let ((qualifiers (if (listp m)
                               (early-method-qualifiers m)
                               (safe-method-qualifiers m))))
           (cond
             ((null qualifiers) (primary m))
-            ((cdr qualifiers) (invalid generic-function combin m))
+            ((cdr qualifiers) (invalid m))
             ((eq (car qualifiers) :around) (around m))
             ((eq (car qualifiers) :before) (before m))
             ((eq (car qualifiers) :after) (after m))
-            (t (invalid generic-function combin m))))))
-    (cond ((null (primary))
-           `(%no-primary-method ',generic-function .args.))
+            (t (invalid m))))))
+    (cond ((null applicable-methods)
+           ;; APPLICABLE-METHODS is normally non-null in effective
+           ;; method computation, but COMPUTE-APPLICABLE-METHODS can
+           ;; in principle be called by MetaObject Protocol programmers.
+           `(method-combination-error
+             "No applicable method found for ~S"
+             ',generic-function))
+          ((null (primary))
+           ;; PCL checks for no primary method before method
+           ;; combination, but a MetaObject Protocol programmer could
+           ;; call COMPUTE-EFFECTIVE-METHOD themselves and end up
+           ;; here.
+           `(method-combination-error
+             "No primary method found for ~S among applicable methods: ~S"
+             ',generic-function (list ,@(mapcar (lambda (m) `(quote ,m)) applicable-methods))))
           ((and (null (before)) (null (after)) (null (around)))
            ;; By returning a single call-method `form' here we enable
            ;; an important implementation-specific optimization; that
@@ -444,13 +478,12 @@
            ;; perform this checking in fast-method-functions given
            ;; that they are not solely used for effective method
            ;; functions, but also in combination, when they should not
-           ;; perform argument checks.
-           (let ((call-method
-                  `(call-method ,(first (primary)) ,(rest (primary)))))
-             (if (gf-requires-emf-keyword-checks generic-function)
-                 ;; the PROGN inhibits the above optimization
-                 `(progn ,call-method)
-                 call-method)))
+           ;; perform argument checks.  We still return the bare
+           ;; CALL-METHOD, but the caller is responsible for ensuring
+           ;; that keyword applicability is checked if this is a fast
+           ;; method function used in an effective method.  (See
+           ;; WRAP-WITH-APPLICABLE-KEYWORD-CHECK below).
+           `(call-method ,(first (primary)) ,(rest (primary))))
           (t
            (let ((main-effective-method
                    (if (or (before) (after))
@@ -481,18 +514,18 @@
         (order (car (method-combination-options combin)))
         (around ())
         (primary ()))
-    (flet ((invalid (gf combin m)
+    (flet ((invalid (method)
              (return-from short-compute-effective-method
-               `(%invalid-qualifiers ',gf ',combin ',m))))
+               `(invalid-qualifiers ',generic-function ',combin ',method))))
       (dolist (m applicable-methods)
         (let ((qualifiers (method-qualifiers m)))
-          (cond ((null qualifiers) (invalid generic-function combin m))
-                ((cdr qualifiers) (invalid generic-function combin m))
+          (cond ((null qualifiers) (invalid m))
+                ((cdr qualifiers) (invalid m))
                 ((eq (car qualifiers) :around)
                  (push m around))
                 ((eq (car qualifiers) type-name)
                  (push m primary))
-                (t (invalid generic-function combin m))))))
+                (t (invalid m))))))
     (setq around (nreverse around))
     (ecase order
       (:most-specific-last) ; nothing to be done, already in correct order
@@ -504,34 +537,21 @@
                 `(call-method ,(car primary) ())
                 `(,operator ,@(mapcar (lambda (m) `(call-method ,m ()))
                                       primary)))))
-      (cond ((null primary)
-             ;; As of sbcl-0.8.0.80 we don't seem to need to do
-             ;; anything messy like
-             ;;        `(APPLY (FUNCTION (IF AROUND
-             ;;                              'NO-PRIMARY-METHOD
-             ;;                              'NO-APPLICABLE-METHOD)
-             ;;                           ',GENERIC-FUNCTION
-             ;;                           .ARGS.)
-             ;; here because (for reasons I don't understand at the
-             ;; moment -- WHN) control will never reach here if there
-             ;; are no applicable methods, but instead end up
-             ;; in NO-APPLICABLE-METHODS first.
-             ;;
-             ;; FIXME: The way that we arrange for .ARGS. to be bound
-             ;; here seems weird. We rely on EXPAND-EFFECTIVE-METHOD-FUNCTION
-             ;; recognizing any form whose operator is %NO-PRIMARY-METHOD
-             ;; as magical, and carefully surrounding it with a
-             ;; LAMBDA form which binds .ARGS. But...
-             ;;   1. That seems fragile, because the magicalness of
-             ;;      %NO-PRIMARY-METHOD forms is scattered around
-             ;;      the system. So it could easily be broken by
-             ;;      locally-plausible maintenance changes like,
-             ;;      e.g., using the APPLY expression above.
-             ;;   2. That seems buggy w.r.t. to MOPpish tricks in
-             ;;      user code, e.g.
-             ;;         (DEFMETHOD COMPUTE-EFFECTIVE-METHOD :AROUND (...)
-             ;;           `(PROGN ,(CALL-NEXT-METHOD) (INCF *MY-CTR*)))
-             `(%no-primary-method ',generic-function .args.))
+      (cond ((null applicable-methods)
+             ;; APPLICABLE-METHODS is normally non-null in effective
+             ;; method computation, but COMPUTE-APPLICABLE-METHODS can
+             ;; in principle be called by MetaObject Protocol programmers.
+             `(method-combination-error
+               "No applicable method found for ~S"
+               ',generic-function))
+            ((null primary)
+             ;; PCL checks for no primary method before method
+             ;; combination, but a MetaObject Protocol programmer could
+             ;; call COMPUTE-EFFECTIVE-METHOD themselves and end up
+             ;; here.
+             `(method-combination-error
+               "No primary method found for ~S among applicable methods: ~S"
+               ',generic-function (list ,@(mapcar (lambda (m) `(quote ,m)) applicable-methods))))
             ((null around) main-method)
             (t
              `(call-method ,(car around)
@@ -548,7 +568,7 @@
                  (setq any-keyp t))
                (values nopt (ll-kwds-allowp llks) keys))))
       (multiple-value-bind (nopt allowp keys)
-          (analyze (generic-function-lambda-list gf))
+          (analyze (gf-lambda-list gf))
         (dolist (method methods)
           (let ((ll (if (consp method)
                         (early-method-lambda-list method)
@@ -595,6 +615,14 @@
                ((eq t valid-keys))
                ((not (memq key valid-keys)) (invalid key))))
            (incf i))))))
+
+(defun wrap-with-applicable-keyword-check (effective valid-keys keyargs-start)
+  `(let ((.valid-keys. ',valid-keys)
+         (.keyargs-start. ',keyargs-start))
+     (multiple-value-bind (.more-context. .more-count.) (sb-c::%rest-context .rest.)
+      (check-applicable-keywords
+       .keyargs-start. .valid-keys. .more-context. .more-count.))
+     ,effective))
 
 ;;;; the STANDARD method combination type. This is coded by hand
 ;;;; (rather than with DEFINE-METHOD-COMBINATION) for bootstrapping
@@ -613,15 +641,20 @@
                                      combin
                                      applicable-methods))
 
+;;; not INVALID-METHOD-ERROR as that would violate CLHS 11.1.2.1.1
+(define-condition invalid-method-program-error (program-error simple-condition)
+  ())
 (defun invalid-method-error (method format-control &rest format-arguments)
   (let ((sb-debug:*stack-top-hint* (find-caller-frame)))
-    (error "~@<invalid method error for ~2I~_~S ~I~_method: ~2I~_~?~:>"
-           method
-           format-control
-           format-arguments)))
+    (error 'invalid-method-program-error
+           :format-control "~@<invalid method error for ~2I~_~S ~I~_method: ~2I~_~?~:>"
+           :format-arguments (list method format-control format-arguments))))
 
+;;; not METHOD-COMBINATION-ERROR as that would violate CLHS 11.1.2.1.1
+(define-condition method-combination-program-error (program-error simple-condition)
+  ())
 (defun method-combination-error (format-control &rest format-arguments)
   (let ((sb-debug:*stack-top-hint* (find-caller-frame)))
-    (error "~@<method combination error in CLOS dispatch: ~2I~_~?~:>"
-           format-control
-           format-arguments)))
+    (error 'method-combination-program-error
+           :format-control "~@<method combination error in CLOS dispatch: ~2I~_~?~:>"
+           :format-arguments (list format-control format-arguments))))
