@@ -27,10 +27,16 @@ lispobj* atomic_bump_static_space_free_ptr(int nbytes)
 {
     gc_assert((nbytes & LOWTAG_MASK) == 0);
     lispobj* claimed_ptr = static_space_free_pointer;
+#ifdef LISP_FEATURE_X86_64
+    // Must not clobber some constant data on the final page of static space
+    lispobj* limit = static_space_trailer_start;
+#else
+    lispobj* limit = (lispobj*)STATIC_SPACE_END;
+#endif
     do {
         lispobj* new = (lispobj*)((char*)claimed_ptr + nbytes);
         // Fail if space exhausted or bogusly wrapped around
-        if (new > (lispobj*)STATIC_SPACE_END || new < claimed_ptr) return 0;
+        if (new > limit || new < claimed_ptr) return 0;
         lispobj* actual_old = __sync_val_compare_and_swap(&static_space_free_pointer,
                                                           claimed_ptr, new);
         if (actual_old == claimed_ptr) return claimed_ptr;
@@ -203,7 +209,7 @@ void close_code_region() {
 void gc_close_thread_regions(__attribute__((unused)) struct thread* th,
                              int locking) {
     close_region_arg argv[] = {
-#ifdef LISP_FEATURE_SB_THREAD
+#if defined LISP_FEATURE_SB_THREAD || defined LISP_FEATURE_X86_64
       { &th->mixed_tlab, PAGE_TYPE_MIXED },
       { &th->cons_tlab, PAGE_TYPE_CONS },
       { &th->sys_mixed_tlab, PAGE_TYPE_MIXED },
@@ -283,13 +289,11 @@ lispobj alloc_code_object(unsigned total_words, unsigned boxed)
     // 'boxed_size' is an untagged word expressing the number of *bytes* in the boxed section
     // (so CODE-INSTRUCTIONS can simply add rather than shift and add).
     code->boxed_size = boxed * N_WORD_BYTES;
-    code->debug_info = 0;
-    code->fixups = 0;
-    lispobj* p = &code->constants[0], *end = (lispobj*)code + boxed;
-    /* Must intialize to an arbitrary non-pointer value so that GC doesn't crash after the
-     * size is assigned (at some point prior to storing the constants) */
-    for ( ; p < end ; ++p) *p = 0;
-    *p = 0; // 'p' now points to the jump table count word which must be 0
+    // GC mustn't see uninitialized data. And one word past the boxed words (which holds the
+    // count of following words containing absolute jump addresses) must also be pre-zeroed.
+    lispobj* begin = 1 + &code->boxed_size, *end = (lispobj*)code + boxed + 1;
+    memset(begin, 0, (char*)end - (char*)begin);
+
     ((lispobj*)code)[total_words-1] = 0; // zeroize the simple-fun table count
     THREAD_JIT_WP(1);
     return make_lispobj(code, OTHER_POINTER_LOWTAG);
@@ -326,10 +330,11 @@ NO_SANITIZE_MEMORY lispobj alloc_funinstance(sword_t nbytes)
 /* Make a list that couldn't be inline-allocated. Break it up into contiguous
  * blocks of conses not to exceed one GC page each. */
 NO_SANITIZE_MEMORY lispobj
-make_list(lispobj element, sword_t nbytes, int sys) {
+make_list(lispobj element, sword_t nelts, int sys) {
     // Technically this overflow handler could permit garbage collection
     // between separate allocation. For now the entire thing is pseudo-atomic.
     struct thread *self = get_sb_vm_thread();
+    sword_t nbytes = nelts << WORD_SHIFT; // fixnum input
     PREPARE_LIST_ALLOCATION();
     lispobj result, *tail = &result;
     do {
@@ -354,9 +359,10 @@ make_list(lispobj element, sword_t nbytes, int sys) {
 /* Convert a &MORE context to a list. Split it up like make_list if we have to */
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
 NO_SANITIZE_MEMORY lispobj
-listify_rest_arg(lispobj* context, sword_t nbytes, int sys) {
+listify_rest_arg(lispobj* context, sword_t nelts, int sys) {
     // same comment as above in make_list() applies about the scope of pseudo-atomic
     struct thread *self = get_sb_vm_thread();
+    sword_t nbytes = nelts << WORD_SHIFT; // fixnum input
     PREPARE_LIST_ALLOCATION();
     lispobj result, *tail = &result;
     do {
