@@ -338,6 +338,14 @@ print_entry_points (struct code *code, FILE *f)
     });
 }
 
+/* SBCL itself uses print_lisp_backtrace() with one of stdout or
+ * stderr. This is the old interface, in case people debug with it. */
+void
+lisp_backtrace(int nframes)
+{
+    void print_lisp_backtrace(int frames, FILE *f);
+    print_lisp_backtrace(nframes, stdout);
+}
 
 #if !(defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64))
 
@@ -475,7 +483,7 @@ int lisp_frame_previous(struct thread *thread, struct call_info *info)
 }
 
 void
-lisp_backtrace(int nframes)
+print_lisp_backtrace(int nframes, FILE *f)
 {
     struct thread *thread = get_sb_vm_thread();
     struct call_info info;
@@ -491,41 +499,41 @@ lisp_backtrace(int nframes)
     do {
         if (!lisp_frame_previous(thread, &info)) {
             if (info.frame) // 0 is normal termination of the call chain
-                printf("Bad frame pointer %p [valid range=%p..%p]\n", info.frame,
-                       thread->control_stack_start, thread->control_stack_end);
+                fprintf(f, "Bad frame pointer %p [valid range=%p..%p]\n", info.frame,
+                        thread->control_stack_start, thread->control_stack_end);
             break;
         }
-        printf("%4d: ", i);
+        fprintf(f, "%4d: ", i);
         // Print spaces to keep the alignment nice
         if (info.interrupted
 #ifdef reg_LRA
             || info.lra == NIL
 #endif
             ) {
-            putchar('[');
-            if (info.interrupted) { footnotes |= 1; putchar('I'); }
+            putc('[', f);
+            if (info.interrupted) { footnotes |= 1; putc('I', f); }
 #ifdef reg_LRA
-            if (info.lra == NIL) { footnotes |= 2; putchar('*'); }
+            if (info.lra == NIL) { footnotes |= 2; putc('*', f); }
 #endif
-            putchar(']');
-            if (!(info.lra == NIL && info.interrupted)) putchar(' ');
+            putc(']', f);
+            if (!(info.lra == NIL && info.interrupted)) putc(' ', f);
         } else {
-            printf("    ");
+            fprintf(f, "    ");
         }
-        printf("%p ", info.frame);
+        fprintf(f, "%p ", info.frame);
         void* absolute_pc = 0;
         if (info.code) {
             absolute_pc = (char*)info.code + info.pc;
-            printf("pc=%p {%p+%04x} ", absolute_pc, info.code, (int)info.pc);
+            fprintf(f, "pc=%p {%p+%04x} ", absolute_pc, info.code, (int)info.pc);
         } else {
             absolute_pc = (char*)info.pc;
-            printf("pc=%p ", absolute_pc);
+            fprintf(f, "pc=%p ", absolute_pc);
         }
 #ifdef reg_LRA
         // If LRA does not match the PC, print it. This should not happen.
         if (info.lra != make_lispobj(absolute_pc, OTHER_POINTER_LOWTAG)
             && info.lra != NIL)
-            printf("LRA=%p ", (void*)info.lra);
+            fprintf(f, "LRA=%p ", (void*)info.lra);
 #endif
 
         int fpvalid = (lispobj*)info.frame >= thread->control_stack_start
@@ -533,31 +541,31 @@ lisp_backtrace(int nframes)
 
         // If the FP is invalid, then quite likely we'd crash trying to find a
         // compiled-debug-fun because info.code is a wild pointer
-        if (!fpvalid) { printf(" BAD FRAME\n"); break; }
+        if (!fpvalid) { fprintf(f, " BAD FRAME\n"); break; }
 
         if (info.code) {
             lispobj name;
             if (absolute_pc &&
                 (name = debug_function_name_from_pc((struct code *)info.code, absolute_pc)))
-                print_entry_name(barrier_load(&name), stdout);
+                print_entry_name(barrier_load(&name), f);
             else
                 // I can't imagine a scenario where we have info.code
                 // but do not have an absolute_pc, or debug-fun can't be found.
                 // Anyway, we can uniquely identify code by serial# now.
-                printf("{code_serialno=%x}", code_serialno(info.code));
+                fprintf(f, "{code_serialno=%x}", code_serialno(info.code));
         }
 
-        putchar('\n');
+        putc('\n', f);
 
     } while (++i <= nframes);
-    if (footnotes) printf("Note: [I] = interrupted"
+    if (footnotes) fprintf(f, "Note: [I] = interrupted"
 #ifdef reg_LRA
-                          ", [*] = no LRA"
+                           ", [*] = no LRA"
 #endif
-                          "\n");
+                           "\n");
 }
 
-#else
+#else /* (defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64)) */
 
 static int
 altstack_pointer_p(__attribute__((unused)) struct thread* thread,
@@ -696,10 +704,12 @@ static void print_backtrace_frame(char *pc, void *fp, int i, FILE *f) {
 
 /* This function has been split from lisp_backtrace() to enable Lisp
  * backtraces from gdb with call backtrace_from_fp(...). Useful for
- * example when debugging threading deadlocks.
+ * example when debugging threading deadlocks. (SBCL internals call
+ * print_backtrace_from_fp() however, because the right stream to
+ * write to is context-dependent.)
  */
 void NO_SANITIZE_MEMORY
-log_backtrace_from_fp(struct thread* th, void *fp, int nframes, int start, FILE *f)
+print_backtrace_from_fp(struct thread* th, void *fp, int nframes, int start, FILE *f)
 {
   int i = start;
 
@@ -715,24 +725,24 @@ log_backtrace_from_fp(struct thread* th, void *fp, int nframes, int start, FILE
   fflush(f);
 }
 void backtrace_from_fp(void *fp, int nframes, int start) {
-    log_backtrace_from_fp(get_sb_vm_thread(), fp, nframes, start, stdout);
+    print_backtrace_from_fp(get_sb_vm_thread(), fp, nframes, start, stdout);
 }
 
 void print_backtrace_from_context(os_context_t *context, int nframes, FILE* file) {
     void *fp = (void *)os_context_frame_pointer(context);
     print_backtrace_frame((void *)os_context_pc(context), fp, 0, file);
-    log_backtrace_from_fp(get_sb_vm_thread(), fp, nframes - 1, 1, file);
+    print_backtrace_from_fp(get_sb_vm_thread(), fp, nframes - 1, 1, file);
 }
 
 void
-lisp_backtrace(int nframes)
+print_lisp_backtrace(int nframes, FILE *f)
 {
     struct thread *thread = get_sb_vm_thread();
     int free_ici = fixnum_value(read_TLS(FREE_INTERRUPT_CONTEXT_INDEX,thread));
 
     if (free_ici) {
         os_context_t *context = nth_interrupt_context(free_ici - 1, thread);
-        print_backtrace_from_context(context, nframes, stdout);
+        print_backtrace_from_context(context, nframes, f);
     } else {
         void *fp;
 
@@ -741,7 +751,7 @@ lisp_backtrace(int nframes)
 #elif defined (LISP_FEATURE_X86_64)
         asm("movq %%rbp,%0" : "=g" (fp));
 #endif
-        backtrace_from_fp(fp, nframes, 0);
+        print_backtrace_from_fp(get_sb_vm_thread(), fp, nframes, 0, f);
     }
 }
 #endif
@@ -759,16 +769,16 @@ int simple_fun_index_from_pc(struct code* code, char *pc)
     return -1;
 }
 
-static bool __attribute__((unused)) print_lisp_fun_name(char* pc)
+static bool __attribute__((unused)) print_lisp_fun_name(char* pc, FILE* f)
 {
   struct code* code;
   if (gc_managed_heap_space_p((uword_t)pc) &&
       (code = (void*)component_ptr_from_pc(pc)) != 0) {
       lispobj name = debug_function_name_from_pc(code, pc);
       if (name) {
-          fprintf(stderr, " %p [", pc);
-          print_entry_name(barrier_load(&name), stderr);
-          fprintf(stderr, "]\n");
+          fprintf(f, " %p [", pc);
+          print_entry_name(barrier_load(&name), f);
+          fprintf(f, "]\n");
           return 1;
       }
   }
@@ -778,6 +788,7 @@ static bool __attribute__((unused)) print_lisp_fun_name(char* pc)
 #ifdef HAVE_LIBUNWIND
 #define UNW_LOCAL_ONLY
 #include <libunwind.h>
+int sbcl_have_libunwind() { return 1; }
 int get_sizeof_unw_context() { return sizeof (unw_context_t); }
 int get_sizeof_unw_cursor() { return sizeof (unw_cursor_t); }
 #ifdef LISP_FEATURE_DARWIN // slightly different libunwind. And it doesn't work for me
@@ -790,6 +801,10 @@ int sb_unw_get_pc(void* a, void* b) { return unw_get_reg(a, UNW_TDEP_IP, b); }
 int sb_unw_get_proc_name(void* a, void* b, int c, unw_word_t* d) { return unw_get_proc_name(a, b, c, d); }
 int sb_unw_step(void* a) { return unw_step(a); }
 #else
+#ifdef LISP_FEATURE_BACKTRACE_ON_SIGNAL
+# include <ucontext.h>
+#endif
+int sbcl_have_libunwind() { return 0; }
 int get_sizeof_unw_context() { return 0; }
 int get_sizeof_unw_cursor() { return 0; }
 int sb_unw_init(void* a, void* b) { lose("unw_init %p %p", a, b); }
@@ -799,25 +814,44 @@ int sb_unw_get_pc(void* a, void* b) { lose("unw_get_pc %p %p", a, b); }
 #endif
 
 #ifdef LISP_FEATURE_BACKTRACE_ON_SIGNAL
-static __attribute__((unused))int backtrace_completion_pipe[2] = {-1,-1};
-void libunwind_backtrace(struct thread *th, os_context_t *context)
+static sem_t bt_suspend_sem;
+static os_context_t *bt_suspension_context;
+void libunwind_backtrace(struct thread *th, FILE* f)
 {
-    fprintf(stderr, "Lisp thread @ %p, tid %d", th, (int)th->os_kernel_tid);
-#ifdef LISP_FEATURE_SB_THREAD
+    ucontext_t ucontext;
+    static int initialized;
+    if (th != get_sb_vm_thread()) {
+        if (!initialized) sem_init(&bt_suspend_sem, 0, 0);
+        initialized = 1;
+        gc_assert(bt_suspension_context == 0);
+        // TODO: combine stop-for-gc and stop-for-backtrace into one signal, informing
+        // the signaled thread of a reason. Backtrace should never defer suspension.
+        pthread_kill(th->os_thread, SIGXCPU);
+        sem_wait(&bt_suspend_sem);
+        gc_assert(bt_suspension_context != 0);
+    } else {
+#ifdef HAVE_LIBUNWIND
+        unw_getcontext(&ucontext);
+#else
+        getcontext(&ucontext);
+#endif
+        bt_suspension_context = &ucontext;
+    }
+    fprintf(f, "Lisp thread @ %p, tid %d", th, (int)th->os_kernel_tid);
     // the TLS area is not used if #-sb-thread. And if so, it must be "main thread"
     struct thread_instance* lispthread = (void*)native_pointer(th->lisp_thread);
     if (lispthread->_name != NIL) {
-        fprintf(stderr, " (\"");
-        print_string(VECTOR(lispthread->_name), stderr);
-        fprintf(stderr, "\")");
+        fprintf(f, " (\"");
+        print_string(VECTOR(lispthread->_name), f);
+        fprintf(f, "\")");
     }
-    putc('\n', stderr);
+    putc('\n', f);
     // In case you get no backtrace whatsoever, maybe at least see where the
     // signal was received, probably in a function without the standard
     // frame pointer setup.
-    fprintf(stderr, " interrupted @ PC %p\n", (void*)OS_CONTEXT_PC(context));
+    fprintf(f, " interrupted @ PC %p\n", (void*)OS_CONTEXT_PC(bt_suspension_context));
     if (lispthread->waiting_for != NIL) {
-        fprintf(stderr, "waiting for %p", (void*)lispthread->waiting_for);
+        fprintf(f, "waiting for %p", (void*)lispthread->waiting_for);
         if (instancep(lispthread->waiting_for)) {
             // THREAD-WAITING-FOR can be a mutex or a waitqueue (if not a cons).
             // Accessing it as if it's a mutex works because both a waitqueue
@@ -826,84 +860,55 @@ void libunwind_backtrace(struct thread *th, os_context_t *context)
             // "Use this only if you know what you're doing"
             struct lispmutex* lispmutex = (void*)native_pointer(lispthread->waiting_for);
             if (lispmutex->name != NIL) {
-                fprintf(stderr, " (MUTEX:\"");
+                fprintf(f, " (MUTEX:\"");
                 print_string(VECTOR(lispmutex->name), stderr);
-                fprintf(stderr, "\")");
+                fprintf(f, "\")");
             }
         }
-        putc('\n', stderr);
+        putc('\n', f);
     }
-#endif
 #ifdef HAVE_LIBUNWIND
     char procname[100];
     unw_cursor_t cursor;
-    // "unw_init_local() is thread-safe as well as safe to use from a signal handler."
-    // "unw_get_proc_name() is thread-safe. If cursor cp is in the local address-space,
-    //  this routine is also safe to use from a signal handler."
-    if (context) {
-        unw_init_local(&cursor, context);
-    } else {
-        unw_context_t here;
-        unw_getcontext(&here);
-        unw_init_local(&cursor, &here);
-    }
+    unw_init_local(&cursor, bt_suspension_context);
+    // This thread performs the backtrace of the other thread, so it matters not
+    // whether libunwind functions are interrupt-safe (though they do claim to be)
     do {
         uword_t offset;
         char *pc;
         unw_get_reg(&cursor, UNW_TDEP_IP, (uword_t*)&pc);
-        if (print_lisp_fun_name(pc)) {
+        if (print_lisp_fun_name(pc, f)) {
             // printed
         } else if (!unw_get_proc_name(&cursor, procname, sizeof procname, &offset)) {
-            fprintf(stderr, " %p [%s]\n", pc, procname);
+            fprintf(f, " %p [%s]\n", pc, procname);
         } else {
-            fprintf(stderr, " %p ?\n", pc);
+            fprintf(f, " %p ?\n", pc);
         }
     } while (unw_step(&cursor));
 #else
     // If you don't have libunwind, this will almost surely not work,
     // because we can't figure out how to get backwards past a signal frame.
-    log_backtrace_from_fp(th, (void*)*os_context_fp_addr(context), 100, 0, stderr);
+    print_backtrace_from_fp(th, (void*)*os_context_fp_addr(context), 100, 0, stderr);
 #endif
+    bt_suspension_context = 0;
+    if (th != get_sb_vm_thread()) pthread_kill(th->os_thread, SIGXCPU);
 }
-void backtrace_lisp_threads(int __attribute__((unused)) signal,
-                                   siginfo_t __attribute__((unused)) *info,
-                                   os_context_t *context)
+__thread int suspended_for_bt;
+void suspend_for_backtrace(int signal, siginfo_t __attribute__((unused)) *info,
+                           os_context_t *context)
 {
-    struct thread* this_thread = get_sb_vm_thread();
-#ifdef LISP_FEATURE_SB_THREAD
-    if (backtrace_completion_pipe[1] >= 0) {
-        libunwind_backtrace(get_sb_vm_thread(), context);
-        write(backtrace_completion_pipe[1], context /* any random byte */, 1);
-        return;
+    if (suspended_for_bt) {
+        suspended_for_bt = 0;
+    } else {
+        gc_assert(!bt_suspension_context);
+        bt_suspension_context = context;
+        sem_post(&bt_suspend_sem);
+        suspended_for_bt = 1;
+        sigset_t mask;
+        sigfillset(&mask);
+        sigdelset(&mask, signal);
+        sigsuspend(&mask);
     }
-    struct thread *th;
-    int nthreads = 0;
-    for_each_thread(th) { ++nthreads; }
-    if (signal)
-        fprintf(stderr, "Caught backtrace-all signal in tid %d, %d threads\n",
-                (int)this_thread->os_kernel_tid, nthreads);
-    // Would be nice if we could forcibly stop all the other threads,
-    // but pthread_mutex_trylock is not safe to use in a signal handler.
-    if (nthreads > 1) {
-        pipe(backtrace_completion_pipe);
-    }
-    for_each_thread(th) {
-        if (th == this_thread)
-            libunwind_backtrace(th, context);
-        else {
-            char junk;
-            pthread_kill(th->os_thread, SIGXCPU);
-            read(backtrace_completion_pipe[0], &junk, 1);
-        }
-    }
-    if (nthreads > 1) {
-        close(backtrace_completion_pipe[1]);
-        close(backtrace_completion_pipe[0]);
-        backtrace_completion_pipe[0] = backtrace_completion_pipe[1] = -1;
-    }
-#else
-    libunwind_backtrace(this_thread, context);
-#endif
 }
 static int watchdog_pipe[2] = {-1,-1};
 static pthread_t watchdog_tid;
@@ -919,7 +924,8 @@ static void* watchdog_thread(void* arg) {
         if (nfds < 1) {
             // Ensure this message comes out in one piece even if nothing following it does.
             char msg[] = "Watchdog timer expired\n"; write(2, msg, sizeof msg-1);
-            backtrace_lisp_threads(0, 0, 0);
+            struct thread* th;
+            for_each_thread(th) { libunwind_backtrace(th, stderr); }
             _exit(1); // cause the test suite to exit with failure
         }
         if (!FD_ISSET(watchdog_pipe[0], &fds))
