@@ -232,6 +232,7 @@ struct heap_adjust {
     int n_ranges;
     int n_relocs_abs; // absolute
     int n_relocs_rel; // relative
+    uword_t symhash_adjust;
 };
 
 #include "genesis/gc-tables.h"
@@ -284,6 +285,14 @@ set_adjustment(struct cons* pair, int id, uword_t actual_addr)
     sword_t len = spaces[id].len;
     sword_t delta = len ? actual_addr - desired_addr : 0;
     if (!delta) return;
+#ifdef LISP_FEATURE_X86_64
+    if (id == STATIC_CORE_SPACE_ID) {
+        lispobj nil_want = (lispobj)desired_addr + NIL_VALUE_OFFSET;
+        lispobj nil_have = (lispobj)actual_addr + NIL_VALUE_OFFSET;
+        adj->symhash_adjust = (nil_want ^ nil_have) & SYMBOL_HASH_MASK;
+        len = STATIC_SPACE_SIZE; // T and NIL are above the supplied 'len'
+    }
+#endif
     int j = adj->n_ranges;
     adj->range[j].start = (lispobj)desired_addr;
     adj->range[j].end   = (lispobj)desired_addr + len;
@@ -419,6 +428,7 @@ static void fix_space(uword_t start, lispobj* end, struct heap_adjust* adj)
             // Modeled on scav_symbol() in gc-common
             struct symbol* s = (void*)where;
 #ifdef LISP_FEATURE_64_BIT
+            s->hash ^= adj->symhash_adjust;
             lispobj name = decode_symbol_name(s->name);
             lispobj adjusted_name = adjust_word(adj, name);
             // writeback the name if it changed
@@ -529,11 +539,19 @@ static void fix_space(uword_t start, lispobj* end, struct heap_adjust* adj)
 
         // Other
         case SAP_WIDETAG:
+#if defined LISP_FEATURE_X86_64
+            if (((uint32_t*)where)[1]) { // high half of header != 0 ==> static-space-relative
+                FIXUP(where[1] = STATIC_SPACE_START + ((uint32_t*)where)[1],
+                      where+1);
+            } else
+#endif
+            // guess whether the SAP looks like it should be adjusted
             if ((delta = calc_adjustment(adj, where[1])) != 0) {
-                fprintf(stderr,
-                        "WARNING: SAP at %p -> %p in relocatable core\n",
-                        where, (void*)where[1]);
-                FIXUP(where[1] += delta, where+1);
+                if (!lisp_startup_options.noinform) {
+                    fprintf(stderr, "WARNING: SAP at %p -> %p in relocatable core\n",
+                            where, (void*)where[1]);
+                }
+                // FIXUP(where[1] += delta, where+1);
             }
             continue;
         default:
@@ -577,13 +595,17 @@ static void relocate_heap(struct heap_adjust* adj)
                         (char*)adj->range[i].start + adj->range[i].delta,
                         (char*)adj->range[i].end + adj->range[i].delta);
     }
-#ifdef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
+#if defined LISP_FEATURE_RELOCATABLE_STATIC_SPACE && !defined LISP_FEATURE_X86_64
     fix_space(READ_ONLY_SPACE_START, read_only_space_free_pointer, adj);
     // Relocate the CAR slot of nil-as-a-list, which needs to point to
     // itself.
     adjust_pointers((void*)(NIL - LIST_POINTER_LOWTAG), 1, adj);
+#endif
+#ifdef T_SYMBOL_SLOTS_START
+    fix_space((uword_t)T_SYMBOL_SLOTS_START, T_SYMBOL_SLOTS_END, adj);
 #endif
     fix_space((uword_t)NIL_SYMBOL_SLOTS_START, NIL_SYMBOL_SLOTS_END, adj);
+    CONS(NIL)->car = NIL;
     fix_space(STATIC_SPACE_OBJECTS_START, static_space_free_pointer, adj);
 #ifdef LISP_FEATURE_PERMGEN
     fix_space(PERMGEN_SPACE_START, permgen_space_free_pointer, adj);
@@ -594,10 +616,8 @@ static void relocate_heap(struct heap_adjust* adj)
     fix_space(DYNAMIC_SPACE_START, (lispobj*)dynamic_space_highwatermark(), adj);
     fix_space(TEXT_SPACE_START, text_space_highwatermark, adj);
 #if defined LISP_FEATURE_X86_64 && defined LISP_FEATURE_IMMOBILE_SPACE
-    /* Update the static-space asm routine indirection vector
-     * which immediately follows LFLIST_TAIL_ATOM in memory */
-    lispobj* tail_atom = native_pointer(LFLIST_TAIL_ATOM);
-    struct vector* v = (void*)(tail_atom + object_size(tail_atom));
+    /* Update the static-space asm routine indirection vector */
+    struct vector* v = (void*)static_space_trailer_start;
     gc_assert(widetag_of((lispobj*)v) == SIMPLE_ARRAY_UNSIGNED_BYTE_64_WIDETAG);
     struct code* c = (void*)asm_routines_start;
     lispobj* jump_table = code_jumptable_start(c);
@@ -665,6 +685,13 @@ void calc_immobile_space_bounds()
         {FIXEDOBJ_SPACE_START, FIXEDOBJ_SPACE_START + FIXEDOBJ_SPACE_SIZE};
     struct range range2 =
         {TEXT_SPACE_START, TEXT_SPACE_START + text_space_size};
+    if (range1.start == 0) {
+        immobile_space_lower_bound  = range2.start;
+        immobile_space_max_offset   = range2.end - range2.start;
+        immobile_range_1_max_offset = immobile_space_max_offset;
+        immobile_range_2_min_offset = immobile_space_max_offset;
+        return;
+    }
     if (range2.start < range1.start) { // swap
         struct range temp = range1;
         range1 = range2;
@@ -677,27 +704,16 @@ void calc_immobile_space_bounds()
 }
 #endif
 
-#ifdef LISP_FEATURE_LINKAGE_SPACE
-#define LISP_LINKAGE_SPACE_SIZE (1<<(N_LINKAGE_INDEX_BITS+WORD_SHIFT))
-#endif
-
-#if defined LISP_FEATURE_IMMOBILE_SPACE && defined LISP_FEATURE_ARM64
-#define LISP_LINKAGE_SPACE_SIZE 0
-#endif
-static os_vm_address_t reserve_space(int space_id, int attr,
-                                     os_vm_address_t addr, os_vm_size_t size)
+#if defined LISP_FEATURE_X86_64 || defined LISP_FEATURE_ARM64
+extern
+os_vm_address_t coreparse_alloc_space(int space_id, int attr,
+                                      os_vm_address_t addr, os_vm_size_t size);
+#else
+static os_vm_address_t coreparse_alloc_space(int space_id, int attr,
+                                             os_vm_address_t addr, os_vm_size_t size)
 {
-    __attribute__((unused)) int extra_request = 0;
-#ifdef LISP_FEATURE_IMMOBILE_SPACE
-    if (space_id == IMMOBILE_TEXT_CORE_SPACE_ID) {
-        extra_request = ALIEN_LINKAGE_SPACE_SIZE + LISP_LINKAGE_SPACE_SIZE;
-        size += extra_request;
-        addr -= extra_request; // compensate to try to put text space start where expected
-    }
-#endif
     if (size == 0) return addr;
-    // 64-bit already allocated a trap page when the GC card mark table was made
-#if defined(LISP_FEATURE_SB_SAFEPOINT) && !defined(LISP_FEATURE_X86_64)
+#ifdef LISP_FEATURE_SB_SAFEPOINT // this is only for 32-bit x86, I think
     if (space_id == STATIC_CORE_SPACE_ID) {
         // Allocate space for the safepoint page.
         addr = os_alloc_gc_space(space_id, attr, addr - BACKEND_PAGE_BYTES, size + BACKEND_PAGE_BYTES) + BACKEND_PAGE_BYTES;
@@ -706,15 +722,9 @@ static os_vm_address_t reserve_space(int space_id, int attr,
 #endif
         addr = os_alloc_gc_space(space_id, attr, addr, size);
     if (!addr) lose("Can't allocate %#"OBJ_FMTX" bytes for space %d", size, space_id);
-#ifdef LISP_FEATURE_IMMOBILE_SPACE
-    if (space_id == IMMOBILE_TEXT_CORE_SPACE_ID) {
-      ALIEN_LINKAGE_SPACE_START = (uword_t)addr;
-      linkage_space = (void*)((char*)addr + ALIEN_LINKAGE_SPACE_SIZE);
-      addr += extra_request;
-    }
-#endif
     return addr;
 }
+#endif
 
 static __attribute__((unused)) uword_t corespace_checksum(uword_t* base, int nwords)
 {
@@ -762,23 +772,8 @@ process_directory(int count, struct ndir_entry *entry,
 #endif
         // unprotect the pages
         os_protect((void*)TEXT_SPACE_START, text_space_size, OS_VM_PROT_ALL);
-
-#ifdef LISP_FEATURE_IMMOBILE_SPACE
-        // ELF core without immobile space has alien linkage space below static space.
-        ALIEN_LINKAGE_SPACE_START =
-            (uword_t)os_alloc_gc_space(ALIEN_LINKAGE_TABLE_CORE_SPACE_ID, 0, 0,
-                                       ALIEN_LINKAGE_SPACE_SIZE);
-#endif
-    } else
-#endif
-    // NB: The preceding 'else', if it's there, needs at least an empty
-    // statement following it if there is no immobile space.
-    {
-#ifdef LISP_FEATURE_IMMOBILE_SPACE
-        spaces[IMMOBILE_FIXEDOBJ_CORE_SPACE_ID].desired_size +=
-            text_space_size + ALIEN_LINKAGE_SPACE_SIZE;
-#endif
     }
+#endif
 
     for ( ; --count >= 0 ; ++entry) {
         long id = entry->identifier; // FIXME: is this 'long' and not 'int' for a reason?
@@ -813,10 +808,10 @@ process_directory(int count, struct ndir_entry *entry,
 #else
         if (id == READ_ONLY_CORE_SPACE_ID) {
 #endif
-            if (len) // There is no "nominal" size of static or
+            if (len) { // There is no "nominal" size of static or
                      // readonly space, so give them a size
-                spaces[id].desired_size = len;
-            else { // Assign some address, so free_pointer does enclose [0 .. addr+0]
+                if (len > spaces[id].desired_size) spaces[id].desired_size = len;
+            } else { // Assign some address, so free_pointer does enclose [0 .. addr+0]
                 if (id == STATIC_CORE_SPACE_ID)
                     lose("Static space size is 0?");
                 READ_ONLY_SPACE_START = READ_ONLY_SPACE_END = addr;
@@ -831,8 +826,8 @@ process_directory(int count, struct ndir_entry *entry,
 #ifdef LISP_FEATURE_ASLR
             addr = 0;
 #endif
-            addr = (uword_t)reserve_space(id, sub_2gb_flag ? MOVABLE_LOW : MOVABLE,
-                                          (os_vm_address_t)addr, request);
+            addr = (uword_t)coreparse_alloc_space(id, sub_2gb_flag ? MOVABLE_LOW : MOVABLE,
+                                                  (os_vm_address_t)addr, request);
             switch (id) {
             case PERMGEN_CORE_SPACE_ID:
                 PERMGEN_SPACE_START = addr;
@@ -840,7 +835,6 @@ process_directory(int count, struct ndir_entry *entry,
             case STATIC_CORE_SPACE_ID:
 #ifdef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
                 STATIC_SPACE_START = addr;
-                STATIC_SPACE_END = addr + len;
 #endif
                 break;
             case READ_ONLY_CORE_SPACE_ID:
@@ -923,9 +917,9 @@ process_directory(int count, struct ndir_entry *entry,
 
 #ifdef LISP_FEATURE_LINKAGE_SPACE
     if (linkage_table_count) {
-        // Only #-immobile-space should allocate linkage space now,
-        // because otherwise it must be contiguous with text space.
-        if (linkage_space == 0) linkage_space = (void*)os_allocate(LISP_LINKAGE_SPACE_SIZE);
+#ifdef LISP_FEATURE_PPC64
+        linkage_space = (void*)os_allocate(LISP_LINKAGE_SPACE_SIZE);
+#endif
         off_t filepos = file_offset + (1 + linkage_table_data_page) * os_vm_page_size;
         gc_assert(os_reported_page_size);
         // Linkage space is only 8-byte-aligned in an ELF core. It doesn't need more than that.
@@ -1089,13 +1083,12 @@ void gc_allocate_ptes()
 #endif
 
     long num_gc_cards = 1 + gc_card_table_mask;
-#if defined LISP_FEATURE_SB_SAFEPOINT && defined LISP_FEATURE_X86_64
-    /* The card table is hardware-page-aligned. Preceding it and occupying a whole
-     * "backend page" - which by the way is overkill - is the global safepoint trap page.
-     * The dummy TEST instruction for safepoints encodes shorter this way */
-    void* result = os_alloc_gc_space(0, MOVABLE, 0,
-                                     ALIGN_UP(num_gc_cards, BACKEND_PAGE_BYTES) + BACKEND_PAGE_BYTES);
-    gc_card_mark = (unsigned char*)result + BACKEND_PAGE_BYTES;
+#ifdef LISP_FEATURE_X86_64
+# ifdef LISP_FEATURE_SB_SAFEPOINT
+    gc_card_mark = (void*)(STATIC_SPACE_END + BACKEND_PAGE_BYTES);
+# else
+    gc_card_mark = (void*)STATIC_SPACE_END;
+# endif
 #elif defined LISP_FEATURE_PPC64
     unsigned char* mem = checked_malloc(num_gc_cards + LISP_LINKAGE_SPACE_SIZE);
     gc_card_mark = mem + LISP_LINKAGE_SPACE_SIZE;
@@ -1420,7 +1413,11 @@ load_core_file(char *file, os_vm_offset_t file_offset, int merge_core_pages)
 
     struct coreparse_space defined_spaces[] = {
         {READ_ONLY_CORE_SPACE_ID, 0, 0, READ_ONLY_SPACE_START, &read_only_space_free_pointer},
+#ifdef LISP_FEATURE_X86_64 // specify the static space size in order to allocate it
+        {STATIC_CORE_SPACE_ID, STATIC_SPACE_SIZE, 0, STATIC_SPACE_START, &static_space_free_pointer},
+#else                      // must not specify the static space size
         {STATIC_CORE_SPACE_ID, 0, 0, STATIC_SPACE_START, &static_space_free_pointer},
+#endif
 #ifdef LISP_FEATURE_PERMGEN
         {PERMGEN_CORE_SPACE_ID, PERMGEN_SPACE_SIZE | 1, 0,
          PERMGEN_SPACE_START, &permgen_space_free_pointer},
@@ -1493,6 +1490,14 @@ load_core_file(char *file, os_vm_offset_t file_offset, int merge_core_pages)
                 lose("core was built for runtime \"%.*s\" but this is \"%s\"",
                      (int)stringlen, (char*)ptr, build_id);
             break;
+        case STATIC_CONSTANTS_CORE_ENTRY_TYPE_CODE: {
+            int nwords = remaining_len;
+            lispobj* base = (lispobj*)
+              ((STATIC_SPACE_START+STATIC_SPACE_SIZE) - (nwords<<WORD_SHIFT));
+            static_space_trailer_start = base;
+            memcpy(base, ptr, nwords<<WORD_SHIFT);
+            break;
+        }
         case DIRECTORY_CORE_ENTRY_TYPE_CODE:
             process_directory(remaining_len / NDIR_ENTRY_LENGTH,
                               (struct ndir_entry*)ptr,
@@ -1590,9 +1595,47 @@ void asm_routine_poke(const char* routine, int offset, char byte)
         address[offset] = byte;
 }
 
-static void trace_sym(lispobj, struct symbol*, struct grvisit_context*);
+static void graph_visit(lispobj referer, lispobj ptr, struct grvisit_context* context);
+
+#ifdef LISP_FEATURE_LINKAGE_SPACE
+static void trace_linkage_cells(lispobj packed_int, lispobj codeblob,
+                                struct grvisit_context* context)
+{
+    const unsigned int smallvec_elts =
+        (GENCGC_PAGE_BYTES - offsetof(struct vector,data)) / N_WORD_BYTES;
+    if (!packed_int) return; // do no work for leaf codeblobs
+    lispobj name_map = SYMBOL(LINKAGE_NAME_MAP)->value;
+    if (name_map == UNBOUND_MARKER_WIDETAG) return; /* cold-init perhaps */
+    gc_assert(simple_vector_p(name_map));
+    struct vector* outer_vector = VECTOR(name_map);
+    struct varint_unpacker unpacker;
+    varint_unpacker_init(&unpacker, packed_int);
+    int prev_index = 0, index;
+    while (varint_unpack(&unpacker, &index) && index != 0) {
+        index += prev_index;
+        prev_index = index;
+        int index_high = (unsigned int)index / smallvec_elts;
+        int index_low = (unsigned int)index % smallvec_elts;
+        lispobj smallvec = outer_vector->data[index_high];
+        gc_assert(other_pointer_p(smallvec));
+        struct vector* inner_vector = VECTOR(smallvec);
+        lispobj name = inner_vector->data[index_low];
+        gc_assert(is_lisp_pointer(name));
+        graph_visit(codeblob, name, context);
+    }
+}
+#else
+#define trace_linkage_cells(dummy1,dummy2,dummy3)
+#endif
 
 #define RECURSE(x) if(is_lisp_pointer(x))graph_visit(ptr,x,context)
+static void trace_sym(lispobj ptr, struct symbol* sym, struct grvisit_context* context)
+{
+    RECURSE(decode_symbol_name(sym->name));
+    RECURSE(sym->value);
+    RECURSE(sym->info);
+    RECURSE(sym->fdefn);
+}
 
 /* Despite this being a nice concise expression of a pointer tracing algorithm,
  * it turns out to be almost unusable in any sufficiently complicated object graph
@@ -1616,10 +1659,11 @@ static void graph_visit(lispobj referer, lispobj ptr, struct grvisit_context* co
         RECURSE(CONS(ptr)->cdr);
     } else switch (widetag_of(obj = native_pointer(ptr))) {
         case SIMPLE_VECTOR_WIDETAG:
-            {
-            struct vector* v = (void*)obj;
-            sword_t len = vector_len(v);
-            for(i=0; i<len; ++i) RECURSE(v->data[i]);
+            if (vector_is_weak_not_hashing_p(VECTOR(ptr)->header)) {
+            } else {
+                struct vector* v = (void*)obj;
+                sword_t len = vector_len(v);
+                for(i=0; i<len; ++i) RECURSE(v->data[i]);
             }
             break;
         case INSTANCE_WIDETAG:
@@ -1642,6 +1686,7 @@ static void graph_visit(lispobj referer, lispobj ptr, struct grvisit_context* co
         case CODE_HEADER_WIDETAG:
             nwords = code_header_words((struct code*)obj);
             for(i=2; i<nwords; ++i) RECURSE(obj[i]);
+            trace_linkage_cells(((struct code*)obj)->fixups, ptr, context);
             break;
         // In all the remaining cases, 'nwords' is the count of payload words
         // (following the header), so we iterate up to and including that
@@ -1659,16 +1704,18 @@ static void graph_visit(lispobj referer, lispobj ptr, struct grvisit_context* co
             break;
         case SYMBOL_WIDETAG:
             trace_sym(ptr, SYMBOL(ptr), context);
+#ifdef LISP_FEATURE_LINKAGE_SPACE
+            RECURSE(linkage_cell_function(symbol_linkage_index(SYMBOL(ptr))));
+#endif
             break;
         case FDEFN_WIDETAG:
-            RECURSE(((struct fdefn*)obj)->name);
-            RECURSE(((struct fdefn*)obj)->fun);
-            RECURSE(decode_fdefn_rawfun((struct fdefn*)obj));
+            RECURSE(FDEFN(ptr)->name);
+            RECURSE(FDEFN(ptr)->fun);
+            RECURSE(decode_fdefn_rawfun(FDEFN(ptr)));
             break;
         default:
-            // weak-pointer can be considered an ordinary boxed object.
-            // the 'next' link looks like a fixnum.
-            if (!leaf_obj_widetag_p(widetag_of(obj))) {
+            if (!leaf_obj_widetag_p(widetag_of(obj))
+                && widetag_of(obj) != WEAK_POINTER_WIDETAG) {
                 sword_t size = headerobj_size(obj);
                 for(i=1; i<size; ++i) RECURSE(obj[i]);
             }
@@ -1676,14 +1723,6 @@ static void graph_visit(lispobj referer, lispobj ptr, struct grvisit_context* co
       --context->depth;
 }
 
-static void trace_sym(lispobj ptr, struct symbol* sym, struct grvisit_context* context)
-{
-    RECURSE(decode_symbol_name(sym->name));
-    RECURSE(sym->value);
-    RECURSE(sym->info);
-    RECURSE(sym->fdefn);
-}
-
 /* Caller must provide an uninitialized hopscotch table.
  * This function will initialize it and perform a graph visit.
  * Caller may subsequently inspect the table and/or visit other objects as
@@ -1704,7 +1743,8 @@ visit_heap_from_static_roots(struct hopscotch_table* reached,
     context->action = action;
     context->data = data;
     context->depth = context->maxdepth = 0;
-    trace_sym(NIL, SYMBOL(NIL), context);
+    // ppc64 does not like SYMBOL(NIL). I wonder if this is a cause of problems elsewhere
+    trace_sym(NIL, (void*)NIL_SYMBOL_SLOTS_START, context);
     lispobj* where = (lispobj*)STATIC_SPACE_OBJECTS_START;
     lispobj* end = static_space_free_pointer;
     while (where<end) {
