@@ -9,9 +9,10 @@
  * files for more information.
  */
 
-/* All parse routines take a char ** as their only argument */
-extern boolean more_p(char **ptr);
+#include <stdbool.h>
+#include <stdio.h>
+extern bool more_p(char **ptr);
 extern char *parse_token(char **ptr);
-extern lispobj parse_lispobj(char **ptr);
-extern char *parse_addr(char **ptr);
-extern uword_t parse_number(char **ptr);
+extern int parse_lispobj(char **ptr, lispobj *output);
+extern int parse_addr(char **ptr, bool safely, char **output, FILE*);
+extern int parse_number(char **ptr, int *output);
