@@ -256,6 +256,19 @@ static void unwind_binding_stack(struct thread* th)
     if (verbose) printf("done]\n");
 }
 
+#ifdef LISP_FEATURE_X86_64
+static void write_static_space_constants(FILE *file)
+{
+    lispobj* ptr = (lispobj*)static_space_trailer_start;
+    unsigned int nwords = (lispobj*)STATIC_SPACE_END - ptr;
+    write_lispobj(STATIC_CONSTANTS_CORE_ENTRY_TYPE_CODE, file);
+    // +2 for the core header entry itself
+    write_lispobj(2+nwords, file);
+    /* write_lispobj(STATIC_SPACE_START, file); */
+    if (fwrite(ptr, N_WORD_BYTES, nwords, file) != nwords) perror(GENERAL_WRITE_FAILURE_MSG);
+}
+#endif
+
 bool save_to_filehandle(FILE *file, char *filename, lispobj init_function,
                         bool make_executable,
                         int save_runtime_options,
@@ -295,11 +308,7 @@ bool save_to_filehandle(FILE *file, char *filename, lispobj init_function,
     write_lispobj(2, file);
 #endif
     write_lispobj(gc_card_table_nbits, file);
-#ifdef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
-    write_lispobj(0, file);
-#else
     write_lispobj(NIL, file);
-#endif
     write_lispobj(stringlen, file);
     int nwrote = fwrite(build_id, 1, stringlen, file);
     /* Write padding bytes to align to core_entry_elt_t */
@@ -400,6 +409,10 @@ bool save_to_filehandle(FILE *file, char *filename, lispobj init_function,
                  core_start_pos,
                  core_compression_level), ++count;
 
+#ifdef LISP_FEATURE_X86_64
+    write_static_space_constants(file);
+#endif
+
     write_lispobj(INITIAL_FUN_CORE_ENTRY_TYPE_CODE, file);
     write_lispobj(3, file);
     write_lispobj(init_function, file);
@@ -732,9 +745,10 @@ gc_and_save(char *filename, bool prepend_runtime, bool purify,
     // FIXME: Instead of disabling purify for static space relocation,
     // we should make r/o space read-only after fixing up pointers to
     // static space instead.
+    // BUT: On x86-64 it's ok to use the R/O space. All objects there are leaves.
 #if ((defined LISP_FEATURE_SPARC && defined LISP_FEATURE_LINUX) || \
-     (defined LISP_FEATURE_RELOCATABLE_STATIC_SPACE))
-    /* OS says it'll give you the memory where you want, then it says
+     (defined LISP_FEATURE_RELOCATABLE_STATIC_SPACE && !defined LISP_FEATURE_X86_64))
+    /* SunOS says it'll give you the memory where you want, then it says
      * it won't map over it from the core file.  That's news to me.
      * Fragment of output from 'strace -e mmap2 src/runtime/sbcl --core output/sbcl.core':
      * ...
