@@ -292,18 +292,13 @@ deduce_thread(uword_t pointer, char** pc)
 }
 #endif
 
-static int non_nil_symbolp(lispobj x) {
-    return lowtag_of(x) == OTHER_POINTER_LOWTAG
-      && widetag_of((lispobj*)(x-OTHER_POINTER_LOWTAG)) == SYMBOL_WIDETAG;
-}
-
 static __attribute__((unused)) int tls_index_ok(lispobj tlsindex, struct vector* ignored_objects)
 {
     if (ignored_objects) {
         int i;
         for (i = vector_len(ignored_objects)-1; i >= 0; --i) {
             lispobj x = ignored_objects->data[i];
-            if (non_nil_symbolp(x) && tls_index_of(SYMBOL(x)) == tlsindex) return 0;
+            if (non_nil_symbol_p(x) && tls_index_of(SYMBOL(x)) == tlsindex) return 0;
         }
     }
     return 1; // is OK
@@ -482,7 +477,7 @@ static void maybe_show_object_name(lispobj obj, FILE* stream)
 
 static bool root_p(lispobj ptr, int criterion)
 {
-    if (ptr <= STATIC_SPACE_END) return 1; // always a root
+    if (ptr >= STATIC_SPACE_START && ptr < STATIC_SPACE_END) return 1; // always a root
     // 0 to 2 are in order of weakest to strongest condition for stopping,
     // i.e. criterion 0 implies that that largest number of objects
     // are considered roots.
@@ -873,7 +868,7 @@ static uword_t build_refs(lispobj* where, lispobj* end,
 static void scan_spaces(struct scan_state* ss)
 {
     struct scan_state old = *ss;
-    build_refs((lispobj*)NIL_SYMBOL_SLOTS_START, (lispobj*)NIL_SYMBOL_SLOTS_END, ss);
+    build_refs(NIL_SYMBOL_SLOTS_START, NIL_SYMBOL_SLOTS_END, ss);
     build_refs((lispobj*)STATIC_SPACE_OBJECTS_START, static_space_free_pointer, ss);
     show_tally(old, ss, "static");
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
