@@ -22,16 +22,24 @@
 
 /* constants derived from the fundamental constants in passed by GENESIS */
 #define READ_ONLY_SPACE_SIZE (READ_ONLY_SPACE_END - READ_ONLY_SPACE_START)
-#define STATIC_SPACE_SIZE (STATIC_SPACE_END - STATIC_SPACE_START)
 #define NIL_SYMBOL_SLOTS_START (lispobj*)(STATIC_SPACE_START + NIL_SYMBOL_SLOTS_OFFSET)
 #define NIL_SYMBOL_SLOTS_END (ALIGN_UP(SYMBOL_SIZE,2)+NIL_SYMBOL_SLOTS_START)
 #define STATIC_SPACE_OBJECTS_START (STATIC_SPACE_START + STATIC_SPACE_OBJECTS_OFFSET)
+#define STATIC_SPACE_END (STATIC_SPACE_START + STATIC_SPACE_SIZE)
+
+#ifdef LISP_FEATURE_X86_64
+#define T_SYMBOL_SLOTS_START (NIL_SYMBOL_SLOTS_START-7)
+#define T_SYMBOL_SLOTS_END (T_SYMBOL_SLOTS_START+SYMBOL_SIZE)
+#endif
 
 #ifdef LISP_FEATURE_DARWIN_JIT
 #define STATIC_CODE_SPACE_SIZE (STATIC_CODE_SPACE_END - STATIC_CODE_SPACE_START)
 #endif
 
 #define ALIEN_LINKAGE_SPACE_END (ALIEN_LINKAGE_SPACE_START + ALIEN_LINKAGE_SPACE_SIZE)
+#ifdef LISP_FEATURE_LINKAGE_SPACE
+#define LISP_LINKAGE_SPACE_SIZE (1<<(N_LINKAGE_INDEX_BITS+WORD_SHIFT))
+#endif
 
 #if !defined(__ASSEMBLER__)
 #include <stdbool.h>
