@@ -41,7 +41,6 @@
 #include "lispregs.h"
 #include "runtime.h"
 #include "genesis/primitive-objects.h"
-#include "dynbind.h"
 
 #include <sys/types.h>
 #include <sys/time.h>
@@ -57,13 +56,13 @@
 #include "validate.h"
 #include "thread.h"
 #include "align.h"
-#include "unaligned.h"
 
 #include "gc.h"
 #include <wincrypt.h>
 #include <stdarg.h>
 #include <string.h>
 #include "print.h"
+#include <shlobj.h>
 
 /* missing definitions for modern mingws */
 #ifndef EH_UNWINDING
@@ -114,8 +113,10 @@ static void set_seh_frame(void *frame)
 
 void alloc_gc_page()
 {
+#ifndef LISP_FEATURE_64_BIT // 64-bit uses the page below the card mark table
     gc_assert(VirtualAlloc(GC_SAFEPOINT_PAGE_ADDR, BACKEND_PAGE_BYTES,
                            MEM_RESERVE|MEM_COMMIT, PAGE_READWRITE));
+#endif
 }
 
 /* Permit loads from GC_SAFEPOINT_PAGE_ADDR (NB page state change is
@@ -153,168 +154,6 @@ void unmap_gc_page()
                              PAGE_NOACCESS, &oldProt));
 }
 
-/* This feature has already saved me more development time than it
- * took to implement.  In its current state, ``dynamic RT<->core
- * linking'' is a protocol of initialization of C runtime and Lisp
- * core, populating SBCL linkage table with entries for runtime
- * "foreign" symbols that were referenced in cross-compiled code.
- *
- * How it works: a sketch
- *
- * Last Genesis (resulting in cold-sbcl.core) binds foreign fixups in
- * x-compiled lisp-objs to sequential addresses from the beginning of
- * linkage-table space; that's how it ``resolves'' foreign references.
- * Obviously, this process doesn't require pre-built runtime presence.
- *
- * When the runtime loads the core (cold-sbcl.core initially,
- * sbcl.core later), runtime should do its part of the protocol by (1)
- * traversing a list of ``runtime symbols'' prepared by Genesis and
- * dumped as a static symbol value, (2) resolving each name from this
- * list to an address (stubbing unresolved ones with
- * undefined_alien_address or undefined_alien_function), (3) adding an
- * entry for each symbol somewhere near the beginning of linkage table
- * space (location is provided by the core).
- *
- * The implementation of the part described in the last paragraph
- * follows. C side is currently more ``hackish'' and less clear than
- * the Lisp code; OTOH, related Lisp changes are scattered, and some
- * of them play part in complex interrelations -- beautiful but taking
- * much time to understand --- but my subset of PE-i386 parser below
- * is in one place (here) and doesn't have _any_ non-trivial coupling
- * with the rest of the Runtime.
- *
- * What do we gain with this feature, after all?
- *
- * One things that I have to do rather frequently: recompile and
- * replace runtime without rebuilding the core. Doubtlessly, slam.sh
- * was a great time-saver here, but relinking ``cold'' core and bake a
- * ``warm'' one takes, as it seems, more than 10x times of bare
- * SBCL.EXE build time -- even if everything is recompiled, which is
- * now unnecessary. Today, if I have a new idea for the runtime,
- * getting from C-x C-s M-x ``compile'' to fully loaded SBCL
- * installation takes 5-15 seconds.
- *
- * Another thing (that I'm not currently using, but obviously
- * possible) is delivering software patches to remote system on
- * customer site. As you are doing minor additions or corrections in
- * Lisp code, it doesn't take much effort to prepare a tiny ``FASL
- * bundle'' that rolls up your patch, redumps and -- presto -- 100MiB
- * program is fixed by sending and loading a 50KiB thingie.
- *
- * However, until LISP_FEATURE_ALIEN_LINKAGE_TABLE, if your bug were fixed
- * by modifying two lines of _C_ sources, a customer described above
- * had to be ready to receive and reinstall a new 100MiB
- * executable. With the aid of code below, deploying such a fix
- * requires only sending ~300KiB (when stripped) of SBCL.EXE.
- *
- * But there is more to it: as the common linkage-table is used for
- * DLLs and core, its entries may be overridden almost without a look
- * into SBCL internals. Therefore, ``patching'' C runtime _without_
- * restarting target systems is also possible in many situations
- * (it's not as trivial as loading FASLs into a running daemon, but
- * easy enough to be a viable alternative if any downtime is highly
- * undesirable).
- *
- * During my (rather limited) commercial Lisp development experience
- * I've already been through a couple of situations where such
- * ``deployment'' issues were important; from my _total_ programming
- * experience I know -- _sometimes_ they are a two orders of magnitude
- * more important than those I observed.
- *
- * The possibility of entire runtime ``hot-swapping'' in running
- * process is not purely theoretical, as it could seem. There are 2-3
- * problems whose solution is not obvious (call stack patching, for
- * instance), but it's literally _nothing_ if compared with
- * e.g. LISP_FEATURE_SB_AUTO_FPU_SWITCH.  By the way, one of the
- * problems with ``hot-swapping'', that could become a major one in
- * many other environments, is nonexistent in SBCL: we already have a
- * ``global quiesce point'' that is generally required for this kind
- * of worldwide revolution -- around collect_garbage.
- *
- * If we look at the majority of the ``new style'' code units, it's a
- * common thing to observe how #+-ifdeffery _vanishes_ instead of
- * multiplying: #-sb-xc, #+sb-xc-host and #-sb-xc-host end up
- * needing the same code. Runtime checks of static v. dynamic symbol
- * disappear even faster. STDCALL mangling and leading underscores go
- * out of scope (and GCed, hopefully) instead of surfacing here and
- * there as a ``special case for core static symbols''. What I like
- * the most about CL development in general is a frequency of solving
- * problems and fixing bugs by simplifying code and dropping special
- * cases.
- *
- * Last important thing about the following code: besides resolving
- * symbols provided by the core itself, it detects runtime's own
- * build-time prerequisite DLLs. Any symbol that is unresolved against
- * the core is looked up in those DLLs (normally kernel32, msvcrt,
- * ws2_32... I could forget something). This action (1) resembles
- * implementation of foreign symbol lookup in SBCL itself, (2)
- * emulates shared library d.l. facilities of OSes that use flat
- * dynamic symbol namespace (or default to it). Anyone concerned with
- * portability problems of this PE-i386 stuff below will be glad to
- * hear that it could be ported to most modern Unices _by deletion_:
- * raw dlsym() with null handle usually does the same thing that i'm
- * trying to squeeze out of MS Windows by the brute force.
- *
- * My reason for _desiring_ flat symbol namespace, populated from
- * link-time dependencies, is avoiding any kind of ``requested-by-Lisp
- * symbol lists to be linked statically'', providing core v. runtime
- * independence in both directions. Minimizing future maintenance
- * effort is very important; I had gone for it consistently, starting
- * by turning "CloseHandle@4" into a simple "CloseHandle", continuing
- * by adding intermediate Genesis resulting in autogenerated symbol
- * list (farewell, void scratch(); good riddance), going to take
- * another great step for core/runtime independence... and _without_
- * flat namespace emulation, the ghosts and spirits exiled at the
- * first steps would come and take revenge: well, here are the symbols
- * that are really in msvcrt.dll.. hmm, let's link statically against
- * them, so the entry is pulled from the import library.. and those
- * entry has mangled names that we have to map.. ENOUGH, I though
- * here: fed up with stuff like that.
- *
- * Now here we are, without import libraries, without mangled symbols,
- * and without nm-generated symbol tables. Every symbol exported by
- * the runtime is added to SBCL.EXE export directory; every symbol
- * requested by the core is looked up by GetProcAddress for SBCL.EXE,
- * falling back to GetProcAddress for MSVCRT.dll, etc etc.. All ties
- * between SBCL's foreign symbols with object file symbol tables,
- * import libraries and other pre-linking symbol-resolving entities
- * _having no representation in SBCL.EXE_ were teared.
- *
- * This simplistic approach proved to work well; there is only one
- * problem introduced by it, and rather minor: in real MSVCRT.dll,
- * what's used to be available as open() is now called _open();
- * similar thing happened to many other `lowio' functions, though not
- * every one, so it's not a kind of name mangling but rather someone's
- * evil creative mind in action.
- *
- * When we look up any of those poor `uglified' functions in CRT
- * reference on MSDN, we can see a notice resembling this one:
- *
- * `unixishname()' is obsolete and provided for backward
- * compatibility; new standard-compliant function, `_unixishname()',
- * should be used instead.  Sentences of that kind were there for
- * several years, probably even for a decade or more (a propos,
- * MSVCRT.dll, as the name to link against, predates year 2000, so
- * it's actually possible). Reasoning behing it (what MS people had in
- * mind) always seemed strange to me: if everyone uses open() and that
- * `everyone' is important to you, why rename the function?  If no one
- * uses open(), why provide or retain _open() at all? <kidding>After
- * all, names like _open() are entirely non-informative and just plain
- * ugly; compare that with CreateFileW() or InitCommonControlsEx(),
- * the real examples of beauty and clarity.</kidding>
- *
- * Anyway, if the /standard/ name on Windows is _open() (I start to
- * recall, vaguely, that it's because of _underscore names being
- * `reserved to system' and all other ones `available for user', per
- * ANSI/ISO C89) -- well, if the /standard/ name is _open, SBCL should
- * use it when it uses MSVCRT and not some ``backward-compatible''
- * stuff. Deciding this way, I added a hack to SBCL's syscall macros,
- * so "[_]open" as a syscall name is interpreted as a request to link
- * agains "_open" on win32 and "open" on every other system.
- *
- * Of course, this name-parsing trick lacks conceptual clarity; we're
- * going to get rid of it eventually. */
-
 uint32_t os_get_build_time_shared_libraries(uint32_t excl_maximum,
                                        void* opt_root,
                                        void** opt_store_handles,
@@ -426,17 +265,44 @@ void* os_dlsym_default(char* name)
 {
     unsigned int i;
     void* result = 0;
+
     buildTimeImages[0] = (void*)runtime_module_handle;
     if (buildTimeImageCount == 0) {
         buildTimeImageCount =
             1 + os_get_build_time_shared_libraries(15u,
-            NULL, 1+(void**)buildTimeImages, NULL);
+                                                   NULL, 1+(void**)buildTimeImages, NULL);
     }
     for (i = 0; i<buildTimeImageCount && (!result); ++i) {
         result = GetProcAddress(buildTimeImages[i], name);
     }
     return result;
 }
+BOOL K32EnumProcessModules(HANDLE  hProcess, HMODULE *lphModule, DWORD cb, LPDWORD lpcbNeeded);
+
+void* sb_dlsym(char* name)
+{
+    unsigned int i;
+    void* result = 0;
+
+    if ((result = GetProcAddress(GetModuleHandle(NULL), name)))
+        return result;
+
+    HANDLE process = GetCurrentProcess();
+    HMODULE modules[1024];
+    DWORD needed;
+    if (K32EnumProcessModules(process, modules, sizeof(modules), &needed)) {
+        DWORD fetched = (needed / sizeof(HMODULE));
+
+        if (fetched > sizeof(modules))
+            fetched = sizeof(modules);
+
+        for (i = 0; i < fetched; i++) {
+            if ((result = GetProcAddress(modules[i], name)))
+                return result;
+        }
+    }
+    return result;
+}
 
 /* We want to get a slot in TIB that (1) is available at constant
    offset, (2) is our private property, so libraries wouldn't legally
@@ -838,12 +704,10 @@ os_alloc_gc_space(int space_id, int attributes, os_vm_address_t addr, os_vm_size
 
 void os_commit_memory(os_vm_address_t addr, os_vm_size_t len)
 {
-    if (len) gc_assert(VirtualAlloc(addr, len, MEM_COMMIT, PAGE_EXECUTE_READWRITE));
-}
-
-void os_decommit_mem(os_vm_address_t addr,  os_vm_size_t len) {
-    gc_assert(gc_active_p);
-    gc_assert(VirtualFree(addr, len, MEM_DECOMMIT));
+    if (len) {
+        gc_assert(addr);
+        gc_assert(VirtualAlloc(addr, len, MEM_COMMIT, PAGE_EXECUTE_READWRITE));
+    }
 }
 
 /*
@@ -862,28 +726,38 @@ void os_decommit_mem(os_vm_address_t addr,  os_vm_size_t len) {
 void* load_core_bytes(int fd, os_vm_offset_t offset, os_vm_address_t addr, os_vm_size_t len,
                       int is_readonly_space)
 {
-    os_commit_memory(addr, len);
+    if (addr) {
+        os_commit_memory(addr, len);
+    } else {
+        addr = VirtualAlloc(NULL, len, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
+        gc_assert(addr);
+    }
 #ifdef LISP_FEATURE_64_BIT
     os_vm_offset_t res = _lseeki64(fd, offset, SEEK_SET);
 #else
     os_vm_offset_t res = lseek(fd, offset, SEEK_SET);
 #endif
     gc_assert(res == offset);
-    size_t count;
+    int count;
 
     os_vm_address_t original_addr = addr;
     os_vm_size_t original_len = len;
     while (len) {
         unsigned to_read = len > INT_MAX ? INT_MAX : len;
         count = read(fd, addr, to_read);
+        if (count == -1) {
+            perror("read() failed"); fflush(stderr);
+        }
         addr += count;
         len -= count;
-        gc_assert(count == to_read);
+        gc_assert(count == (int) to_read);
     }
     DWORD old;
     if (is_readonly_space) VirtualProtect(original_addr, original_len, PAGE_READONLY, &old);
-    return (void*)0;
+
+    return original_addr;
 }
+
 static DWORD os_protect_modes[8] = {
     PAGE_NOACCESS,
     PAGE_READONLY,
@@ -977,21 +851,7 @@ handle_breakpoint_trap(os_context_t *ctx, struct thread* self)
 
     WITH_GC_AT_SAFEPOINTS_ONLY() {
         block_blockable_signals(&ctx->sigmask);
-#ifdef LISP_FEATURE_IMMOBILE_SPACE
-        if (trap == trap_UndefinedFunction) {
-            lispobj* fdefn = (lispobj*)(OS_CONTEXT_PC(ctx) & ~LOWTAG_MASK);
-            if (fdefn && widetag_of(fdefn) == FDEFN_WIDETAG) {
-                // Return to undefined-tramp
-                OS_CONTEXT_PC(ctx) = (uword_t)((struct fdefn*)fdefn)->raw_addr;
-                // with RAX containing the FDEFN
-                *os_context_register_addr(ctx,reg_RAX) =
-                    make_lispobj(fdefn, OTHER_POINTER_LOWTAG);
-            }
-        } else
-#endif
-        {
-            handle_trap(ctx, trap);
-        }
+        handle_trap(ctx, trap);
         thread_sigmask(SIG_SETMASK,&ctx->sigmask,NULL);
     }
 
@@ -1045,7 +905,7 @@ handle_access_violation(os_context_t *ctx,
     /* dynamic space */
     page_index_t page = find_page_index(fault_address);
 #ifdef LISP_FEATURE_SOFT_CARD_MARKS
-    if (page >= 0) lose("should not get access violation in dynamic space");
+    if (page >= 0) lose("should not get access violation in dynamic space %p", fault_address);
 #else
     if (page != -1 && !PAGE_WRITEPROTECTED_P(page)) {
         os_commit_memory(PTR_ALIGN_DOWN(fault_address, os_vm_page_size),
@@ -1070,6 +930,8 @@ handle_access_violation(os_context_t *ctx,
     return -1;
 }
 
+void lisp_memory_fault_warning(os_context_t *context, os_vm_address_t addr);
+
 static void
 signal_internal_error_or_lose(os_context_t *ctx,
                               EXCEPTION_RECORD *exception_record,
@@ -1081,8 +943,14 @@ signal_internal_error_or_lose(os_context_t *ctx,
      * set up, or drop to LDB.
      */
 
-    if (internal_errors_enabled) {
+    if ((long int)exception_record->ExceptionCode == EXCEPTION_ACCESS_VIOLATION) {
+        lisp_memory_fault_warning(ctx, fault_address);
+    }
 
+    if (internal_errors_enabled) {
+        /* The exception system doesn't automatically clear pending
+         * exceptions, so we lose as soon as we execute any FP
+         * instruction unless we do this first. */
         asm("fnclex");
         /* We're making the somewhat arbitrary decision that having
          * internal errors enabled means that lisp has sufficient
@@ -1098,9 +966,10 @@ signal_internal_error_or_lose(os_context_t *ctx,
             DX_ALLOC_SAP(exception_record_sap, exception_record);
             thread_sigmask(SIG_SETMASK, &ctx->sigmask, NULL);
 
-            /* The exception system doesn't automatically clear pending
-             * exceptions, so we lose as soon as we execute any FP
-             * instruction unless we do this first. */
+#ifdef LISP_FEATURE_X86_64
+            asm("fninit");
+#endif
+
             /* Call into lisp to handle things. */
             funcall2(StaticSymbolFunction(HANDLE_WIN32_EXCEPTION),
                      context_sap,
@@ -1117,17 +986,7 @@ signal_internal_error_or_lose(os_context_t *ctx,
             (void*)(intptr_t)exception_record->ExceptionCode);
     fprintf(stderr, "Faulting IP: %p.\n",
             (void*)(intptr_t)exception_record->ExceptionAddress);
-    if ((long int)exception_record->ExceptionCode == EXCEPTION_ACCESS_VIOLATION) {
-        MEMORY_BASIC_INFORMATION mem_info;
 
-        if (VirtualQuery(fault_address, &mem_info, sizeof mem_info)) {
-            fprintf(stderr, "page status: 0x%lx.\n", mem_info.State);
-        }
-
-        fprintf(stderr, "Was writing: %p, where: %p.\n",
-                (void*)exception_record->ExceptionInformation[0],
-                fault_address);
-    }
 
     fflush(stderr);
 
@@ -1737,9 +1596,10 @@ win32_write_console(HANDLE handle, void * buf, int count)
     }
 }
 
-int
-win32_unix_write(HANDLE handle, void * buf, int count)
+ssize_t
+win32_unix_write(HANDLE handle, void * buf, size_t requested_count)
 {
+    int count = requested_count > INT_MAX ? INT_MAX : requested_count;
     DWORD written_bytes;
     OVERLAPPED overlapped;
     struct thread * self = get_sb_vm_thread();
@@ -1812,9 +1672,10 @@ win32_unix_write(HANDLE handle, void * buf, int count)
 }
 
 
-int
-win32_unix_read(HANDLE handle, void * buf, int count)
+ssize_t
+win32_unix_read(HANDLE handle, void * buf, size_t requested_count)
 {
+    int count = requested_count > INT_MAX ? INT_MAX : requested_count;
     OVERLAPPED overlapped = {.Internal=0};
     DWORD read_bytes = 0;
     struct thread * self = get_sb_vm_thread();
