@@ -13,7 +13,7 @@
 
 #include <stdio.h>
 
-#include "sbcl.h"
+#include "genesis/sbcl.h"
 #include "runtime.h"
 #include "globals.h"
 #include "validate.h"
@@ -21,21 +21,19 @@
 #include "arch.h"
 #include "lispregs.h"
 #include <signal.h>
-#include "alloc.h"
 #include "interrupt.h"
 #include "interr.h"
 #include "breakpoint.h"
 #include "thread.h"
 #include "pseudo-atomic.h"
-#include "unaligned.h"
+#include "align.h"
 #include "search.h"
 #include "var-io.h"
 
 #include "genesis/static-symbols.h"
 #include "genesis/symbol.h"
-#include "forwarding-ptr.h"
 #include "core.h"
-#include "gc-private.h"
+#include "gc.h"
 
 #define INT3_INST 0xCC
 #define INTO_INST 0xCE
@@ -72,7 +70,20 @@ static void xgetbv(unsigned *eax, unsigned *edx)
 }
 
 #define VECTOR_FILL_T "VECTOR-FILL/T"
+#ifdef LISP_FEATURE_SB_SAFEPOINT
+// the store to card table takes 3 bytes more encode
+static const int vector_fill_offset_to_check = 0x53;
+static const int vector_fill_offset_to_poke  = 0x5A;
+#else
+static const int vector_fill_offset_to_check = 0x50;
+static const int vector_fill_offset_to_poke  = 0x57;
+#endif
+static const unsigned char vector_fill_expect_bytes[] = {
+  0x48, 0x81, 0xF9, 0xBC, 0x02, 0x00, 0x00,
+  0xEB, 0x07
+};
 
+#define FEATURE_BITS_EA (lispobj*)(NIL-LIST_POINTER_LOWTAG-16)
 // Poke in a byte that changes an opcode to enable faster vector fill.
 // Using fixed offsets and bytes is no worse than what we do elsewhere.
 void tune_asm_routines_for_microarch(void)
@@ -101,7 +112,17 @@ void tune_asm_routines_for_microarch(void)
     if (avx2_supported) our_cpu_feature_bits |= 1;
     // POPCNT = ECX bit 23, which gets copied into bit 2 in *CPU-FEATURE-BITS*
     if (cpuid_fn1_ecx & (1<<23)) our_cpu_feature_bits |= 2;
-    SetSymbolValue(CPU_FEATURE_BITS, make_fixnum(our_cpu_feature_bits), 0);
+    *FEATURE_BITS_EA = make_fixnum(our_cpu_feature_bits);
+
+    unsigned char* asm_routine = (void*)get_asm_routine_by_name(VECTOR_FILL_T, 0);
+    if (!asm_routine) return;
+    // Since a particular runtime expects a particular core,
+    // mismatch of the ASM routine is a fatal error.
+    if (memcmp(asm_routine + vector_fill_offset_to_check,
+               vector_fill_expect_bytes,
+               sizeof vector_fill_expect_bytes))
+        lose("%s does not match expectation @ %p",
+             VECTOR_FILL_T, asm_routine + vector_fill_offset_to_check);
 
     // I don't know if this works on Windows
 #ifndef _MSC_VER
@@ -109,7 +130,8 @@ void tune_asm_routines_for_microarch(void)
     if (eax >= 7) {
         cpuid(7, 0, &eax, &ebx, &ecx, &edx);
         if (ebx & (1<<9)) // Enhanced Repeat Movs/Stos
-          asm_routine_poke(VECTOR_FILL_T, 0x12, 0x7C); // Change JMP to JL
+          asm_routine_poke(VECTOR_FILL_T, vector_fill_offset_to_poke,
+                           0x7C); // Change JMP to JL
     }
 #endif
 }
@@ -120,8 +142,11 @@ void tune_asm_routines_for_microarch(void)
    instructions that don't exist on some cpu family members */
 void untune_asm_routines_for_microarch(void)
 {
-    asm_routine_poke(VECTOR_FILL_T, 0x12, 0xEB); // Change JL to JMP
-    SetSymbolValue(CPU_FEATURE_BITS, 0, 0);
+    asm_routine_poke(VECTOR_FILL_T, vector_fill_offset_to_poke,
+                     0xEB); // Change JL to JMP
+    *FEATURE_BITS_EA = 0;
+    // ensure no random value lingering in static space on image save
+    SYMBOL(CALLBACK_WRAPPER_TRAMPOLINE)->value = 0;
 }
 
 #ifndef _WIN64
@@ -145,53 +170,55 @@ arch_get_bad_addr(int __attribute__((unused)) sig,
 // I don't have an easy way to test changes for these OSes, so just use this
 // context visitor based on the old slightly-less-efficient way of doing it.
 #if defined LISP_FEATURE_SUNOS || defined LISP_FEATURE_HAIKU
-void visit_context_registers(void (*proc)(os_context_register_t, int), os_context_t *context)
+void visit_context_registers(void (*proc)(os_context_register_t, void*),
+                             os_context_t *context, void* arg)
 {
-    proc(os_context_pc(context), 1);
-    proc(*os_context_register_addr(context, reg_RAX), 1);
-    proc(*os_context_register_addr(context, reg_RCX), 1);
-    proc(*os_context_register_addr(context, reg_RDX), 1);
-    proc(*os_context_register_addr(context, reg_RBX), 1);
-    proc(*os_context_register_addr(context, reg_RSI), 1);
-    proc(*os_context_register_addr(context, reg_RDI), 1);
-    proc(*os_context_register_addr(context, reg_R8 ), 1);
-    proc(*os_context_register_addr(context, reg_R9 ), 1);
-    proc(*os_context_register_addr(context, reg_R10), 1);
-    proc(*os_context_register_addr(context, reg_R11), 1);
-    proc(*os_context_register_addr(context, reg_R12), 1);
-    proc(*os_context_register_addr(context, reg_R13), 1);
-    proc(*os_context_register_addr(context, reg_R14), 1);
-    proc(*os_context_register_addr(context, reg_R15), 1);
+    proc(os_context_pc(context), arg);
+    proc(*os_context_register_addr(context, reg_RAX), arg);
+    proc(*os_context_register_addr(context, reg_RCX), arg);
+    proc(*os_context_register_addr(context, reg_RDX), arg);
+    proc(*os_context_register_addr(context, reg_RBX), arg);
+    proc(*os_context_register_addr(context, reg_RSI), arg);
+    proc(*os_context_register_addr(context, reg_RDI), arg);
+    proc(*os_context_register_addr(context, reg_R8 ), arg);
+    proc(*os_context_register_addr(context, reg_R9 ), arg);
+    proc(*os_context_register_addr(context, reg_R10), arg);
+    proc(*os_context_register_addr(context, reg_R11), arg);
+    proc(*os_context_register_addr(context, reg_R12), arg);
+    proc(*os_context_register_addr(context, reg_R13), arg);
+    proc(*os_context_register_addr(context, reg_R14), arg);
+    proc(*os_context_register_addr(context, reg_R15), arg);
 }
 #endif
 
-os_context_register_t *
-os_context_flags_addr(os_context_t *context)
-{
 #if defined __linux__
     /* KLUDGE: As of kernel 2.2.14 on Red Hat 6.2, there's code in the
      * <sys/ucontext.h> file to define symbolic names for offsets into
      * gregs[], but it's conditional on __USE_GNU and not defined, so
      * we need to do this nasty absolute index magic number thing
      * instead. */
-    return (os_context_register_t*)&context->uc_mcontext.gregs[17];
+#   define CONTEXT_FLAGS(c) c->uc_mcontext.gregs[17]
 #elif defined LISP_FEATURE_SUNOS
-    return &context->uc_mcontext.gregs[REG_RFL];
+#   define CONTEXT_FLAGS(c) c->uc_mcontext.gregs[REG_RFL]
 #elif defined LISP_FEATURE_FREEBSD || defined(__DragonFly__)
-    return &context->uc_mcontext.mc_rflags;
+#   define CONTEXT_FLAGS(c) c->uc_mcontext.mc_rflags
 #elif defined __HAIKU__
-    return &context->uc_mcontext.rflags;
+#   define CONTEXT_FLAGS(c) c->uc_mcontext.rflags
 #elif defined LISP_FEATURE_DARWIN
-    return CONTEXT_ADDR_FROM_STEM(rflags);
+#   define CONTEXT_FLAGS(c) CONTEXT_SLOT(c,rflags)
 #elif defined __OpenBSD__
-    return &context->sc_rflags;
+#   define CONTEXT_FLAGS(c) c->sc_rflags
 #elif defined __NetBSD__
-    return CONTEXT_ADDR_FROM_STEM(RFLAGS);
+#   define CONTEXT_FLAGS(c) CONTEXT_SLOT(c,RFLAGS)
 #elif defined _WIN64
-    return (os_context_register_t*)&context->win32_context->EFlags;
+#   define CONTEXT_FLAGS(c) c->win32_context->EFlags
 #else
 #error unsupported OS
 #endif
+
+os_context_register_t *os_context_flags_addr(os_context_t *context)
+{
+    return (os_context_register_t*)&(CONTEXT_FLAGS(context));
 }
 
 void arch_skip_instruction(os_context_t *context)
@@ -237,8 +264,6 @@ void arch_skip_instruction(os_context_t *context)
             break;
         }
 
-    FSHOW((stderr,
-           "/[arch_skip_inst resuming at %x]\n", OS_CONTEXT_PC(context)));
 }
 
 unsigned char *
@@ -247,7 +272,7 @@ arch_internal_error_arguments(os_context_t *context)
     return 1 + (unsigned char *)OS_CONTEXT_PC(context);
 }
 
-boolean arch_pseudo_atomic_atomic(struct thread *thread) {
+bool arch_pseudo_atomic_atomic(struct thread *thread) {
     return get_pseudo_atomic_atomic(thread);
 }
 
@@ -266,7 +291,7 @@ void arch_clear_pseudo_atomic_interrupted(struct thread *thread) {
 unsigned int
 arch_install_breakpoint(void *pc)
 {
-    unsigned int result = *(unsigned int*)pc;
+    unsigned int result = UNALIGNED_LOAD32(pc);
 #ifdef LISP_FEATURE_INT4_BREAKPOINTS
     *(char*)pc = INTO_INST;
 #else
@@ -310,7 +335,7 @@ arch_do_displaced_inst(os_context_t *context, unsigned int orig_inst)
     *(pc-2) = 0x00240c81;
     *(pc-1) = 0x9d000001;
 #else
-    *os_context_flags_addr(context) |= 0x100;
+    CONTEXT_FLAGS(context) |= 0x100;
 #endif
 
     single_stepping = pc;
@@ -353,7 +378,7 @@ restore_breakpoint_from_single_step(os_context_t * context)
     *(single_stepping-2) = single_step_save2;
     *(single_stepping-1) = single_step_save3;
 #else
-    *os_context_flags_addr(context) &= ~0x100;
+    CONTEXT_FLAGS(context) &= ~0x100;
 #endif
     /* Re-install the breakpoint if possible. */
     if (((char *)OS_CONTEXT_PC(context) > (char *)single_stepping) &&
@@ -395,22 +420,6 @@ sigtrap_handler(int __attribute__((unused)) signal,
      * number of bytes will follow, the first is the length of the byte
      * arguments to follow. */
     trap = *(unsigned char *)OS_CONTEXT_PC(context);
-#ifdef LISP_FEATURE_IMMOBILE_SPACE
-    if (trap == trap_UndefinedFunction) {
-        // The interrupted PC pins this fdefn. Sigtrap is delivered on the ordinary stack,
-        // not the alternate stack.
-        // (FIXME: an interior pointer to an fdefn _should_ pin it, but doesn't)
-        lispobj* fdefn = (lispobj*)(OS_CONTEXT_PC(context) & ~LOWTAG_MASK);
-        if (fdefn && widetag_of(fdefn) == FDEFN_WIDETAG) {
-            // Return to undefined-tramp
-            OS_CONTEXT_PC(context) = (uword_t)((struct fdefn*)fdefn)->raw_addr;
-            // with RAX containing the FDEFN
-            *os_context_register_addr(context,reg_RAX) =
-                make_lispobj(fdefn, OTHER_POINTER_LOWTAG);
-            return;
-        }
-    }
-#endif
     handle_trap(context, trap);
 }
 
@@ -419,7 +428,7 @@ sigill_handler(int __attribute__((unused)) signal,
                siginfo_t __attribute__((unused)) *siginfo,
                os_context_t *context) {
     unsigned char* pc = (void*)OS_CONTEXT_PC(context);
-    if (*(unsigned short *)pc == UD2_INST) {
+    if (UNALIGNED_LOAD16(pc) == UD2_INST) {
         OS_CONTEXT_PC(context) += 2;
         return sigtrap_handler(signal, siginfo, context);
     }
@@ -516,7 +525,7 @@ arch_install_interrupt_handlers()
 void
 arch_write_linkage_table_entry(int index, void *target_addr, int datap)
 {
-    char *reloc_addr = (char*)ALIEN_LINKAGE_TABLE_SPACE_START + index * ALIEN_LINKAGE_TABLE_ENTRY_SIZE;
+    char *reloc_addr = (char*)ALIEN_LINKAGE_SPACE_START + index * ALIEN_LINKAGE_TABLE_ENTRY_SIZE;
     if (datap) {
         *(uword_t *)reloc_addr = (uword_t)target_addr;
         return;
@@ -582,69 +591,23 @@ arch_set_fp_modes(unsigned int mxcsr)
     asm ("ldmxcsr %0" : : "m" (temp));
 }
 
-static __attribute__((unused)) boolean codeblob_p(lispobj ptr) {
-    return lowtag_of(ptr) == OTHER_POINTER_LOWTAG &&
-           widetag_of((lispobj*)(ptr-OTHER_POINTER_LOWTAG)) == CODE_HEADER_WIDETAG;
-}
-
 /* Return the tagged pointer for which 'entrypoint' is the starting address.
- * This will be one of the following:
- * 1. a fun-pointer which is either
- *   - a simple-fun header located 2 words before the entrypoint.
- *     N.B.: This returns the SIMPLE-FUN, and _not_ its containing codeblob.
- *   - a funcallable-instance with an embedded trampoline that make it
- *     equivalent to a simple-fun
- * 2. a code-component with no simple-fun within it, that makes closures
- *    callable like simple-funs. The code header is at entrypoint minus 4 words.
- *
- * By first reading at (entrypoint - 2*N_WORD_BYTES) this does the right thing
- * when 'entrypoint' is actually type 2, because the word accessed will not be
- * an object header or forwarding marker (it will be whatever %CODE-DEBUG-INFO is).
- * Whereis if this were type 1, but we read at (entrypoint - 4*N_WORD_BYTES) first,
- * then we could perceive random uninitialized bytes of the preceding object.
+ * This result will have SIMPLE_FUN_WIDETAG or FUNCALLABLE_INSTANCE_WIDETAG.
+ * If the thing has been forwarded, we do NOT return the newspace copy.
  */
 lispobj entrypoint_taggedptr(uword_t entrypoint) {
     if (!entrypoint || points_to_asm_code_p(entrypoint)) return 0;
-    // First try
     lispobj* phdr = (lispobj*)(entrypoint - 2*N_WORD_BYTES);
     if (forwarding_pointer_p(phdr)) {
-        gc_dcheck(lowtag_of(forwarding_pointer_value(phdr)) == FUN_POINTER_LOWTAG);
-        return make_lispobj(phdr, FUN_POINTER_LOWTAG);
+        gc_assert(lowtag_of(forwarding_pointer_value(phdr)) == FUN_POINTER_LOWTAG);
+        // We can't assert on the widetag if forwarded, because defragmentation
+        // puts the new logical object at some totally different physical address.
+        // This function doesn't know if defrag is occurring.
+    } else {
+        __attribute__((unused)) unsigned char widetag = widetag_of(phdr);
+        gc_assert(widetag == SIMPLE_FUN_WIDETAG || widetag == FUNCALLABLE_INSTANCE_WIDETAG);
     }
-    int widetag = widetag_of(phdr);
-    if (widetag == FUNCALLABLE_INSTANCE_WIDETAG || widetag == SIMPLE_FUN_WIDETAG) {
-        return make_lispobj(phdr, FUN_POINTER_LOWTAG);
-    }
-    // Second try
-    phdr = (lispobj*)(entrypoint - 4*N_WORD_BYTES);
-    /* It is nearly impossible for forwarding to arise, by this reasoning:
-     * Case A: if some thread references the codeblob that wraps a closure,
-     *  then the codeblob is pinned; hence not forwarded.
-     * Case B: if not executing (or otherwise referenced from stack/registers),
-     *  there can exist no other tagged reference to the codeblob.
-     * Aside from the FDEFN that owns it, the only other untagged reference would
-     * be from the search tree, which isn't scavenged. (The entire tree dies after GC.)
-     * It's conceivable the debugger could store a tagged pointer to this entrypoint
-     * in something, but I tried to make it do so, and couldn't. I was, however, able
-     * to artificially cause forwarding by putting closure trampolines in symbols */
-    if (forwarding_pointer_p(phdr))
-        gc_dcheck(codeblob_p(forwarding_pointer_value(phdr)));
-    else
-        gc_dcheck(widetag_of(phdr) == CODE_HEADER_WIDETAG);
-    return make_lispobj(phdr, OTHER_POINTER_LOWTAG);
-}
-/* Return the lisp object that fdefn's raw_addr slot jumps to.
- * In the event that the referenced object was forwarded, this returns the un-forwarded
- * object (the forwarded value is used to assert some invariants though).
- * If the fdefn jumps to the UNDEFINED-FDEFN routine, then return 0.
- *
- * Some legacy baggage is evident: in the first implementation of immobile fdefns,
- * an fdefn used a 'jmp rel32' (relative to itself), and so you could decode the
- * jump target only given the address of the fdefn. That is no longer true; fdefns use
- * absolute jumps. Therefore it is possible to call entrypoint_taggedptr() with any
- * raw_addr, whether or not you know the fdefn whence the raw_addr was obtained. */
-lispobj decode_fdefn_rawfun(struct fdefn* fdefn) {
-    return entrypoint_taggedptr((uword_t)fdefn->raw_addr);
+    return make_lispobj(phdr, FUN_POINTER_LOWTAG);
 }
 
 #ifdef LISP_FEATURE_SB_THREAD
@@ -663,7 +626,7 @@ extern pthread_mutex_t alloc_profiler_lock;
 
 static unsigned int claim_index(int qty) // qty is 1 or 2
 {
-    static boolean warning_issued;
+    static bool warning_issued;
     unsigned int index = fixnum_value(SYMBOL(N_PROFILE_SITES)->value);
     SYMBOL(N_PROFILE_SITES)->value += make_fixnum(qty);
     if (fixnum_value(SYMBOL(N_PROFILE_SITES)->value) <= max_alloc_point_counters)
@@ -675,7 +638,7 @@ static unsigned int claim_index(int qty) // qty is 1 or 2
     return 0; // use the overflow bin(s)
 }
 
-static boolean NO_SANITIZE_MEMORY
+static bool NO_SANITIZE_MEMORY
 instrumentp(uword_t* sp, uword_t** pc, uword_t* old_word)
 {
     int __attribute__((unused)) ret = mutex_acquire(&alloc_profiler_lock);
@@ -703,7 +666,7 @@ instrumentp(uword_t* sp, uword_t** pc, uword_t* old_word)
 
 // logical index 'index' in the metadata array stores the code component
 // and pc-offset relative to the component base address
-static void record_pc(char* pc, unsigned int index, boolean sizedp)
+static void record_pc(char* pc, unsigned int index, bool sizedp)
 {
     lispobj *code = component_ptr_from_pc(pc);
     if (!code) {
@@ -745,8 +708,7 @@ static void record_pc(char* pc, unsigned int index, boolean sizedp)
     }
 }
 
-void AMD64_SYSV_ABI
-allocation_tracker_counted(uword_t* sp)
+void allocation_tracker_counted(uword_t* sp)
 {
     uword_t *pc, word_at_pc;
     if (instrumentp(sp, &pc, &word_at_pc)) {
@@ -777,8 +739,7 @@ allocation_tracker_counted(uword_t* sp)
     gc_assert(ret);
 }
 
-void AMD64_SYSV_ABI
-allocation_tracker_sized(uword_t* sp)
+void allocation_tracker_sized(uword_t* sp)
 {
     uword_t *pc, word_at_pc;
     if (instrumentp(sp, &pc, &word_at_pc)) {
@@ -823,4 +784,77 @@ lispobj call_into_lisp_first_time(lispobj fun, lispobj *args, int nargs) {
     return call_into_lisp_first_time_(fun, args, nargs, get_sb_vm_thread());
 }
 
+/*
+ * On x86-64 we try to place the alien and lisp linkage tables in such a way
+ * that avoids extra load instructions when calling, but also allows those tables
+ * to be fully relocatable. It is best achieved by using PC-relative addressng,
+ * which works only for immobile text space. Failing that, we can place the
+ * linkage tables below NIL and use NIL-relative addressing.
+ * The core file makes no indication of the "effective size" of static space
+ * or text space, so we have to oversize them.
+ * It's a little confusing, so here are the possibilities:
+ *
+ * Supports           |   extra allocation amount
+ * Immobile | elfcode |      text         | static
+ * ---------|-----------------------------|-------------------------------------
+ *   Yes    |   No    | +AL +LL below     | none
+ *   Yes    |   Yes   |     n/a           | +AL below, +GC cards above
+ *   No     |   No    |     none          | +AL+LL below, +GC cards above
+ *   No     |   Yes   |     n/a           | +AL below, +GC cards above
+ *
+ * AL = alien linkage
+ * LL = lisp linkage
+ * n/a means the call does not occur for that space
+ *
+ * For #+immobile-space we want to end up with text space having both linkage subspaces
+ * (unless code-in-ELF)
+ *   | LISP LINKAGE | ALIEN LINKAGE | CODE OBJECTS ...
+ *   |<------------>|<------------->| ....
+ * For code-in-ELF then the lisp linkage space was preallocated to a .bss section,
+ * so we only oversize the static space by the alien linkage space size.
+ * If there is no text space (i.e. for #-immobile-space) then the linkage tables
+ * are below static space.
+ */
+os_vm_address_t coreparse_alloc_space(int space_id, int attr,
+                                      os_vm_address_t addr, os_vm_size_t size)
+{
+    if (size == 0) return addr;
+
+    int extra_below = 0, extra_above = 0;
+    extern int lisp_code_in_elf();
+
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+# define LINKAGE_TABLE_CONTAINER IMMOBILE_TEXT_CORE_SPACE_ID
+#else
+# define LINKAGE_TABLE_CONTAINER STATIC_CORE_SPACE_ID
+#endif
+    if (!lisp_code_in_elf()) { // a normal core
+        if (space_id == LINKAGE_TABLE_CONTAINER)
+            extra_below = LISP_LINKAGE_SPACE_SIZE + ALIEN_LINKAGE_SPACE_SIZE;
+    } else { // code-in-ELF core
+        if (space_id == STATIC_CORE_SPACE_ID)
+            extra_below = ALIEN_LINKAGE_SPACE_SIZE;
+    }
+    if (space_id == STATIC_CORE_SPACE_ID) {
+        extra_above =
+# ifdef LISP_FEATURE_SB_SAFEPOINT // should just add 1 OS page but instead
+            BACKEND_PAGE_BYTES +  // it's a ridiculously generous bump up
+# endif
+                ALIGN_UP((1+gc_card_table_mask), os_reported_page_size);
+    }
+
+    addr -= extra_below; // endeavor to return the requested address as it was
+    size += extra_below + extra_above;
+    //fprintf(stderr, "requesting space for space_id %d, below=%x above=%x\n", space_id, extra_below, extra_above);
+    addr = os_alloc_gc_space(space_id, attr, addr, size);
+    if (!addr) lose("Can't allocate %#"OBJ_FMTX" bytes for space %d", size, space_id);
+
+    if (extra_below) { // it contains at least alien linkage if not also lisp linkage
+        if (extra_below > ALIEN_LINKAGE_SPACE_SIZE) linkage_space = (void*)addr;
+        addr += extra_below;
+        ALIEN_LINKAGE_SPACE_START = (uword_t)addr - ALIEN_LINKAGE_SPACE_SIZE;
+    }
+    return addr;
+}
+
 #include "x86-arch-shared.inc"
