@@ -40,12 +40,11 @@ rm -fr sandbox/scratch
  "PRINT.BACKQUOTE.RANDOM.2" "PRINT.BACKQUOTE.RANDOM.3"
  "PRINT.BACKQUOTE.RANDOM.4" "PRINT.BACKQUOTE.RANDOM.5" "PROCLAIM.ERROR.7"
  "SHIFTF.7"
- "SUBTYPEP-COMPLEX.8"
  "SXHASH.17" "SXHASH.18" "SXHASH.19" "PRINT-STRUCTURE.1"
- "UNION.FOLD.1"
+ "UNION.FOLD.1" "SUBTYPEP-COMPLEX.8" "FORMAT./.17"
  (append #+win32 (list "PROBE-FILE.4")
-         #+x86 (list "CIS.4" "CEILING.16" "CEILING.18")
-         #+(and arm64 (not darwin))
+         #+x86 (list "CIS.4")
+         #+(or arm (and arm64 (not darwin)))
            (list "EXP.ERROR.4" "EXP.ERROR.5" "EXP.ERROR.6" "EXP.ERROR.7" "EXPT.ERROR.4"
                  "EXPT.ERROR.5" "EXPT.ERROR.6" "EXPT.ERROR.7")
          #-sb-unicode (list "MISC.638")
