@@ -623,3 +623,100 @@
    (lambda (a)
      (* a #c(1d0 0d0)))
     (complex double-float)))
+
+(with-test (:name :bignum-ash-modarith)
+  (checked-compile-and-assert
+      ()
+      `(lambda (a)
+         (declare (bignum a))
+         (logand (ash a -1) 1))
+    (((expt 2 129)) 0)
+    (((+ (expt 2 129) 2)) 1)))
+
+(with-test (:name :two-fixnum-eq-mask)
+  (checked-compile-and-assert
+      ()
+      `(lambda (x)
+         (declare (fixnum x))
+         (typep x '(member -1 ,most-positive-fixnum)))
+    ((-1) t)
+    ((most-positive-fixnum) t)
+    ((1) nil)
+    ((0) nil)
+    ((most-negative-fixnum) nil)
+    (((1+ most-negative-fixnum)) nil)
+    (((1- most-positive-fixnum)) nil)))
+
+(with-test (:name :logior-derive-negative)
+  (assert-type
+   (lambda (x)
+     (declare (fixnum x))
+     (logior x 1))
+   (or (integer #.(1+ most-negative-fixnum) -1)
+       (integer 1 #.most-positive-fixnum)))
+  (assert-type
+   (lambda (x y)
+     (declare (fixnum x)
+              ((and (signed-byte 16) (not (eql 0))) y))
+     (logior y x))
+   (or (integer #.(1+ most-negative-fixnum) -1)
+       (integer 1 #.most-positive-fixnum)))
+  (assert-type
+   (lambda (x)
+     (declare (fixnum x))
+     (logior x 10))
+   (or (integer #.(+ most-negative-fixnum 10) -1)
+       (integer 10 #.most-positive-fixnum))))
+
+(with-test (:name :range<=low-address)
+  (checked-compile-and-assert
+   ()
+   `(lambda (a)
+      (typep a '(integer 1 13)))
+    ((0.0) nil)
+    ((0) nil)
+    ((1) t)
+    ((6) t)
+    ((13) t)
+    ((14) nil)))
+
+(with-test (:name :ratio+integer-type)
+  (assert-type
+   (lambda (x r)
+     (declare (integer x)
+              (ratio r))
+     (+ x r))
+   ratio)
+  (assert-type
+   (lambda (x r)
+     (declare (integer x)
+              (ratio r))
+     (+ r x))
+   ratio)
+  (assert-type
+   (lambda (x r)
+     (declare (integer x)
+              (ratio r))
+     (/ r x))
+   ratio))
+
+(with-test (:name :fixnum-gcd-overflow)
+  (checked-compile-and-assert
+      ()
+      `(lambda (a b)
+         (gcd a b))
+    ((most-negative-fixnum most-negative-fixnum) (- most-negative-fixnum))
+    ((most-negative-fixnum 48) 16)))
+
+(with-test (:name :signed-word-minus1-division)
+  (checked-compile-and-assert
+      ()
+      `(lambda (a b)
+         (truncate
+          (the sb-vm:signed-word a)
+          (the (member -8 -1) b)))
+    ((-2 -1) (values 2 0))
+    ((-2 -8) (values 0 -2))
+    (((- #1=(expt 2 (1- sb-vm:n-word-bits))) -1) (values #1# 0))
+    (((- #1#) -8) (values (- (ash (- #1#) -3))
+                          0))))
