@@ -22,11 +22,11 @@
               (2 2)
               (3 3))
 
-        ("Duplicate key 1 in CASE form, occurring in the first clause:  (1 1), and the second clause:  (1 2)"
+        ("Duplicate key 1 in CASE form, occurring in clause 1:  (1 1), and clause 2:  (1 2)"
          case (1 1)
               (1 2))
 
-        ("Duplicate key 2 in CASE form, occurring in the first clause:  ((1 2) 1), and the second clause:  ((2 3) 2)"
+        ("Duplicate key 2 in CASE form, occurring in clause 1:  ((1 2) 1), and clause 2:  ((2 3) 2)"
          case ((1 2) 1)
               ((2 3) 2))
 
@@ -71,3 +71,9 @@
                          (a 1)
                          (2 10))))
                    :allow-notes nil))
+
+(with-test (:name :duplicate-typecase)
+  (assert (nth-value 3
+                     (checked-compile '(lambda (x)
+                                        (typecase x (number 1) (integer 2)))
+                                      :allow-style-warnings t))))
