@@ -139,7 +139,7 @@
 
 (defparameter *breakpoint-tracing-expectations*
   '(:fails-on :arm
-    :broken-on (or :freebsd :ppc :ppc64)))
+    :broken-on (or :freebsd)))
 
 ;;; bug 379
 (with-test (:name (trace :encapsulate nil)
@@ -456,6 +456,51 @@
                  '((0 (flet body :in (compiler-macro fn-with-cmac)) :enter)
                    (0 (flet body :in (compiler-macro fn-with-cmac)) :exit 42)))))
 
+(defun fn-known-values-return (x)
+  (flet ((g (x)
+           x))
+    (g x)
+    (g x)
+    (g x)))
+
+(with-test (:name (trace :flet :known-values-return)
+                  . #.*breakpoint-tracing-expectations*)
+  (assert (equal (collecting-traces ((flet g :in fn-known-values-return))
+                   (fn-known-values-return 9))
+                 '((0 (flet g :in fn-known-values-return) :enter 9)
+                   (0 (flet g :in fn-known-values-return) :exit 9)
+                   (0 (flet g :in fn-known-values-return) :enter 9)
+                   (0 (flet g :in fn-known-values-return) :exit 9)
+                   (0 (flet g :in fn-known-values-return) :enter 9)
+                   (0 (flet g :in fn-known-values-return) :exit 9)))))
+
+(defun fn-known-values-return-multiple (x)
+  (flet ((g (x)
+           (values x x x x x x x)))
+    (g x)
+    (g x)
+    (g x)))
+
+(with-test (:name (trace :flet :known-values-return-multiple)
+                  . #.*breakpoint-tracing-expectations*)
+  (assert
+   (equal
+    (collecting-traces ((flet g :in fn-known-values-return-multiple))
+      (fn-known-values-return-multiple 9))
+    '((0 (flet g :in fn-known-values-return-multiple) :enter 9)
+      (0 (flet g :in fn-known-values-return-multiple) :exit 9 9 9 9 9 9 9)
+      (0 (flet g :in fn-known-values-return-multiple) :enter 9)
+      (0 (flet g :in fn-known-values-return-multiple) :exit 9 9 9 9 9 9 9)
+      (0 (flet g :in fn-known-values-return-multiple) :enter 9)
+      (0 (flet g :in fn-known-values-return-multiple) :exit 9 9 9 9 9 9 9)))))
+
+(with-test (:name (trace :flet :within-compiler-macro)
+                  . #.*breakpoint-tracing-expectations*)
+  (assert (equal (collecting-traces ((flet body :in (compiler-macro fn-with-cmac)))
+                                    (compile nil '(lambda () (fn-with-cmac 0))))
+                 '((0 (flet body :in (compiler-macro fn-with-cmac)) :enter)
+                   (0 (flet body :in (compiler-macro fn-with-cmac)) :exit 42)))))
+
 (defun call-with-compiler-macro-redefined (fn)
   (eval `(define-compiler-macro fn-with-cmac (x)
            (declare (ignore x) (optimize (debug 3)))
