@@ -33,10 +33,14 @@
                         ;; the start of static space, but it holds a zero.
                         (simple-vector 2))))))
     (assert (not (funcall f (sb-kernel:make-unbound-marker)))))
-  (assert (not (equalp (sb-kernel:make-unbound-marker) "")))
-  (let ((a (- (sb-kernel:get-lisp-obj-address (sb-kernel:make-unbound-marker))
-              sb-vm:other-pointer-lowtag)))
-    (assert (> a sb-vm:static-space-start))))
+  (let ((ubm (opaque-identity (sb-kernel:make-unbound-marker)))
+        (thing (opaque-identity "")))
+    (assert (not (eql thing ubm)))
+    (assert (not (equal thing ubm)))
+    (assert (not (equalp thing ubm)))
+    (assert (not (eql ubm thing)))
+    (assert (not (equal ubm thing)))
+    (assert (not (equalp ubm thing)))))
 
 (sb-vm::define-vop (tryme)
   (:generator 1 (sb-assem:inst mov :byte (sb-vm::ea :gs sb-vm::rax-tn) 0)))
@@ -175,7 +179,7 @@
          (c-call (find "os_deallocate" lines :test #'search)))
     ;; Depending on #+immobile-code it's either direct or memory indirect.
     #+immobile-code (assert (search "CALL #x" c-call))
-    #-immobile-code (assert (search "CALL [#x" c-call))))
+    #-immobile-code (assert (search "LEA RBX, [R12-" c-call))))
 
 (with-test (:name :set-symbol-value-imm)
   (let (success)
@@ -1365,3 +1369,38 @@
     (assert
      (loop for line in lines
            thereis (and (search "PUSH #x" line) (search "':FOO" line))))))
+
+(defun check-hasnot (instructions lexpr)
+  (let ((f (compile nil lexpr)))
+    (dolist (inst (get-simple-fun-instruction-model f))
+      (let ((mnemonic (second inst)))
+        (assert (not (find mnemonic instructions :test 'string=)))))))
+
+(with-test (:name :compute-lisp-bool-from-c-bool :skipped-on (:not :sb-thread))
+  (check-hasnot '("JMP")
+                '(lambda (x)
+                  (declare (optimize (sb-c::verify-arg-count 0)))
+                  (= (truly-the fixnum x) 1))))
+
+(with-test (:name :compute-c-bool-from-lisp-bool)
+  (check-hasnot '("JMP" "CMOV")
+                '(lambda (x)
+                  (declare (optimize (sb-c::verify-arg-count 0)))
+                  (if x 1 0))))
+
+(with-test (:name :pcl-ctor-slotv-allocator)
+  (flet ((get-asm-lines (n)
+           (disassembly-lines
+            (compile nil
+             `(lambda () (make-array ,n :initial-element sb-pcl:+slot-unbound+))))))
+    (let ((lines (get-asm-lines 1)))
+      (assert (loop for line in lines
+                    thereis (and (search "MOV QWORD PTR" line)
+                                 (search "], 9" line))))) ; UNBOUND-MARKER-WIDETAG
+    (loop for nelts from 2 to 10
+          do (let ((lines (get-asm-lines nelts)))
+               (assert (loop for line in lines
+                             thereis (search "MOVDDUP XMM" line)))))
+    (let ((lines (get-asm-lines 11)))
+      (assert (loop for line in lines
+                    thereis (search "REPE STOSQ" line))))))
