@@ -145,10 +145,14 @@
 ;;; At execution time the function will have virtual address LOAD-ADDR.
 #+x86-64
 (defun list-textual-instructions (sap length core load-addr emit-cfi)
-  (let ((insts (simple-collect-inst-model sap length load-addr))
-        (alien-linkage-end
-         (+ (bounds-low (core-linkage-bounds core)) alien-linkage-space-size))
-        (result))
+  (let* ((insts (simple-collect-inst-model sap length load-addr))
+         (alien-linkage-bounds
+          (make-bounds (- (bounds-high (core-linkage-bounds core)) alien-linkage-space-size)
+                       (bounds-high (core-linkage-bounds core))))
+         (lisp-linkage-bounds
+          (make-bounds (bounds-low (core-linkage-bounds core))
+                       (bounds-low alien-linkage-bounds)))
+         (result))
     (flet ((pc-relative-ea-p (x)
              (when (consp x) (setq x (car x)))
              (and (typep x 'machine-ea) (eq (machine-ea-base x) :rip)))
@@ -167,40 +171,41 @@
                     (pc-relative-ea-p ea))
             (cond ((and (integerp ea) (in-bounds-p ea (core-linkage-bounds core)))
                    (aver (eq (cadr inst) 'sb-x86-64-asm::call))
-                   (aver (< ea alien-linkage-end)) ; CALL via alien linkage
+                   (aver (in-bounds-p ea alien-linkage-bounds)) ; CALL via alien linkage
                    (aver (= inst-len 5))
                    (push (list* (car inst) 5 :call ea) result))
                   ((pc-relative-ea-p ea)
                    (let* ((next-pc (+ load-addr (caadr insts)))
                           (ea (+ next-pc (ea-disp-of ea)))
-                          (table-offset (- ea alien-linkage-end)))
-                     (when (in-bounds-p ea (core-linkage-bounds core))
-                       (cond ((< ea alien-linkage-end) ; alien linkage
-                              (aver (= inst-len 7))
-                              (let ((op (ecase (cadr inst)
-                                          (sb-x86-64-asm::mov :mov-gotpcrel)
-                                          (sb-x86-64-asm::lea :lea)))
-                                    (args (list ea (string-downcase (princ-to-string (third inst))))))
-                                (push (list* (car inst) 7 op args) result)))
-                             ((eq (cadr inst) 'sb-x86-64-asm::lea)
+                          (table-offset (- ea (bounds-low lisp-linkage-bounds))))
+                     (cond ((in-bounds-p ea alien-linkage-bounds)
+                            (aver (= inst-len 7))
+                            (let ((op (ecase (cadr inst)
+                                        (sb-x86-64-asm::mov :mov-gotpcrel)
+                                        (sb-x86-64-asm::lea :lea)))
+                                  (args (list ea (string-downcase (princ-to-string (third inst))))))
+                              (push (list* (car inst) 7 op args) result)))
+                           ((and (in-bounds-p ea lisp-linkage-bounds)
+                                 (eq (cadr inst) 'sb-x86-64-asm::lea))
                               ;; Get ADDRESS of lisp linkage cell in stepping-enabled code
-                              (aver (eq (third inst) (get-gpr :qword 0))) ; %rax
-                              (aver (= inst-len 7))
-                              (push (list* (car inst) 7 :lea (format nil "(fntbl+~d)(%rip),%rax"
-                                                                      table-offset))
-                                    result))
-                             (t ; lisp CALL or JMP
-                              (aver (= inst-len 6))
-                              (let ((new-inst
-                                     (format nil "~a *(fntbl+~d)(%rip)"
-                                             (string-downcase (cadr inst))
-                                             table-offset)))
-                                (push (list* (car inst) 6 :lispcall new-inst) result))))))))))))
+                            (aver (eq (third inst) (get-gpr :qword 0))) ; %rax
+                            (aver (= inst-len 7))
+                            (push (list* (car inst) 7 :lea (format nil "(fntbl+~d)(%rip),%rax"
+                                                                   table-offset))
+                                  result))
+                           ((in-bounds-p ea lisp-linkage-bounds) ; lisp CALL or JMP
+                            (aver (= inst-len 6))
+                            (let ((new-inst
+                                   (format nil "~a *(fntbl+~d)(%rip)"
+                                           (string-downcase (cadr inst))
+                                           table-offset)))
+                              (push (list* (car inst) 6 :lispcall new-inst) result)))))))))))
       (nreverse result)))
 
 (defun c-linkage-sym-from-addr (addr core)
-  (let ((entry-index (/ (- addr (bounds-low (core-linkage-bounds core)))
-                        (core-alien-linkage-entry-size core))))
+  ;; assumption: alien-linkage-table-growth-direction is :UP for the platform
+  (let* ((alien-ls-start (- (bounds-high (core-linkage-bounds core)) alien-linkage-space-size))
+         (entry-index (/ (- addr alien-ls-start) (core-alien-linkage-entry-size core))))
     (setf (bit (core-alien-linkage-symbol-usedp core) entry-index) 1)
     (let ((symbol (aref (core-alien-linkage-symbols core) entry-index)))
       (if (listp symbol)
@@ -1116,17 +1121,6 @@ lisp_fun_linkage_space: .zero ~:*~D
              ;; boxed objects that can reference code/simple-funs
              ((#.value-cell-widetag #.symbol-widetag #.weak-pointer-widetag)
               (scanptrs vaddr obj 1 (1- nwords))))))
-      ;; Fix the assembly routine entrypoint vector in static space.
-      ;; It's an unboxed vector so it would not get adjusted automatically.
-      (let ((sym (%find-target-symbol (package-id "SB-FASL") "*ASM-ROUTINE-VECTOR*"
-                                      spacemap)))
-        (when sym
-          (let* ((obj (symbol-global-value sym)) ; a tagged ptr to static space
-                 (physical-obj (translate obj spacemap))
-                 (vaddr (logandc2 (get-lisp-obj-address obj) lowtag-mask)))
-            (scanptrs vaddr physical-obj sb-vm:vector-data-offset
-                      (+ sb-vm:vector-data-offset (1- (length obj)))
-                      t)))) ; T -> do not require is-lisp-pointer of each word
       (dolist (space (cdr spacemap))
         (unless (= (space-id space) immobile-text-core-space-id)
           (let* ((logical-addr (space-addr space))
@@ -1180,6 +1174,7 @@ lisp_fun_linkage_space: .zero ~:*~D
           (page-adjust 0)
           (linkage-space-info (vector 0 0 0 0 0))
           (code-start-fixup-ofs 0) ; where to fixup the core header
+          (static-consts-fixup-ofs 0)
           (space-list)
           (copy-actions)
           (fixedobj-range) ; = (START . SIZE-IN-BYTES)
@@ -1189,10 +1184,6 @@ lisp_fun_linkage_space: .zero ~:*~D
   ;; Remove old files
   (ignore-errors (delete-file asm-pathname))
   (ignore-errors (delete-file elf-core-pathname))
-  ;; Crappy assumption: we already called redirect-text-space-calls which
-  ;; assigned a value to *nil-taggedptr*. Anyway, why doesn't this use
-  ;; PARSE-CORE-HEADER like it should?
-  (assert (/= *nil-taggedptr* 0))
   ;; Ensure that all files can be opened
   (with-open-file (input input-pathname :element-type '(unsigned-byte 8))
     (with-open-file (asm-file asm-pathname :direction :output :if-exists :supersede)
@@ -1241,6 +1232,8 @@ lisp_fun_linkage_space: .zero ~:*~D
                                (make-array count :element-type 'word)))
                  (push (cons data-page (* npages +backend-page-bytes+)) copy-actions)
                  (incf original-total-npages npages))))
+            (#.static-constants-core-entry-type-code
+             (setq static-consts-fixup-ofs ptr))
             (#.page-table-core-entry-type-code
              (aver (= len 3))
              (symbol-macrolet ((n-ptes (%vector-raw-bits core-header (+ ptr 0)))
@@ -1305,6 +1298,7 @@ lisp_fun_linkage_space: .zero ~:*~D
                  (/= (%vector-raw-bits core-header 1) runtime-options-magic))
         (let ((added-words 5))
           (incf (linkage-space-header-ptr linkage-space-info) added-words)
+          (incf static-consts-fixup-ofs added-words)
           (incf code-start-fixup-ofs added-words)))
       (unless enable-pie
         ;; This fixup sets the 'address' field of the core directory entry
@@ -1322,21 +1316,36 @@ lisp_fun_linkage_space: .zero ~:*~D
                 (delete immobile-text-core-space-id (reverse space-list)
                         :key #'space-id))
                (spacemap (cons sap (sort (copy-list space-list) #'> :key #'space-addr)))
+               (new-header (change-dynamic-space-size core-header dynamic-space-size))
                (pte-nbytes (cdar copy-actions)))
           (collect-relocations spacemap relocs enable-pie)
           (with-open-file (output elf-core-pathname
                                   :direction :output :if-exists :supersede
                                   :element-type '(unsigned-byte 8))
+            #+x86-64 ; adjust the words of the core header that fill in *ASM-ROUTINE-VECTOR*
+            (let ((code-start (bounds-low (space-bounds immobile-text-core-space-id spacemap)))
+                  (static-asm-jmpvec-nelts
+                   (ash (%vector-raw-bits new-header (1+ static-consts-fixup-ofs))
+                        (- n-fixnum-tag-bits))))
+              (aver (= (%vector-raw-bits new-header static-consts-fixup-ofs)
+                       simple-array-unsigned-byte-64-widetag))
+              (loop for i from (+ static-consts-fixup-ofs vector-data-offset)
+                    repeat static-asm-jmpvec-nelts
+                    do (let ((word (%vector-raw-bits new-header i)))
+                         (unless (= word 0)
+                           (setf (%vector-raw-bits new-header i) 0)
+                           (vector-push-extend
+                            `#(,R_ABS64 ,(ash i word-shift) 1 ,(- word code-start))
+                            relocs)))))
             (prepare-elf (+ (apply #'+ (mapcar #'space-nbytes-aligned data-spaces))
                             (* (linkage-space-npages linkage-space-info) +backend-page-bytes+)
                             +backend-page-bytes+ ; core header
                             pte-nbytes)
                          relocs output enable-pie)
-            (let ((new-header (change-dynamic-space-size core-header dynamic-space-size)))
-              ;; This word will be fixed up by the system linker
-              (setf (%vector-raw-bits new-header code-start-fixup-ofs)
-                    (if enable-pie +code-space-nominal-address+ 0))
-              (write-sequence new-header output))
+            ;; This word will be fixed up by the system linker
+            (setf (%vector-raw-bits new-header code-start-fixup-ofs)
+                  (if enable-pie +code-space-nominal-address+ 0))
+            (write-sequence new-header output)
             (force-output output)
             ;; ELF cores created from #-immobile-space cores use +required-foreign-symbols+.
             ;; But if #+immobile-space the alien-linkage-table values are computed
@@ -1472,7 +1481,7 @@ lisp_fun_linkage_space: .zero ~:*~D
                       (unless (plusp (decf remaining n)) (return))))
               (aver (zerop remaining)))))))))
 
-;; The extra copy of ASM routines, particularly C-calling trampolines, that now reside in text
+;; The extra copy of ASM routines, particularly C-calling trampolines, that will reside in text
 ;; space have to be modified to correctly reference their C functions. They assume that static
 ;; space is near alien-linkage space, and so they use this form:
 ;;   xxxx: E8A1F0EFFF  CALL #x50000060 ; alloc
@@ -1483,8 +1492,8 @@ lisp_fun_linkage_space: .zero ~:*~D
 ;; but is callable from the text-space code.
 ;;; I don't feel like programmatically scanning the asm code to determine these.
 ;;; Hardcoded is good enough (until it isn't)
-(defparameter *c-linkage-redirects*
-  (mapcar #'list
+(defglobal *c-linkage-redirects*
+  (mapcar (lambda (x) (cons nil (concatenate 'string #+darwin "_" x)))
           '("switch_to_arena"
             "alloc"
             "alloc_list"
@@ -1505,7 +1514,6 @@ lisp_fun_linkage_space: .zero ~:*~D
              (header-bytes (ash (code-header-words old-code) word-shift))
              (old-insts-vaddr (+ old-code-vaddr header-bytes))
              (new-insts-vaddr (+ new-code-vaddr header-bytes))
-             (items *c-linkage-redirects*)
              (inst-buffer (make-array 8 :element-type '(unsigned-byte 8)))
              (code-offsets-vector-size (primitive-object-size code-offsets-vector))
              (c-linkage-vector-vaddr (+ (space-addr text-space) code-offsets-vector-size))
@@ -1513,52 +1521,49 @@ lisp_fun_linkage_space: .zero ~:*~D
               (%make-lisp-obj (logior (sap-int (sap+ (space-physaddr text-space spacemap)
                                                      code-offsets-vector-size))
                                       other-pointer-lowtag)))
-             (target-alien-lss ; linkage-space-start
-              (symbol-global-value
-               (find-target-symbol (package-id "SB-VM") "ALIEN-LINKAGE-SPACE-START"
-                                   spacemap))))
-    (aver (<= (length items) (length c-linkage-vector)))
+             (alien-ls-start (- (space-addr (or text-space static-space))
+                                alien-linkage-space-size))
+             (alien-ls-end (1- (+ alien-ls-start alien-linkage-space-size))))
+    (aver (<= (length *c-linkage-redirects*) (length c-linkage-vector)))
+    ;; (format t "~&alien linkage range = ~x .. ~x~%" alien-ls-start alien-ls-end)
     (dolist (x *c-linkage-redirects*)
-      (let* ((index (position (car x) (core-alien-linkage-symbols core)
+      (let* ((index (position (cdr x) (core-alien-linkage-symbols core)
                               :test (lambda (a b) (and (stringp b) (string= a b)))))
-             (addr (+ target-alien-lss (* (core-alien-linkage-entry-size core) index))))
-        (rplacd x (int-sap addr))))
+             (addr (+ alien-ls-start (* (core-alien-linkage-entry-size core) index))))
+        (rplaca x addr)))
     (with-pinned-objects (inst-buffer)
       (do ((sap (vector-sap inst-buffer))
            (item-index 0 (1+ item-index))
-           (items items (cdr items)))
+           (items *c-linkage-redirects* (cdr items)))
           ((null items))
         ;; Each new quasi-linkage-table entry takes 8 bytes to encode.
         ;; The JMP is 7 bytes, followed by a nop.
         ;; FF2425nnnnnnnn = JMP [ea]
-        (setf (sap-ref-8 sap 0) #xFF
-              (sap-ref-8 sap 1) #x24
-              (sap-ref-8 sap 2) #x25
-              (sap-ref-32 sap 3) (sap-int (sap+ (cdar items) 8))
+        (setf (sap-ref-32 sap 0) #x2524FF
+              (sap-ref-32 sap 3) (+ (caar items) 8) ; addr within the real alien linkage entry
               (sap-ref-8 sap 7) #x90) ; nop
         (setf (aref c-linkage-vector item-index) (%vector-raw-bits inst-buffer 0))))
     ;; Produce a model of the instructions. It doesn't really matter whether we scan
     ;; OLD-CODE or NEW-CODE since we're supplying the proper virtual address either way.
-    (let ((insts (get-code-instruction-model old-code old-code-vaddr core)))
-;;  (dovector (inst insts) (write inst :base 16 :pretty nil :escape nil) (terpri))
-      (dovector (inst insts)
-        ;; Look for any call to a linkage table entry.
+    (let* ((skip (ash (code-jump-table-words old-code) word-shift))
+           (insts-start (sap+ (code-instructions old-code) skip))
+           (insts-len (- (%code-text-size old-code) skip))
+           (insts (simple-collect-inst-model insts-start insts-len (+ old-insts-vaddr skip)))
+           (textspace-insts-sap (code-instructions new-code)))
+      (dolist (inst insts)
+        ;; Look for any call to an alien linkage table entry.
         (when (eq (second inst) 'call)
           (let ((operand (third inst)))
-            (when (and (integerp operand)
-                       (>= operand target-alien-lss)
-                       (< operand (+ target-alien-lss alien-linkage-space-size)))
-              (let* ((index (position (int-sap operand) *c-linkage-redirects*
-                                      :key #'cdr :test #'sap=))
+            (when (and (integerp operand) (<= alien-ls-start operand alien-ls-end))
+              (let* ((index (position operand *c-linkage-redirects* :key #'car))
                      (branch-target (+ c-linkage-vector-vaddr
                                        (ash vector-data-offset word-shift)
                                        ;; each new linkage entry takes up exactly 1 word
                                        (* index n-word-bytes)))
-                     (old-next-ip-abs (int-sap (inst-end inst))) ; virtual
-                     (next-ip-rel (sap- old-next-ip-abs (int-sap old-insts-vaddr)))
-                     (new-next-ip (+ new-insts-vaddr next-ip-rel)))
-                (setf (signed-sap-ref-32 (code-instructions new-code) (- next-ip-rel 4))
-                      (- branch-target new-next-ip))))))))))
+                     (rel-pc (+ skip (+ (car inst) 1))))
+                (setf (signed-sap-ref-32 textspace-insts-sap rel-pc)
+                      ;; I played with this math until it came out right.
+                      (- branch-target (+ new-insts-vaddr rel-pc 4)))))))))))
 
 (defconstant smallvec-elts
   (- (ash gencgc-page-bytes (- word-shift)) vector-data-offset))
@@ -1725,6 +1730,9 @@ lisp_fun_linkage_space: .zero ~:*~D
          (progn
            (ecase *heap-arrangement*
              (:gencgc
+              ;; FIXME: should depend on target features but oh well
+              #-immobile-space (move-dynamic-code-to-text-space input-pathname tmp)
+              #+immobile-space
               ;; input core could be readonly
               (run-program "cp" `("--no-preserve=mode" ,input-pathname ,tmp)
                            :search t))
